# 13 "tools/coqdoc/cpretty.mll"
 
  open Printf
  open Lexing

  (* A list function we need *)
  let rec take n ls =
    if n = 0 then [] else
      match ls with
        | [] -> []
        | (l :: ls) -> l :: (take (n-1) ls)

  (* count the number of spaces at the beginning of a string *)
  let count_spaces s =
    let n = String.length s in
    let rec count c i =
      if i == n then c,i else match s.[i] with
	| '\t' -> count (c + (8 - (c mod 8))) (i + 1)
	| ' ' -> count (c + 1) (i + 1)
	| _ -> c,i
    in
      count 0 0

  let remove_newline s =
    let n = String.length s in
    let rec count i = if i == n || s.[i] <> '\n' then i else count (i + 1) in
    let i = count 0 in
    i, String.sub s i (n - i)

  let count_dashes s =
    let c = ref 0 in
    for i = 0 to String.length s - 1 do if s.[i] = '-' then incr c done;
    !c

  let cut_head_tail_spaces s =
    let n = String.length s in
    let rec look_up i = if i == n || s.[i] <> ' ' then i else look_up (i+1) in
    let rec look_dn i = if i == -1 || s.[i] <> ' ' then i else look_dn (i-1) in
    let l = look_up 0 in
    let r = look_dn (n-1) in
    if l <= r then String.sub s l (r-l+1) else s

  let sec_title s =
    let rec count lev i =
      if s.[i] = '*' then
	count (succ lev) (succ i)
      else
	let t = String.sub s i (String.length s - i) in
	lev, cut_head_tail_spaces t
    in
    count 0 (String.index s '*')

  let strip_eol s =
    let eol = s.[String.length s - 1] = '\n' in
    (eol, if eol then String.sub s 1 (String.length s - 1) else s)


  let formatted = ref false
  let brackets = ref 0
  let comment_level = ref 0
  let in_proof = ref None

  let in_env start stop = 
    let r = ref false in
    let start_env () = r := true; start () in
    let stop_env () = if !r then stop (); r := false in
      (fun x -> !r), start_env, stop_env
      
  let _, start_emph, stop_emph = in_env Output.start_emph Output.stop_emph
  let in_quote, start_quote, stop_quote = in_env Output.start_quote Output.stop_quote

  let url_buffer = Buffer.create 40
  let url_name_buffer = Buffer.create 40

  let backtrack lexbuf = lexbuf.lex_curr_pos <- lexbuf.lex_start_pos;
    lexbuf.lex_curr_p <- lexbuf.lex_start_p

  let backtrack_past_newline lexbuf =
    let buf = lexeme lexbuf in
    let splits = Str.bounded_split_delim (Str.regexp "['\n']") buf 2 in
      match splits with
        | [] -> ()
        | (_ :: []) -> ()
        | (s1 :: rest :: _) ->
            let length_skip = 1 + String.length s1 in
              lexbuf.lex_curr_pos <- lexbuf.lex_start_pos + length_skip

  (* saving/restoring the PP state *)

  type state = {
    st_gallina : bool;
    st_light : bool
  }

  let state_stack = Stack.create ()

  let save_state () =
    Stack.push { st_gallina = !Cdglobals.gallina; st_light = !Cdglobals.light } state_stack

  let restore_state () =
    let s = Stack.pop state_stack in
    Cdglobals.gallina := s.st_gallina;
    Cdglobals.light := s.st_light

  let begin_show () = save_state (); Cdglobals.gallina := false; Cdglobals.light := false
  let end_show () = restore_state ()

  (* Reset the globals *)

  let reset () =
    formatted := false;
    brackets := 0;
    comment_level := 0

  (* erasing of Section/End *)

  let section_re = Str.regexp "[ \t]*Section"
  let end_re = Str.regexp "[ \t]*End"
  let is_section s = Str.string_match section_re s 0
  let is_end s = Str.string_match end_re s 0

  let sections_to_close = ref 0

  let section_or_end s =
    if is_section s then begin
      incr sections_to_close; true
    end else if is_end s then begin
      if !sections_to_close > 0 then begin
	decr sections_to_close; true
      end else
	false
    end else
      true

  (* for item lists *)
  type list_compare =
      | Before
      | StartLevel of int
      | InLevel of int * bool

  (* Before : we're before any levels
     StartLevel : at the same column as the dash in a level
     InLevel : after the dash of this level, but before any deeper dashes.
               bool is true if this is the last level *)
  let find_level levels cur_indent =
    match levels with
    | [] -> Before
    | (l::ls) ->
        if cur_indent < l then Before
        else
          (* cur_indent will never be less than the head of the list *)
          let rec findind ls n =
            match ls with
            | [] -> InLevel (n,true)
            | (l :: []) -> if cur_indent = l then StartLevel n
                           else InLevel (n,true)
            | (l1 :: l2 :: ls) ->
                if cur_indent = l1 then StartLevel n
                else if cur_indent < l2 then InLevel (n,false)
                     else findind (l2 :: ls) (n+1)
          in
            findind (l::ls) 1

  type is_start_list =
    | Rule
    | List of int
    | Neither

  let check_start_list str =
    let n_dashes = count_dashes str in
    let (n_spaces,_) = count_spaces str in
      if n_dashes >= 4 && not !Cdglobals.plain_comments then
        Rule
      else
        if n_dashes = 1 && not !Cdglobals.plain_comments then
          List n_spaces
        else
          Neither

   (* examine a string for subtitleness *)
  let subtitle m s =
    match Str.split_delim (Str.regexp ":") s with
    | [] -> false
    | (name::_) ->
        if (cut_head_tail_spaces name) = m then
          true
        else
          false


  (* tokens pretty-print *)

  let token_buffer = Buffer.create 1024

  let token_re =
    Str.regexp "[ \t]*(\\*\\*[ \t]+printing[ \t]+\\([^ \t]+\\)"
  let printing_token_re =
    Str.regexp
      "[ \t]*\\(\\(%\\([^%]*\\)%\\)\\|\\(\\$[^$]*\\$\\)\\)?[ \t]*\\(#\\(\\(&#\\|[^#]\\)*\\)#\\)?"

  let add_printing_token toks pps =
    try
      if Str.string_match token_re toks 0 then
	let tok = Str.matched_group 1 toks in
	if Str.string_match printing_token_re pps 0 then
	  let pp =
	    (try Some (Str.matched_group 3 pps) with _ ->
             try Some (Str.matched_group 4 pps) with _ -> None),
	    (try Some (Str.matched_group 6 pps) with _ -> None)
	  in
	  Output.add_printing_token tok pp
    with _ ->
      ()

  let remove_token_re =
    Str.regexp
      "[ \t]*(\\*\\*[ \t]+remove[ \t]+printing[ \t]+\\([^ \t]+\\)[ \t]*\\*)"

  let remove_printing_token toks =
    try
      if Str.string_match remove_token_re toks 0 then
	let tok = Str.matched_group 1 toks in
	Output.remove_printing_token tok
    with _ ->
      ()

  let output_indented_keyword s lexbuf =
    let nbsp,isp = count_spaces s in
    Output.indentation nbsp;
    let s = String.sub s isp (String.length s - isp) in
    Output.keyword s (lexeme_start lexbuf + isp)

  let only_gallina () =
    !Cdglobals.gallina && !in_proof <> None

  let parse_comments () =
    !Cdglobals.parse_comments && not (only_gallina ())


# 241 "tools/coqdoc/cpretty.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\235\255\236\255\000\000\000\000\000\000\001\000\000\000\
    \000\000\002\000\002\000\000\000\007\000\005\000\013\000\017\000\
    \017\000\015\000\004\000\034\000\032\000\026\000\029\000\046\000\
    \000\000\001\000\002\000\143\000\001\000\025\000\010\000\033\000\
    \049\000\054\000\057\000\062\000\050\000\073\000\082\000\063\000\
    \068\000\077\000\075\000\060\000\133\000\091\000\054\000\086\000\
    \135\000\001\000\005\000\227\000\222\000\053\000\066\000\185\000\
    \007\000\254\255\225\000\072\000\078\000\233\000\094\000\094\000\
    \089\000\101\000\097\000\115\000\009\000\012\000\013\000\014\000\
    \015\000\124\000\130\000\126\000\144\000\016\000\239\000\135\000\
    \146\000\142\000\137\000\150\000\148\000\157\000\017\000\018\000\
    \019\000\021\000\022\000\026\001\011\000\240\255\158\000\158\000\
    \156\000\024\000\013\001\168\000\168\000\176\000\178\000\037\001\
    \014\000\050\001\252\255\009\000\169\000\162\000\052\001\016\000\
    \061\001\251\255\019\000\182\000\027\000\053\001\180\000\174\000\
    \168\000\056\001\017\000\067\001\250\255\024\000\245\000\245\255\
    \178\000\232\000\002\001\175\000\242\000\180\000\217\000\223\000\
    \243\000\252\000\246\000\241\000\245\000\097\001\253\255\253\000\
    \000\001\012\001\255\000\249\000\248\255\251\000\006\001\008\001\
    \015\001\011\001\006\001\023\001\007\001\026\001\010\001\022\001\
    \012\001\030\001\031\001\032\001\035\001\019\001\019\001\023\001\
    \037\001\019\001\024\001\042\001\037\001\032\001\034\001\030\001\
    \031\001\046\001\249\255\033\001\038\001\040\001\046\001\053\001\
    \056\001\046\001\043\001\057\001\050\001\042\001\058\001\046\001\
    \053\001\064\001\066\001\064\001\069\001\059\001\069\001\162\001\
    \163\001\072\001\060\001\075\001\069\001\066\001\074\001\060\001\
    \080\001\068\001\072\001\075\001\073\001\078\001\073\001\085\001\
    \092\001\087\001\077\001\187\001\188\001\084\001\096\001\083\001\
    \092\001\101\001\093\001\088\001\090\001\090\001\110\001\108\001\
    \113\001\103\001\205\001\208\001\096\001\102\001\106\001\103\001\
    \101\001\116\001\108\001\122\001\112\001\114\001\112\001\111\001\
    \118\001\113\001\225\001\229\001\130\001\133\001\132\001\135\001\
    \141\001\150\001\132\001\138\001\154\001\142\001\148\001\144\001\
    \149\001\148\001\159\001\149\001\165\001\159\001\150\001\168\001\
    \161\001\170\001\172\001\168\001\177\001\159\001\175\001\180\001\
    \168\001\164\001\168\001\184\001\169\001\184\001\176\001\171\001\
    \191\001\176\001\173\001\190\001\190\001\182\001\186\001\195\001\
    \187\001\184\001\184\001\187\001\186\001\205\001\189\001\195\001\
    \193\001\194\001\215\001\197\001\205\001\200\001\217\001\208\001\
    \222\001\205\001\215\001\219\001\216\001\221\001\211\001\231\001\
    \215\001\230\001\067\002\068\002\229\001\234\001\220\001\226\001\
    \224\001\227\001\236\001\236\001\227\001\234\001\243\001\233\001\
    \244\001\231\001\083\002\084\002\234\001\250\001\240\001\250\001\
    \250\001\241\001\255\001\094\002\095\002\003\002\004\002\010\002\
    \254\001\250\001\254\001\014\002\254\001\013\002\007\002\002\002\
    \003\002\111\002\114\002\012\002\008\002\019\002\015\002\012\002\
    \024\002\020\002\028\002\123\002\124\002\026\002\025\002\025\002\
    \036\002\024\002\040\002\035\002\030\002\032\002\032\002\043\002\
    \031\002\051\002\039\002\039\002\056\002\043\002\051\002\048\002\
    \053\002\057\002\064\002\059\002\049\002\160\002\161\002\052\002\
    \065\002\054\002\063\002\072\002\065\002\060\002\062\002\079\002\
    \070\002\083\002\084\002\071\002\085\002\087\002\077\002\179\002\
    \181\002\086\002\071\002\081\002\101\002\089\002\101\002\100\002\
    \094\002\102\002\105\002\094\002\104\002\103\002\111\002\109\002\
    \100\002\208\002\113\002\111\002\109\002\105\002\124\002\108\002\
    \122\002\110\002\128\002\129\002\134\002\122\002\134\002\133\002\
    \136\002\120\002\120\002\138\002\138\002\235\002\236\002\120\002\
    \130\002\142\002\126\002\243\255\137\002\136\002\150\002\140\002\
    \137\002\149\002\141\002\163\002\153\002\158\002\160\002\167\002\
    \163\002\158\002\176\002\160\002\172\002\167\002\169\002\183\002\
    \173\002\017\003\018\003\185\002\181\002\185\002\178\002\187\002\
    \194\002\188\002\179\002\197\002\173\002\178\002\188\002\196\002\
    \033\003\034\003\189\002\204\002\202\002\198\002\214\002\196\002\
    \209\002\200\002\201\002\216\002\202\002\213\002\201\002\222\002\
    \059\003\060\003\206\002\211\002\214\002\232\002\231\002\215\002\
    \227\002\222\002\224\002\221\002\239\002\224\002\223\002\222\002\
    \240\002\242\002\079\003\080\003\232\002\230\002\252\002\234\002\
    \246\002\241\002\243\002\001\003\255\002\239\002\005\003\245\002\
    \251\002\249\002\254\002\006\003\011\003\010\003\253\002\009\003\
    \004\003\006\003\021\003\003\003\015\003\011\003\013\003\009\003\
    \017\003\014\003\016\003\012\003\026\003\030\003\017\003\032\003\
    \019\003\020\003\028\003\041\003\026\003\023\003\040\003\038\003\
    \041\003\048\003\030\003\042\003\037\003\039\003\035\003\244\255\
    \040\003\054\003\049\003\047\003\060\003\045\003\059\003\152\003\
    \154\003\054\003\054\003\066\003\050\003\060\003\068\003\055\003\
    \055\003\075\003\063\003\075\003\074\003\066\003\076\003\062\003\
    \079\003\246\255\077\003\072\003\075\003\075\003\076\003\090\003\
    \089\003\094\003\082\003\078\003\082\003\098\003\082\003\096\003\
    \088\003\083\003\089\003\087\003\101\003\094\003\088\003\100\003\
    \107\003\200\003\201\003\096\003\092\003\114\003\096\003\108\003\
    \103\003\105\003\115\003\109\003\114\003\105\003\122\003\103\003\
    \125\003\109\003\109\003\121\003\117\003\119\003\135\003\131\003\
    \173\003\138\003\142\003\131\003\126\003\128\003\126\003\132\003\
    \142\003\132\003\148\003\149\003\007\000\143\003\155\003\138\003\
    \139\003\139\003\158\003\142\003\142\003\146\003\160\003\161\003\
    \154\003\163\003\147\003\165\003\152\003\167\003\153\003\170\003\
    \157\003\167\003\163\003\171\003\161\003\175\003\175\003\179\003\
    \167\003\169\003\180\003\170\003\183\003\178\003\181\003\176\003\
    \172\003\029\004\247\255\030\004\181\003\192\003\182\003\195\003\
    \195\003\191\003\199\003\204\003\188\003\204\003\206\003\011\000\
    \196\003\197\003\197\003\213\003\202\003\047\004\048\004\197\003\
    \203\003\205\003\221\003\199\003\224\003\225\003\212\003\227\003\
    \226\003\225\003\238\003\241\003\253\003\227\003\251\003\243\003\
    \029\004\217\003\231\003\254\003\251\003\247\003\245\003\005\004\
    \254\003\255\003\251\003\255\003\025\004\009\004\034\004\026\004\
    \021\004\022\004\034\004\041\004\132\004\133\004\033\004\029\004\
    \054\004\036\004\048\004\043\004\045\004\055\004\060\004\051\004\
    \049\004\064\004\065\004\063\004\050\004\067\004\059\004\050\004\
    \072\004\056\004\061\004\063\004\070\004\066\004\061\004\079\004\
    \063\004\075\004\070\004\072\004\086\004\076\004\176\004\177\004\
    \088\004\084\004\088\004\081\004\090\004\095\004\091\004\082\004\
    \101\004\077\004\082\004\090\004\099\004\192\004\193\004\092\004\
    \108\004\106\004\101\004\117\004\099\004\112\004\103\004\104\004\
    \119\004\105\004\116\004\104\004\122\004\217\004\218\004\108\004\
    \113\004\116\004\134\004\133\004\117\004\129\004\124\004\126\004\
    \121\004\141\004\125\004\124\004\144\004\132\004\146\004\174\004\
    \132\004\142\004\141\004\147\004\142\004\146\004\147\004\159\004\
    \143\004\138\004\141\004\150\004\152\004\159\004\155\004\150\004\
    \167\004\151\004\171\004\155\004\167\004\155\004\173\004\175\004\
    \176\004\179\004\163\004\163\004\167\004\181\004\163\004\173\004\
    \191\004\186\004\181\004\183\004\179\004\180\004\185\004\187\004\
    \193\004\200\004\203\004\193\004\190\004\204\004\197\004\189\004\
    \205\004\193\004\208\004\195\004\205\004\201\004\201\004\211\004\
    \213\004\201\004\203\004\214\004\204\004\216\004\211\004\213\004\
    \209\004\205\004\228\004\212\004\228\004\230\004\012\000\220\004\
    \221\004\221\004\218\004\238\004\227\004\236\004\222\004\238\004\
    \226\004\226\004\231\004\246\004\236\004\247\004\249\004\239\004\
    \083\005\084\005\240\004\236\004\236\004\000\005\244\004\000\005\
    \255\004\246\004\003\005\251\004\255\004\252\004\002\005\248\004\
    \004\005\255\004\001\005\015\005\255\004\013\005\108\005\109\005\
    \008\005\010\005\006\005\006\005\026\005\014\005\026\005\025\005\
    \027\005\011\005\021\005\029\005\019\005\035\005\029\005\021\005\
    \038\005\008\000\029\005\041\005\024\005\027\005\026\005\045\005\
    \029\005\029\005\033\005\047\005\048\005\041\005\050\005\038\005\
    \037\005\049\005\044\005\046\005\059\005\055\005\095\005\060\005\
    \064\005\052\005\047\005\049\005\047\005\053\005\063\005\053\005\
    \069\005\074\005\072\005\056\005\077\005\061\005\067\005\066\005\
    \071\005\079\005\084\005\084\005\068\005\080\005\076\005\078\005\
    \092\005\074\005\086\005\081\005\084\005\080\005\088\005\077\005\
    \097\005\099\005\086\005\096\005\105\005\105\005\096\005\089\005\
    \105\005\092\005\096\005\108\005\111\005\104\005\101\005\097\005\
    \117\005\102\005\114\005\102\005\120\005\121\005\126\005\114\005\
    \126\005\126\005\129\005\113\005\113\005\131\005\131\005\224\005\
    \225\005\114\005\124\005\136\005\128\005\123\005\135\005\125\005\
    \121\005\135\005\139\005\146\005\141\005\131\005\136\005\138\005\
    \134\005\147\005\151\005\138\005\153\005\140\005\144\005\149\005\
    \161\005\157\005\152\005\155\005\150\005\149\005\151\005\168\005\
    \152\005\162\005\170\005\175\005\160\005\157\005\174\005\171\005\
    \161\005\162\005\182\005\171\005\180\005\166\005\182\005\170\005\
    \170\005\170\005\167\005\172\005\024\006\025\006\193\005\184\005\
    \188\005\191\005\198\005\180\005\192\005\187\005\189\005\203\005\
    \185\005\197\005\192\005\194\005\200\005\206\005\191\005\203\005\
    \199\005\211\005\209\005\199\005\200\005\216\005\218\005\205\005\
    \205\005\207\005\213\005\216\005\211\005\207\005\227\005\229\005\
    \216\005\226\005\235\005\235\005\226\005\219\005\235\005\222\005\
    \238\005\222\005\232\005\240\005\237\255\077\006\241\255\079\006\
    \002\000\236\005\251\005\238\005\252\005\252\005\001\006\250\005\
    \245\005\022\006\004\006\013\006\067\006\012\006\001\006\016\006\
    \255\005\068\006\031\006\024\006\065\006\251\255\047\000\159\006\
    \027\000\001\000\119\006\170\006\028\000\252\255\048\000\177\006\
    \254\255\021\006\005\006\095\006\003\006\010\006\080\006\051\006\
    \161\006\165\006\067\006\083\006\075\006\094\006\090\006\087\006\
    \095\006\081\006\102\006\090\006\094\006\097\006\095\006\100\006\
    \099\006\109\006\116\006\111\006\101\006\210\006\211\006\104\006\
    \118\006\105\006\114\006\123\006\115\006\110\006\112\006\250\255\
    \112\006\132\006\130\006\139\006\131\006\236\006\237\006\124\006\
    \140\006\144\006\141\006\138\006\151\006\143\006\157\006\147\006\
    \149\006\147\006\146\006\154\006\149\006\001\007\002\007\159\006\
    \161\006\160\006\159\006\165\006\174\006\157\006\162\006\178\006\
    \166\006\171\006\167\006\172\006\170\006\180\006\171\006\187\006\
    \183\006\174\006\192\006\185\006\194\006\196\006\192\006\201\006\
    \183\006\199\006\204\006\192\006\188\006\192\006\208\006\192\006\
    \206\006\198\006\193\006\213\006\198\006\195\006\213\006\213\006\
    \205\006\209\006\218\006\209\006\205\006\205\006\208\006\206\006\
    \225\006\210\006\217\006\215\006\215\006\235\006\221\006\225\006\
    \221\006\238\006\230\006\229\006\244\006\248\006\231\006\241\006\
    \245\006\242\006\247\006\237\006\001\007\241\006\000\007\093\007\
    \094\007\255\006\004\007\246\006\252\006\250\006\253\006\006\007\
    \006\007\253\006\004\007\013\007\003\007\014\007\001\007\109\007\
    \110\007\004\007\020\007\010\007\020\007\020\007\011\007\025\007\
    \120\007\121\007\029\007\030\007\036\007\024\007\020\007\024\007\
    \040\007\024\007\039\007\033\007\028\007\029\007\137\007\140\007\
    \038\007\034\007\045\007\041\007\038\007\050\007\046\007\054\007\
    \149\007\150\007\052\007\051\007\051\007\062\007\050\007\066\007\
    \061\007\056\007\058\007\058\007\069\007\057\007\077\007\065\007\
    \065\007\082\007\069\007\077\007\074\007\079\007\083\007\090\007\
    \085\007\075\007\186\007\187\007\078\007\091\007\080\007\089\007\
    \098\007\091\007\086\007\088\007\089\007\104\007\090\007\247\255\
    \108\007\099\007\112\007\113\007\100\007\114\007\117\007\108\007\
    \212\007\213\007\114\007\110\007\110\007\130\007\118\007\130\007\
    \129\007\123\007\131\007\134\007\125\007\130\007\134\007\141\007\
    \137\007\127\007\237\007\142\007\140\007\138\007\134\007\154\007\
    \139\007\151\007\139\007\157\007\158\007\163\007\151\007\163\007\
    \162\007\165\007\149\007\149\007\167\007\167\007\005\008\006\008\
    \151\007\161\007\173\007\155\007\246\255\166\007\161\007\173\007\
    \163\007\163\007\177\007\168\007\191\007\175\007\187\007\189\007\
    \196\007\192\007\187\007\205\007\189\007\201\007\196\007\198\007\
    \212\007\202\007\046\008\047\008\214\007\210\007\214\007\207\007\
    \216\007\223\007\217\007\208\007\226\007\202\007\207\007\217\007\
    \225\007\062\008\063\008\218\007\233\007\231\007\227\007\243\007\
    \225\007\238\007\229\007\230\007\245\007\232\007\243\007\231\007\
    \251\007\088\008\089\008\235\007\240\007\243\007\005\008\004\008\
    \244\007\000\008\251\007\253\007\250\007\012\008\253\007\252\007\
    \251\007\013\008\015\008\108\008\109\008\005\008\003\008\025\008\
    \007\008\019\008\014\008\016\008\030\008\028\008\012\008\034\008\
    \018\008\024\008\022\008\027\008\035\008\040\008\039\008\026\008\
    \038\008\033\008\035\008\050\008\032\008\044\008\040\008\042\008\
    \038\008\046\008\043\008\045\008\041\008\055\008\059\008\046\008\
    \061\008\048\008\049\008\057\008\070\008\055\008\052\008\069\008\
    \067\008\070\008\077\008\059\008\071\008\066\008\068\008\064\008\
    \245\255\083\008\071\008\112\008\086\008\071\008\073\008\075\008\
    \084\008\081\008\087\008\082\008\086\008\087\008\099\008\084\008\
    \078\008\081\008\090\008\092\008\099\008\095\008\090\008\107\008\
    \091\008\110\008\094\008\106\008\094\008\113\008\114\008\115\008\
    \118\008\102\008\102\008\106\008\120\008\103\008\108\008\126\008\
    \122\008\117\008\119\008\119\008\121\008\127\008\134\008\137\008\
    \127\008\124\008\138\008\131\008\123\008\139\008\127\008\131\008\
    \145\008\140\008\138\008\151\008\135\008\149\008\242\008\243\008\
    \144\008\144\008\156\008\140\008\150\008\158\008\145\008\145\008\
    \165\008\153\008\165\008\164\008\156\008\166\008\152\008\169\008\
    \248\255\167\008\162\008\166\008\165\008\166\008\180\008\179\008\
    \184\008\172\008\168\008\172\008\188\008\172\008\186\008\178\008\
    \173\008\179\008\177\008\191\008\184\008\178\008\190\008\197\008\
    \033\009\034\009\186\008\184\008\204\008\186\008\198\008\193\008\
    \195\008\205\008\199\008\204\008\195\008\212\008\193\008\215\008\
    \199\008\199\008\212\008\207\008\213\008\225\008\223\008\007\009\
    \228\008\232\008\221\008\216\008\218\008\216\008\222\008\232\008\
    \222\008\238\008\239\008\028\000\233\008\245\008\228\008\229\008\
    \229\008\248\008\232\008\232\008\236\008\250\008\251\008\244\008\
    \253\008\237\008\255\008\241\008\238\008\243\008\095\009\096\009\
    \008\009\255\008\003\009\006\009\013\009\251\008\008\009\003\009\
    \005\009\249\255\019\009\001\009\013\009\008\009\010\009\021\009\
    \008\009\018\009\014\009\022\009\012\009\029\009\030\009\030\009\
    \126\009\127\009\022\009\034\009\024\009\038\009\038\009\034\009\
    \042\009\032\009\032\009\043\009\033\009\045\009\040\009\042\009\
    \037\009\033\009\056\009\040\009\056\009\060\009\020\000\050\009\
    \051\009\051\009\067\009\056\009\157\009\182\009\050\009\056\009\
    \058\009\076\009\049\009\071\009\080\009\067\009\072\009\085\009\
    \086\009\086\009\091\009\107\009\078\009\111\009\082\009\109\009\
    \083\009\099\009\102\009\102\009\098\009\098\009\116\009\110\009\
    \110\009\106\009\110\009\127\009\111\009\125\009\117\009\113\009\
    \115\009\127\009\134\009\225\009\228\009\124\009\120\009\141\009\
    \123\009\135\009\130\009\132\009\155\009\170\009\161\009\162\009\
    \175\009\176\009\174\009\159\009\175\009\168\009\159\009\181\009\
    \165\009\170\009\172\009\179\009\175\009\170\009\188\009\172\009\
    \184\009\179\009\181\009\195\009\185\009\029\010\030\010\197\009\
    \193\009\197\009\190\009\199\009\204\009\200\009\191\009\210\009\
    \186\009\191\009\202\009\209\009\046\010\047\010\203\009\218\009\
    \220\009\210\009\226\009\208\009\221\009\212\009\213\009\228\009\
    \214\009\225\009\213\009\231\009\068\010\071\010\217\009\222\009\
    \225\009\243\009\242\009\226\009\238\009\233\009\235\009\230\009\
    \250\009\234\009\233\009\253\009\241\009\255\009\029\010\239\009\
    \251\009\250\009\000\010\251\009\255\009\000\010\012\010\252\009\
    \247\009\250\009\003\010\005\010\012\010\008\010\003\010\020\010\
    \004\010\023\010\008\010\020\010\010\010\028\010\030\010\031\010\
    \034\010\018\010\018\010\022\010\036\010\018\010\025\010\043\010\
    \039\010\034\010\036\010\032\010\033\010\038\010\040\010\046\010\
    \053\010\056\010\046\010\043\010\057\010\050\010\042\010\058\010\
    \046\010\061\010\048\010\058\010\054\010\054\010\064\010\066\010\
    \054\010\056\010\067\010\057\010\069\010\064\010\066\010\062\010\
    \058\010\081\010\065\010\081\010\083\010\108\010\074\010\075\010\
    \075\010\072\010\092\010\081\010\090\010\076\010\092\010\080\010\
    \080\010\085\010\100\010\090\010\101\010\103\010\093\010\193\010\
    \194\010\094\010\090\010\090\010\110\010\098\010\110\010\109\010\
    \100\010\113\010\105\010\109\010\106\010\112\010\102\010\114\010\
    \109\010\111\010\125\010\109\010\123\010\218\010\219\010\118\010\
    \120\010\116\010\116\010\136\010\124\010\136\010\135\010\137\010\
    \121\010\131\010\139\010\129\010\145\010\139\010\131\010\148\010\
    \180\010\140\010\152\010\137\010\138\010\137\010\156\010\140\010\
    \140\010\144\010\158\010\159\010\152\010\161\010\149\010\148\010\
    \160\010\155\010\158\010\170\010\166\010\206\010\171\010\175\010\
    \163\010\158\010\160\010\158\010\164\010\174\010\164\010\180\010\
    \185\010\183\010\167\010\188\010\172\010\178\010\177\010\182\010\
    \190\010\196\010\195\010\179\010\192\010\187\010\189\010\203\010\
    \185\010\197\010\192\010\195\010\191\010\199\010\188\010\208\010\
    \210\010\197\010\207\010\216\010\216\010\207\010\200\010\216\010\
    \203\010\207\010\219\010\222\010\215\010\212\010\208\010\228\010\
    \213\010\225\010\213\010\231\010\232\010\237\010\225\010\237\010\
    \237\010\240\010\224\010\224\010\242\010\242\010\079\011\080\011\
    \225\010\235\010\247\010\239\010\234\010\246\010\236\010\232\010\
    \246\010\250\010\001\011\252\010\242\010\247\010\249\010\245\010\
    \002\011\006\011\249\010\008\011\251\010\255\010\004\011\016\011\
    \012\011\007\011\010\011\005\011\004\011\006\011\023\011\007\011\
    \017\011\025\011\030\011\015\011\012\011\029\011\026\011\016\011\
    \017\011\037\011\026\011\035\011\021\011\037\011\025\011\025\011\
    \036\011\042\011\027\011\039\011\034\011\046\011\042\011\032\011\
    \033\011\049\011\051\011\039\011\036\011\038\011\046\011\049\011\
    \044\011\040\011\060\011\062\011\049\011\060\011\069\011\069\011\
    \060\011\053\011\069\011\056\011\072\011\056\011\066\011\074\011\
    \244\255\167\011\249\255\250\255\251\255\118\011\136\011\173\011\
    \175\011\181\011\183\011\185\011\182\011\222\011\194\011\197\011\
    \201\011\202\011\206\011\177\011\255\255\228\011\212\011\236\011\
    \249\255\231\011\179\011\106\011\140\011\255\255\242\011\233\011\
    \240\011\246\011\253\255\191\011\252\255\192\011\248\011\250\255\
    \251\255\252\011\013\012\233\255\234\255\206\012\113\011\127\011\
    \206\011\235\011\221\011\233\011\096\012\012\012\176\011\184\011\
    \177\011\046\012\034\013\255\255\239\255\252\255\129\012\017\012\
    \254\255\012\012\253\255\019\012\251\255\249\255\134\012\247\011\
    \248\011\136\012\141\012\143\012\148\012\249\011\248\255\244\255\
    \242\255\240\255\027\012\237\255\128\012\252\255\253\011\254\255\
    \255\255\253\255\009\012\252\255\254\011\254\255\255\255\253\255\
    \157\012\250\255\001\012\252\255\009\012\022\012\253\255\254\255\
    \251\255\172\012\249\255\250\255\017\012\034\012\043\012\047\012\
    \130\012\086\012\093\012\177\012\255\255\149\012\187\012\253\255\
    \252\255\251\255\044\012\253\255\035\012\255\255\046\012\254\255\
    \255\255\149\013\248\255\192\012\078\012\146\013\213\013\022\014\
    \233\011\175\014\251\255\122\012\123\012\254\255\255\255\253\255\
    \252\255\233\014\093\012\059\015\106\015\170\015\243\011\144\013\
    \234\015\016\013\217\013\100\012\025\016\072\016\136\016\249\011\
    \033\017\108\012\011\017\097\017\161\017\250\011\255\014\225\017\
    \180\014\243\014\018\015\033\018\009\015\045\015\075\017\097\018\
    \113\018\101\017\036\019\251\255\207\012\169\012\164\012\041\019\
    \173\013\254\255\170\012\252\255\204\012\251\255\252\255\143\012\
    \147\012\255\255\065\013\254\255\183\012\158\013\048\019\241\255\
    \242\255\203\012\218\012\244\255\183\012\184\012\192\012\144\012\
    \202\012\205\012\255\255\206\014\254\255\239\012\064\017\254\012\
    \251\255\241\012\249\255\247\255\245\255\007\019\252\255\211\012\
    \254\255\046\019\255\255\000\013\054\019\253\255\057\019\249\255\
    \152\013\082\019\226\012\254\255\089\019\255\255\032\013\091\019\
    \253\255\252\255\066\013\096\019\039\019\254\255\154\013\120\019\
    \241\255\242\255\243\255\219\012\232\019\184\019\040\020\148\012\
    \193\020\100\021\037\013\234\019\252\255\243\012\152\019\071\013\
    \001\013\084\019\052\021\116\013\102\019\254\255\120\013\109\019\
    \253\255\150\013\244\255\251\255\072\013\156\013\240\019\080\013\
    \127\019\250\255\164\013\133\013\135\019\248\255\158\021\047\013\
    \240\021\178\020\048\022\249\012\201\022\108\023\015\024\178\024\
    \242\020\196\022\105\021\153\019\079\013\086\023\249\023\173\024\
    \000\013\085\025\209\013\063\025\149\025\213\025\002\013\180\021\
    \021\026\151\021\188\019\238\021\085\026\163\021\168\021\150\023\
    \149\026\009\024\226\021\237\019\253\255\254\255\255\255\199\014\
    \253\255\154\013\255\255\200\014\253\255\176\013\255\255\205\013\
    \254\255\171\013\255\255\111\019\131\013\143\013\134\015\144\019\
    \183\013\189\013\239\013\085\019\060\014\137\019\207\014\018\019\
    \254\255\255\255\177\014\000\015\210\014\076\027\253\255\129\014\
    \255\255\225\014\245\019\226\014\247\019\144\014\002\015\254\255\
    \001\015\078\027\254\255\052\019\251\019\255\019\080\027\082\027\
    \255\255\030\020\127\021\087\027\241\021\088\027\255\255\003\025\
    \094\027\005\025\108\025\254\255\089\027\096\027\101\027\037\019\
    \254\255\026\015\027\015\126\019\013\020\135\015\099\019\123\027\
    \254\255\218\016\101\027\217\027\025\028\141\014\199\028\117\019\
    \123\019\001\029\165\029\090\017\143\029\229\029\037\030\142\014\
    \134\025\101\030\021\024\153\025\041\019\148\030\195\030\003\031\
    \144\014\177\031\173\019\155\031\241\031\049\032\183\014\000\027\
    \113\032\224\026\028\027\165\027\177\032\041\028\042\027\023\029\
    \241\032\204\028\191\027\224\016\255\255\042\019\252\019";
  Lexing.lex_backtrk =
   "\255\255\255\255\255\255\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\000\000\020\000\020\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\017\000\255\255\255\255\255\255\
    \255\255\255\255\001\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\014\000\255\255\013\000\
    \013\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\016\000\
    \255\255\255\255\255\255\015\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\010\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\010\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\010\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\010\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\011\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \010\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\009\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\011\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\011\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\011\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\011\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\011\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\011\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\011\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \010\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\010\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\010\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\012\000\
    \014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\
    \014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\
    \014\000\014\000\014\000\014\000\014\000\255\255\014\000\000\000\
    \014\000\255\255\255\255\255\255\255\255\255\255\002\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\008\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\008\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \008\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \008\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\010\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\008\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\007\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\010\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \010\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\010\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\010\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \010\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\010\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\010\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\008\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \008\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\008\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\006\000\001\000\002\000\
    \006\000\006\000\006\000\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\000\000\255\255\255\255\003\000\255\255\
    \255\255\006\000\006\000\006\000\006\000\255\255\006\000\001\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\005\000\255\255\255\255\255\255\022\000\022\000\020\000\
    \022\000\014\000\012\000\010\000\022\000\022\000\022\000\005\000\
    \000\000\022\000\017\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\009\000\255\255\
    \255\255\008\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\019\000\255\255\255\255\255\255\003\000\255\255\
    \255\255\255\255\255\255\255\255\003\000\255\255\255\255\255\255\
    \255\255\255\255\005\000\255\255\005\000\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\006\000\006\000\006\000\006\000\
    \006\000\255\255\255\255\001\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\001\000\255\255\255\255\255\255\
    \255\255\006\000\255\255\006\000\007\000\007\000\007\000\007\000\
    \007\000\005\000\255\255\007\000\007\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \005\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\005\000\002\000\005\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\004\000\
    \004\000\255\255\002\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\014\000\012\000\255\255\009\000\007\000\005\000\003\000\
    \014\000\014\000\255\255\002\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\003\000\
    \255\255\003\000\255\255\255\255\001\000\255\255\005\000\255\255\
    \005\000\004\000\006\000\255\255\006\000\255\255\255\255\001\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\014\000\014\000\014\000\014\000\014\000\
    \010\000\010\000\014\000\014\000\255\255\014\000\000\000\014\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\004\000\255\255\006\000\255\255\
    \255\255\255\255\255\255\008\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\010\000\010\000\010\000\009\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\010\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\001\000\255\255\255\255\255\255\001\000\255\255\255\255\
    \255\255\001\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\001\000\000\000\255\255\255\255\255\255\002\000\
    \255\255\002\000\002\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\001\000\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\001\000\001\000\001\000\001\000\001\000\001\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\001\000\255\255";
  Lexing.lex_default =
   "\001\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\070\000\070\000\070\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\088\000\
    \088\000\088\000\255\255\088\000\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\097\002\000\000\097\002\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\142\004\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \144\006\144\006\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\186\008\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\203\008\255\255\
    \255\255\255\255\203\008\203\008\000\000\255\255\255\255\208\008\
    \000\000\223\008\255\255\255\255\255\255\000\000\208\008\255\255\
    \255\255\255\255\000\000\255\255\000\000\255\255\223\008\000\000\
    \000\000\223\008\230\008\000\000\000\000\011\009\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\244\008\000\000\000\000\000\000\255\255\255\255\
    \000\000\255\255\000\000\255\255\000\000\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\000\000\255\255\000\000\013\009\000\000\255\255\000\000\
    \000\000\000\000\019\009\000\000\255\255\000\000\000\000\000\000\
    \025\009\000\000\255\255\000\000\255\255\255\255\000\000\000\000\
    \000\000\034\009\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\000\000\
    \000\000\000\000\051\009\000\000\255\255\000\000\055\009\000\000\
    \000\000\058\009\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\000\000\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\109\009\000\000\109\009\109\009\114\009\109\009\
    \109\009\000\000\114\009\000\000\117\009\000\000\000\000\255\255\
    \255\255\000\000\255\255\000\000\255\255\255\255\127\009\000\000\
    \000\000\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\000\000\255\255\255\255\255\255\
    \000\000\255\255\000\000\000\000\000\000\150\009\000\000\255\255\
    \000\000\255\255\000\000\255\255\255\255\000\000\159\009\000\000\
    \255\255\255\255\255\255\000\000\255\255\000\000\255\255\255\255\
    \000\000\000\000\255\255\255\255\173\009\000\000\255\255\176\009\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \000\000\255\255\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\252\009\000\000\000\000\000\000\000\010\
    \000\000\255\255\000\000\004\010\000\000\255\255\000\000\008\010\
    \000\000\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\024\010\
    \000\000\000\000\255\255\255\255\255\255\030\010\000\000\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\043\010\000\000\043\010\043\010\043\010\043\010\049\010\
    \000\000\049\010\049\010\052\010\052\010\052\010\000\000\052\010\
    \052\010\052\010\052\010\000\000\052\010\052\010\052\010\064\010\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\072\010\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\117\010\000\000\118\010\118\010";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\027\000\025\000\025\000\025\000\026\000\050\000\025\000\
    \058\000\057\000\069\000\106\000\057\000\071\000\072\000\071\000\
    \072\000\078\000\087\000\087\000\090\000\113\000\090\000\090\000\
    \027\000\098\000\124\000\003\000\117\000\167\004\171\004\058\000\
    \024\000\069\000\052\000\052\000\071\000\072\000\071\000\072\000\
    \078\000\087\000\087\000\090\000\093\000\090\000\090\000\105\000\
    \098\000\112\000\123\000\117\000\089\000\089\000\091\000\089\000\
    \092\000\007\000\006\000\023\000\014\000\017\000\019\000\021\000\
    \016\000\020\000\157\002\210\003\022\000\004\000\009\000\015\000\
    \010\000\008\000\018\000\012\000\013\000\011\000\005\000\208\002\
    \151\003\174\004\175\004\140\004\140\004\184\008\173\004\204\006\
    \015\007\128\004\112\004\176\002\113\004\109\004\103\002\090\004\
    \135\002\232\001\132\002\128\004\136\002\104\002\172\002\136\004\
    \125\004\089\004\073\002\058\002\177\002\108\004\131\002\105\002\
    \111\004\175\002\248\000\133\002\134\002\028\002\102\002\072\002\
    \029\002\059\002\192\000\211\001\169\001\027\002\155\001\030\002\
    \136\004\154\001\049\002\210\001\194\000\171\001\170\001\128\000\
    \125\004\193\000\127\000\112\004\237\000\113\004\126\000\209\001\
    \051\000\025\000\129\000\109\004\050\000\130\000\090\004\176\002\
    \172\002\232\001\155\001\115\000\073\002\154\001\058\002\094\000\
    \089\004\111\004\135\002\108\004\132\002\073\000\136\002\051\000\
    \177\002\072\002\028\000\103\002\059\002\175\002\248\000\049\000\
    \131\002\028\002\104\002\192\000\029\002\133\002\134\002\169\001\
    \062\000\027\002\055\000\030\002\105\002\194\000\049\002\063\000\
    \171\001\170\001\193\000\102\002\064\000\065\000\066\000\067\000\
    \032\000\031\000\048\000\039\000\042\000\044\000\046\000\041\000\
    \045\000\055\000\068\000\047\000\029\000\034\000\040\000\035\000\
    \033\000\043\000\037\000\038\000\036\000\030\000\211\001\055\000\
    \128\000\074\000\061\000\127\000\051\000\025\000\210\001\126\000\
    \050\000\075\000\061\000\129\000\076\000\077\000\130\000\237\000\
    \078\000\080\000\209\001\081\000\082\000\083\000\055\000\084\000\
    \002\000\061\000\085\000\051\000\086\000\095\000\028\000\096\000\
    \056\000\061\000\097\000\049\000\255\255\255\255\255\255\078\000\
    \108\000\101\000\255\255\255\255\102\000\255\255\098\000\103\000\
    \109\000\110\000\116\000\054\000\119\000\120\000\053\000\121\000\
    \180\000\171\000\155\000\090\000\032\000\031\000\048\000\039\000\
    \042\000\044\000\046\000\041\000\045\000\098\000\103\000\047\000\
    \029\000\034\000\040\000\035\000\033\000\043\000\037\000\038\000\
    \036\000\030\000\090\000\105\000\106\000\110\000\117\000\107\000\
    \054\000\121\000\149\000\053\000\089\000\103\000\112\000\113\000\
    \134\000\175\000\114\000\133\000\123\000\124\000\143\000\104\000\
    \125\000\060\000\105\000\059\000\110\000\117\000\186\000\176\000\
    \121\000\060\000\187\000\059\000\163\000\112\000\111\000\079\000\
    \137\000\138\000\122\000\123\000\139\000\140\000\162\000\131\000\
    \141\000\144\000\142\000\142\000\145\000\146\000\142\000\136\000\
    \132\000\147\000\148\000\150\000\135\000\100\000\151\000\152\000\
    \153\000\154\000\127\000\156\000\157\000\158\000\159\000\160\000\
    \099\000\142\000\161\000\127\000\169\000\164\000\165\000\166\000\
    \167\000\168\000\127\000\170\000\127\000\172\000\173\000\174\000\
    \127\000\179\000\177\000\178\000\127\000\181\000\182\000\183\000\
    \184\000\185\000\127\000\189\000\188\000\127\000\190\000\191\000\
    \127\000\151\001\152\001\195\000\178\000\196\000\197\000\199\000\
    \118\000\127\000\127\000\200\000\200\000\104\001\095\001\198\000\
    \056\001\055\001\050\001\248\000\244\000\239\000\237\000\230\000\
    \212\000\213\000\215\000\229\000\193\000\214\000\216\000\217\000\
    \218\000\219\000\200\000\200\000\220\000\220\000\222\000\223\000\
    \224\000\225\000\226\000\227\000\228\000\178\000\178\000\231\000\
    \178\000\232\000\233\000\234\000\193\000\238\000\235\000\178\000\
    \240\000\235\000\241\000\220\000\220\000\242\000\243\000\178\000\
    \245\000\246\000\247\000\178\000\249\000\250\000\204\000\203\000\
    \207\000\209\000\251\000\206\000\211\000\235\000\251\000\210\000\
    \235\000\040\001\205\000\036\001\032\001\208\000\201\000\202\000\
    \036\000\023\001\016\001\007\001\003\001\221\000\004\001\005\001\
    \006\001\251\000\178\000\008\001\013\001\251\000\010\001\009\001\
    \011\001\012\001\178\000\014\001\015\001\178\000\017\001\018\001\
    \019\001\020\001\021\001\022\001\178\000\024\001\025\001\026\001\
    \027\001\028\001\255\255\029\001\236\000\030\001\031\001\178\000\
    \033\001\034\001\035\001\178\000\037\001\038\001\039\001\178\000\
    \252\000\041\001\002\001\042\001\043\001\044\001\000\001\045\001\
    \046\001\047\001\048\001\049\001\254\000\178\000\053\001\001\001\
    \051\001\255\000\253\000\178\000\052\001\054\001\178\000\176\000\
    \057\001\058\001\090\001\063\001\060\001\061\001\062\001\178\000\
    \064\001\065\001\059\001\066\001\067\001\067\001\087\001\080\001\
    \078\001\072\001\073\001\074\001\075\001\076\001\077\001\178\000\
    \079\001\178\000\081\001\082\001\083\001\083\001\085\001\086\001\
    \178\000\088\001\089\001\067\001\067\001\082\001\091\001\092\001\
    \092\001\094\001\178\000\096\001\097\001\098\001\099\001\100\001\
    \101\001\178\000\102\001\083\001\083\001\103\001\178\000\105\001\
    \106\001\135\001\109\001\106\001\110\001\111\001\092\001\092\001\
    \112\001\113\001\114\001\115\001\116\001\116\001\134\001\127\001\
    \120\001\121\001\122\001\123\001\124\001\125\001\126\001\106\001\
    \069\001\071\001\106\001\178\000\128\001\129\001\068\001\070\001\
    \130\001\131\001\132\001\116\001\116\001\133\001\178\000\178\000\
    \136\001\137\001\138\001\139\001\140\001\141\001\144\001\084\001\
    \145\001\142\001\142\001\146\001\147\001\148\001\093\001\149\001\
    \150\001\178\000\178\000\153\001\148\000\168\001\156\001\157\001\
    \158\001\159\001\163\001\107\001\160\001\117\001\160\001\119\001\
    \142\001\142\001\108\001\212\000\162\001\164\001\165\001\166\001\
    \167\001\127\000\148\000\204\001\179\001\181\001\172\001\178\001\
    \174\001\118\001\175\001\160\001\173\001\160\001\176\001\214\000\
    \177\001\220\000\202\001\194\001\180\001\187\001\182\001\183\001\
    \184\001\188\001\143\001\185\001\186\001\127\000\192\001\189\001\
    \190\001\191\001\127\000\193\001\127\000\195\001\196\001\197\001\
    \220\000\200\001\201\001\127\000\198\001\198\001\203\001\205\001\
    \204\000\203\000\207\000\209\000\206\001\206\000\161\001\207\001\
    \208\001\210\000\127\000\218\001\205\000\213\001\212\001\208\000\
    \201\000\202\000\036\000\198\001\198\001\148\000\214\001\215\001\
    \216\001\217\001\127\000\219\001\220\001\221\001\222\001\223\001\
    \224\001\225\001\226\001\226\001\228\001\229\001\230\001\231\001\
    \127\000\011\002\250\001\239\001\238\001\234\001\237\001\148\000\
    \249\001\240\001\241\001\241\001\243\001\244\001\245\001\235\001\
    \012\002\226\001\226\001\240\000\236\001\233\001\246\001\247\001\
    \248\001\127\000\252\001\251\001\127\000\253\001\254\001\255\001\
    \199\001\241\001\241\001\000\002\001\002\001\002\003\002\004\002\
    \005\002\006\002\007\002\008\002\009\002\010\002\127\000\015\002\
    \013\002\127\000\014\002\127\000\016\002\017\002\018\002\021\002\
    \019\002\019\002\022\002\001\002\001\002\023\002\024\002\025\002\
    \026\002\203\001\048\002\127\000\042\002\227\001\032\002\036\002\
    \033\002\034\002\035\002\127\000\037\002\038\002\242\001\019\002\
    \019\002\039\002\040\002\041\002\127\000\245\000\043\002\044\002\
    \045\002\031\002\046\002\047\002\127\000\127\000\050\002\051\002\
    \052\002\002\002\053\002\054\002\055\002\056\002\127\000\127\000\
    \063\002\057\002\060\002\061\002\062\002\178\000\052\001\064\002\
    \065\002\066\002\067\002\068\002\069\002\070\002\071\002\101\002\
    \057\001\075\002\093\002\076\002\074\002\077\002\020\002\078\002\
    \079\002\080\002\059\001\080\002\087\002\083\002\084\002\085\002\
    \086\002\127\000\088\002\089\002\090\002\091\002\092\002\127\000\
    \094\002\096\002\098\002\097\002\073\001\095\002\099\002\100\002\
    \080\002\127\000\080\002\127\000\130\002\118\002\114\002\106\002\
    \107\002\108\002\109\002\110\002\111\002\112\002\113\002\178\000\
    \115\002\116\002\117\002\148\000\119\002\120\002\121\002\124\002\
    \125\002\122\002\122\002\126\002\127\002\128\002\129\002\127\000\
    \127\000\170\002\167\002\161\002\156\002\155\002\137\002\142\002\
    \138\002\139\002\140\002\081\002\141\002\127\000\069\001\082\002\
    \122\002\122\002\143\002\144\002\068\001\070\001\146\002\150\002\
    \147\002\105\001\148\002\149\002\127\000\151\002\152\002\153\002\
    \154\002\127\000\097\002\158\002\159\002\160\002\127\000\145\002\
    \162\002\163\002\164\002\165\002\166\002\127\000\168\002\169\002\
    \127\000\171\002\071\002\173\002\174\002\105\001\127\000\082\004\
    \178\002\081\004\180\002\211\002\203\002\193\002\179\002\123\002\
    \184\002\191\002\182\002\183\002\186\002\187\002\188\002\181\002\
    \189\002\190\002\185\002\148\000\192\002\148\000\195\002\195\002\
    \200\002\198\002\199\002\194\002\201\002\202\002\194\002\204\002\
    \205\002\206\002\207\002\209\002\210\002\127\000\212\002\213\002\
    \214\002\214\002\079\004\075\004\071\004\195\002\195\002\063\004\
    \054\004\064\004\010\004\007\004\019\004\206\003\178\003\205\003\
    \037\003\005\003\249\002\194\002\194\002\018\004\020\004\214\002\
    \214\002\242\003\131\003\177\003\243\003\065\004\207\003\208\003\
    \007\003\241\003\162\003\244\003\016\003\163\003\077\003\241\002\
    \237\002\071\002\006\003\132\003\015\003\076\003\238\002\239\002\
    \130\003\078\003\240\002\071\002\079\003\242\002\243\002\244\002\
    \014\003\218\002\222\002\228\002\225\002\223\002\230\002\226\002\
    \219\002\220\002\245\002\246\002\231\002\217\002\234\002\216\002\
    \227\002\215\002\229\002\224\002\232\002\236\002\221\002\247\002\
    \248\002\071\002\250\002\251\002\252\002\253\002\253\002\235\002\
    \255\002\000\003\255\255\196\002\255\255\197\002\233\002\001\003\
    \002\003\003\003\004\003\071\002\071\002\011\003\009\003\010\003\
    \008\003\071\002\012\003\071\002\253\002\253\002\013\003\071\002\
    \071\002\023\003\018\003\017\003\071\002\019\003\020\003\021\003\
    \022\003\071\002\024\003\025\003\026\003\027\003\028\003\029\003\
    \030\003\031\003\031\003\033\003\034\003\035\003\036\003\071\002\
    \072\003\055\003\039\003\044\003\043\003\071\002\042\003\054\003\
    \045\003\046\003\046\003\048\003\040\003\049\003\050\003\073\003\
    \031\003\031\003\038\003\254\002\041\003\051\003\052\003\053\003\
    \071\002\057\003\056\003\071\002\058\003\059\003\060\003\061\003\
    \046\003\046\003\062\003\062\003\064\003\065\003\066\003\067\003\
    \068\003\069\003\070\003\071\003\071\002\071\002\074\003\075\003\
    \071\002\124\003\118\003\116\003\083\003\125\003\112\003\082\003\
    \104\003\062\003\062\003\096\003\090\003\085\003\086\003\087\003\
    \088\003\089\003\103\003\071\002\032\003\091\003\092\003\093\003\
    \094\003\095\003\071\002\097\003\098\003\047\003\099\003\100\003\
    \101\003\102\003\071\002\080\003\110\003\105\003\106\003\107\003\
    \108\003\109\003\071\002\111\003\081\003\255\255\255\255\063\003\
    \084\003\071\002\113\003\114\003\115\003\071\002\117\003\071\002\
    \119\003\120\003\121\003\122\003\123\003\071\002\127\003\126\003\
    \071\002\128\003\129\003\071\002\071\002\155\003\133\003\154\003\
    \146\003\136\003\137\003\144\003\134\003\135\003\139\003\140\003\
    \141\003\142\003\143\003\071\002\138\003\145\003\071\002\147\003\
    \148\003\149\003\150\003\152\003\153\003\071\002\071\002\156\003\
    \157\003\158\003\159\003\160\003\161\003\071\002\165\003\164\003\
    \071\002\166\003\167\003\168\003\169\003\169\003\171\003\172\003\
    \173\003\174\003\175\003\176\003\071\002\204\003\179\003\187\003\
    \181\003\180\003\182\003\183\003\184\003\185\003\186\003\071\002\
    \188\003\189\003\190\003\169\003\169\003\191\003\191\003\200\003\
    \194\003\195\003\196\003\197\003\198\003\199\003\071\002\201\003\
    \202\003\203\003\071\002\071\002\223\003\220\003\224\003\214\003\
    \209\003\211\003\212\003\213\003\191\003\191\003\071\002\215\003\
    \216\003\217\003\218\003\219\003\071\002\221\003\222\003\071\002\
    \228\003\225\003\226\003\227\003\071\002\170\003\229\003\230\003\
    \232\003\236\003\233\003\234\003\235\003\071\002\237\003\238\003\
    \239\003\240\003\071\002\006\004\071\002\000\004\246\003\250\003\
    \247\003\231\003\248\003\249\003\071\002\251\003\193\003\252\003\
    \253\003\254\003\192\003\255\003\071\002\001\004\002\004\003\004\
    \004\004\245\003\005\004\071\002\071\002\008\004\009\004\071\002\
    \011\004\012\004\013\004\014\004\016\004\071\002\017\004\071\002\
    \049\004\044\004\021\004\023\004\036\004\029\004\024\004\015\004\
    \025\004\030\004\026\004\027\004\028\004\071\002\034\004\031\004\
    \032\004\033\004\022\004\071\002\035\004\071\002\037\004\038\004\
    \039\004\040\004\040\004\042\004\043\004\071\002\045\004\046\004\
    \047\004\048\004\071\002\050\004\051\004\052\004\053\004\071\002\
    \055\004\056\004\057\004\058\004\059\004\060\004\061\004\071\002\
    \040\004\040\004\062\004\071\002\069\004\071\002\066\004\067\004\
    \068\004\071\002\070\004\071\002\072\004\073\004\074\004\071\002\
    \076\004\077\004\078\004\071\002\080\004\071\002\127\000\083\004\
    \084\004\085\004\086\004\087\004\088\004\127\000\103\004\091\004\
    \092\004\093\004\093\004\095\004\096\004\097\004\098\004\099\004\
    \100\004\101\004\102\004\194\002\104\004\105\004\106\004\107\004\
    \203\001\110\004\097\002\127\000\122\004\041\004\119\004\127\000\
    \093\004\093\004\115\004\116\004\117\004\118\004\097\002\121\004\
    \114\004\127\000\120\004\097\002\123\004\124\004\127\000\126\004\
    \127\004\127\000\129\004\130\004\131\004\132\004\134\004\127\000\
    \135\004\127\000\137\004\138\004\139\004\127\000\143\004\167\004\
    \143\004\133\004\168\004\180\008\172\008\169\008\156\008\220\006\
    \157\008\153\008\240\006\140\005\225\005\150\006\144\005\215\004\
    \094\004\219\006\235\004\182\004\151\006\143\004\051\006\143\004\
    \144\004\152\008\120\006\241\006\155\008\166\004\152\006\182\006\
    \239\006\179\006\042\006\183\006\052\006\149\006\021\006\119\006\
    \170\004\022\006\145\005\162\005\148\005\178\006\020\006\147\005\
    \023\006\143\005\180\006\181\006\164\005\163\005\148\004\147\004\
    \156\004\155\004\159\004\161\004\163\004\158\004\162\004\170\004\
    \143\005\164\004\145\004\151\004\157\004\150\004\149\004\160\004\
    \153\004\154\004\152\004\146\004\066\006\204\005\177\004\143\005\
    \170\004\171\004\185\004\065\006\172\004\203\005\185\004\068\006\
    \179\004\140\005\067\006\170\004\171\004\178\004\224\004\172\004\
    \093\005\202\005\176\004\176\004\184\004\084\005\176\004\170\004\
    \180\004\185\004\181\004\045\005\183\004\185\004\042\005\037\005\
    \235\004\231\004\170\004\226\004\224\004\217\004\198\004\199\004\
    \201\004\176\004\178\004\200\004\169\004\216\004\202\004\203\004\
    \204\004\205\004\208\004\206\004\206\004\209\004\210\004\211\004\
    \212\004\213\004\214\004\215\004\215\004\218\004\215\004\219\004\
    \190\004\189\004\193\004\195\004\220\004\192\004\197\004\221\004\
    \178\004\196\004\206\004\206\004\191\004\222\004\222\004\194\004\
    \187\004\188\004\186\004\225\004\169\004\215\004\227\004\228\004\
    \229\004\230\004\215\004\232\004\233\004\234\004\215\004\169\004\
    \236\004\237\004\238\004\238\004\222\004\222\004\027\005\023\005\
    \019\005\010\005\003\005\250\004\207\004\246\004\247\004\248\004\
    \249\004\215\004\251\004\000\005\253\004\254\004\252\004\255\004\
    \215\004\238\004\238\004\001\005\002\005\215\004\004\005\005\005\
    \006\005\007\005\008\005\009\005\215\004\011\005\012\005\013\005\
    \014\005\015\005\016\005\017\005\018\005\215\004\020\005\021\005\
    \022\005\223\004\215\004\024\005\025\005\026\005\215\004\028\005\
    \029\005\030\005\031\005\032\005\033\005\239\004\034\005\245\004\
    \035\005\036\005\215\004\243\004\038\005\165\004\040\005\215\004\
    \039\005\241\004\041\005\215\004\244\004\043\005\242\004\240\004\
    \044\005\215\004\046\005\047\005\079\005\052\005\049\005\050\005\
    \051\005\215\004\053\005\054\005\048\005\055\005\056\005\056\005\
    \076\005\069\005\067\005\061\005\062\005\063\005\064\005\065\005\
    \066\005\215\004\068\005\215\004\070\005\071\005\072\005\072\005\
    \074\005\075\005\215\004\077\005\078\005\056\005\056\005\071\005\
    \080\005\081\005\081\005\083\005\215\004\085\005\086\005\087\005\
    \088\005\089\005\090\005\215\004\091\005\072\005\072\005\092\005\
    \215\004\094\005\095\005\124\005\098\005\095\005\099\005\100\005\
    \081\005\081\005\101\005\102\005\103\005\104\005\105\005\105\005\
    \123\005\116\005\109\005\110\005\111\005\112\005\113\005\114\005\
    \115\005\095\005\058\005\060\005\095\005\215\004\117\005\118\005\
    \057\005\059\005\119\005\120\005\121\005\105\005\105\005\122\005\
    \215\004\215\004\125\005\126\005\127\005\128\005\129\005\130\005\
    \133\005\073\005\134\005\131\005\131\005\135\005\136\005\137\005\
    \082\005\138\005\139\005\215\004\141\005\142\005\094\005\215\004\
    \146\005\143\005\161\005\149\005\150\005\096\005\151\005\106\005\
    \152\005\108\005\131\005\131\005\097\005\153\005\153\005\198\004\
    \155\005\156\005\157\005\158\005\159\005\160\005\143\005\143\005\
    \197\005\172\005\174\005\107\005\171\005\165\005\166\005\167\005\
    \168\005\200\004\169\005\170\005\153\005\153\005\206\004\195\005\
    \187\005\173\005\180\005\175\005\132\005\176\005\181\005\177\005\
    \178\005\179\005\143\005\185\005\182\005\183\005\184\005\143\005\
    \186\005\143\005\188\005\189\005\190\005\206\004\191\005\191\005\
    \193\005\194\005\143\005\196\005\198\005\199\005\200\005\201\005\
    \190\004\189\004\193\004\195\004\143\005\192\004\154\005\211\005\
    \206\005\196\004\205\005\143\005\191\004\191\005\191\005\194\004\
    \187\004\188\004\186\004\207\005\208\005\209\005\210\005\143\005\
    \212\005\213\005\214\005\215\005\216\005\217\005\218\005\219\005\
    \219\005\221\005\222\005\223\005\224\005\143\005\004\006\243\005\
    \232\005\231\005\227\005\230\005\143\005\242\005\233\005\234\005\
    \234\005\236\005\237\005\238\005\228\005\005\006\219\005\219\005\
    \227\004\229\005\226\005\239\005\240\005\241\005\143\005\245\005\
    \244\005\143\005\192\005\246\005\247\005\248\005\234\005\234\005\
    \249\005\250\005\250\005\252\005\253\005\254\005\255\005\000\006\
    \001\006\002\006\003\006\143\005\008\006\006\006\143\005\007\006\
    \143\005\009\006\010\006\011\006\014\006\012\006\012\006\015\006\
    \250\005\250\005\016\006\017\006\018\006\019\006\196\005\041\006\
    \143\005\035\006\220\005\025\006\029\006\026\006\027\006\028\006\
    \143\005\030\006\031\006\235\005\012\006\012\006\032\006\033\006\
    \034\006\143\005\232\004\036\006\037\006\038\006\024\006\039\006\
    \040\006\143\005\143\005\043\006\044\006\045\006\251\005\046\006\
    \047\006\048\006\049\006\143\005\143\005\056\006\050\006\053\006\
    \054\006\055\006\215\004\039\005\057\006\058\006\059\006\060\006\
    \061\006\062\006\063\006\064\006\113\006\107\006\074\006\069\006\
    \114\006\073\006\070\006\013\006\143\005\103\006\095\006\087\006\
    \081\006\076\006\077\006\078\006\079\006\043\005\080\006\143\005\
    \094\006\082\006\083\006\084\006\085\006\086\006\143\005\088\006\
    \089\006\090\006\091\006\092\006\093\006\071\006\143\005\101\006\
    \096\006\097\006\098\006\099\006\100\006\143\005\072\006\102\006\
    \143\005\104\006\075\006\105\006\106\006\143\005\108\006\109\006\
    \110\006\111\006\112\006\143\005\116\006\115\006\143\005\117\006\
    \118\006\143\005\148\006\046\005\122\006\140\006\123\006\121\006\
    \124\006\125\006\126\006\127\006\127\006\048\005\134\006\130\006\
    \131\006\132\006\133\006\143\005\135\006\136\006\137\006\138\006\
    \139\006\143\005\141\006\143\006\145\006\144\006\062\005\142\006\
    \146\006\147\006\127\006\127\006\143\005\143\005\177\006\165\006\
    \161\006\153\006\154\006\155\006\156\006\157\006\158\006\159\006\
    \160\006\215\004\162\006\163\006\164\006\143\005\166\006\167\006\
    \168\006\171\006\169\006\169\006\172\006\173\006\174\006\175\006\
    \176\006\143\005\143\005\217\006\214\006\208\006\203\006\202\006\
    \184\006\189\006\185\006\186\006\128\006\187\006\188\006\058\005\
    \129\006\169\006\169\006\143\005\190\006\057\005\059\005\191\006\
    \193\006\197\006\194\006\094\005\195\006\196\006\143\005\198\006\
    \199\006\200\006\201\006\143\005\144\006\205\006\206\006\207\006\
    \143\005\192\006\209\006\210\006\211\006\212\006\213\006\143\005\
    \215\006\216\006\143\005\218\006\064\006\234\006\221\006\222\006\
    \223\006\223\006\225\006\226\006\227\006\228\006\229\006\230\006\
    \170\006\231\006\232\006\233\006\235\006\236\006\237\006\238\006\
    \196\005\143\005\145\008\242\006\144\008\244\006\018\007\223\006\
    \223\006\243\006\010\007\001\007\248\006\247\006\246\006\249\006\
    \249\006\254\006\245\006\252\006\253\006\002\007\233\006\255\006\
    \000\007\233\006\008\007\003\007\004\007\005\007\006\007\007\007\
    \143\005\009\007\143\005\011\007\012\007\013\007\249\006\249\006\
    \014\007\016\007\017\007\143\005\019\007\020\007\021\007\142\008\
    \138\008\134\008\117\008\082\008\233\006\233\006\126\008\224\006\
    \127\008\073\008\070\008\100\007\081\008\083\008\049\008\070\007\
    \013\008\050\008\012\008\241\007\194\007\021\007\048\008\021\007\
    \051\008\069\007\225\007\068\007\128\008\226\007\056\007\048\007\
    \240\007\014\008\015\008\044\007\140\007\195\007\041\007\064\006\
    \079\007\045\007\193\007\139\007\046\007\043\007\021\007\141\007\
    \078\007\047\007\142\007\064\006\049\007\050\007\051\007\042\007\
    \052\007\053\007\054\007\055\007\077\007\064\006\040\007\057\007\
    \058\007\059\007\060\007\062\007\063\007\060\007\064\007\065\007\
    \066\007\067\007\064\006\255\255\250\006\255\255\251\006\025\007\
    \029\007\035\007\032\007\030\007\037\007\033\007\026\007\027\007\
    \064\006\060\007\038\007\024\007\060\007\023\007\034\007\022\007\
    \036\007\031\007\039\007\074\007\028\007\073\007\071\007\072\007\
    \064\006\075\007\064\006\076\007\064\006\064\006\086\007\081\007\
    \080\007\064\006\082\007\083\007\084\007\085\007\064\006\087\007\
    \088\007\089\007\090\007\091\007\092\007\093\007\094\007\094\007\
    \096\007\097\007\098\007\099\007\064\006\135\007\118\007\102\007\
    \107\007\106\007\061\007\105\007\117\007\064\006\108\007\109\007\
    \109\007\103\007\111\007\112\007\136\007\094\007\094\007\101\007\
    \113\007\104\007\114\007\115\007\116\007\064\006\120\007\119\007\
    \064\006\121\007\122\007\123\007\124\007\125\007\109\007\109\007\
    \125\007\127\007\128\007\129\007\130\007\131\007\132\007\133\007\
    \134\007\064\006\064\006\137\007\138\007\064\006\187\007\181\007\
    \179\007\175\007\188\007\146\007\125\007\167\007\145\007\125\007\
    \159\007\153\007\148\007\149\007\150\007\151\007\152\007\166\007\
    \064\006\095\007\154\007\155\007\156\007\157\007\158\007\064\006\
    \160\007\161\007\162\007\110\007\163\007\164\007\255\255\255\255\
    \165\007\064\006\143\007\173\007\168\007\169\007\170\007\171\007\
    \172\007\064\006\174\007\144\007\126\007\064\006\176\007\147\007\
    \177\007\178\007\064\006\180\007\064\006\182\007\183\007\184\007\
    \185\007\186\007\064\006\190\007\189\007\064\006\191\007\192\007\
    \064\006\064\006\218\007\196\007\217\007\209\007\199\007\200\007\
    \207\007\197\007\198\007\202\007\203\007\204\007\205\007\206\007\
    \064\006\201\007\208\007\064\006\210\007\211\007\212\007\213\007\
    \214\007\215\007\216\007\064\006\064\006\219\007\220\007\221\007\
    \222\007\223\007\224\007\064\006\228\007\227\007\064\006\229\007\
    \230\007\231\007\232\007\232\007\234\007\235\007\236\007\237\007\
    \238\007\239\007\064\006\011\008\242\007\250\007\244\007\243\007\
    \245\007\246\007\247\007\248\007\249\007\064\006\251\007\252\007\
    \253\007\232\007\232\007\254\007\254\007\007\008\001\008\002\008\
    \003\008\004\008\005\008\006\008\064\006\008\008\009\008\010\008\
    \064\006\064\006\030\008\027\008\031\008\021\008\016\008\017\008\
    \018\008\019\008\254\007\254\007\020\008\064\006\022\008\023\008\
    \024\008\025\008\026\008\064\006\028\008\029\008\064\006\035\008\
    \032\008\033\008\034\008\233\007\064\006\036\008\037\008\039\008\
    \043\008\040\008\041\008\042\008\064\006\044\008\045\008\046\008\
    \047\008\064\006\069\008\064\006\063\008\053\008\057\008\054\008\
    \038\008\055\008\056\008\064\006\000\008\058\008\059\008\060\008\
    \255\007\061\008\062\008\064\006\064\008\065\008\066\008\067\008\
    \052\008\068\008\064\006\064\006\071\008\072\008\064\006\074\008\
    \075\008\076\008\077\008\079\008\064\006\080\008\064\006\112\008\
    \107\008\084\008\086\008\099\008\092\008\087\008\078\008\088\008\
    \093\008\089\008\090\008\091\008\064\006\097\008\094\008\095\008\
    \096\008\085\008\064\006\098\008\064\006\100\008\101\008\102\008\
    \103\008\103\008\105\008\106\008\064\006\108\008\109\008\110\008\
    \111\008\064\006\113\008\114\008\115\008\116\008\064\006\118\008\
    \119\008\120\008\121\008\122\008\123\008\124\008\064\006\103\008\
    \103\008\125\008\064\006\132\008\064\006\129\008\130\008\131\008\
    \064\006\133\008\064\006\135\008\136\008\137\008\064\006\139\008\
    \140\008\141\008\064\006\143\008\064\006\143\005\146\008\147\008\
    \148\008\149\008\150\008\151\008\143\005\154\008\144\006\143\005\
    \166\008\163\008\143\005\159\008\160\008\161\008\162\008\144\006\
    \143\005\165\008\144\006\158\008\164\008\167\008\168\008\143\005\
    \170\008\171\008\143\005\173\008\104\008\174\008\175\008\176\008\
    \178\008\143\005\179\008\143\005\181\008\182\008\183\008\143\005\
    \194\008\191\008\206\008\177\008\192\008\190\008\205\008\191\008\
    \205\008\191\008\195\008\255\255\205\008\222\008\198\008\198\008\
    \197\008\191\008\205\008\191\008\195\008\216\008\205\008\194\008\
    \215\008\218\008\220\008\198\008\255\255\205\008\198\008\205\008\
    \242\008\193\008\198\008\198\008\190\008\198\008\198\008\197\008\
    \204\008\205\008\190\008\202\008\190\008\206\008\242\008\199\008\
    \199\008\196\008\198\008\189\008\190\008\198\008\190\008\197\008\
    \191\008\198\008\198\008\195\008\202\008\205\008\205\008\200\008\
    \225\008\205\008\215\008\201\008\206\008\214\008\209\008\255\255\
    \202\008\210\008\218\008\214\008\209\008\219\008\197\008\210\008\
    \220\008\225\008\008\009\221\008\205\008\225\008\187\008\225\008\
    \196\008\215\008\010\009\190\008\214\008\007\009\009\009\242\008\
    \242\008\190\008\214\008\246\008\224\008\245\008\248\008\240\008\
    \225\008\213\008\241\008\250\008\225\008\250\008\251\008\213\008\
    \254\008\255\008\005\009\006\009\010\009\224\008\017\009\023\009\
    \212\008\224\008\032\009\252\008\031\009\242\008\022\009\231\008\
    \233\008\235\008\234\008\020\009\227\008\237\008\253\008\236\008\
    \243\008\030\009\049\009\010\009\242\008\227\008\227\008\227\008\
    \227\008\227\008\227\008\227\008\227\008\227\008\227\008\211\008\
    \242\008\232\008\242\008\217\008\048\009\227\008\227\008\227\008\
    \227\008\227\008\227\008\227\008\227\008\227\008\227\008\227\008\
    \227\008\227\008\227\008\227\008\227\008\227\008\227\008\227\008\
    \227\008\227\008\227\008\227\008\227\008\227\008\227\008\227\008\
    \239\008\047\009\227\008\242\008\229\008\042\009\227\008\227\008\
    \227\008\227\008\227\008\227\008\227\008\227\008\227\008\227\008\
    \227\008\227\008\227\008\227\008\227\008\227\008\227\008\227\008\
    \227\008\227\008\227\008\227\008\227\008\227\008\227\008\227\008\
    \238\008\254\008\255\008\248\008\041\009\242\008\249\008\003\009\
    \001\009\004\009\004\009\002\009\042\009\004\009\004\009\001\009\
    \003\009\001\009\004\009\043\009\002\009\004\009\004\009\044\009\
    \053\009\004\009\065\009\072\009\016\009\071\009\003\009\188\008\
    \004\009\052\009\014\009\056\009\065\009\004\009\000\009\003\009\
    \000\009\255\255\088\009\088\009\004\009\000\009\039\009\000\009\
    \122\009\040\009\046\009\044\009\000\009\121\009\045\009\242\008\
    \029\009\123\009\255\255\028\009\046\009\044\009\113\009\026\009\
    \045\009\059\009\059\009\255\255\113\009\059\009\255\255\104\009\
    \105\009\046\009\103\009\102\009\037\009\128\009\036\009\112\009\
    \115\009\115\009\148\009\046\009\247\008\147\009\081\009\082\009\
    \059\009\080\009\079\009\130\009\146\009\100\009\101\009\255\255\
    \099\009\098\009\038\009\143\009\255\255\096\009\097\009\115\009\
    \095\009\094\009\255\255\139\009\120\009\255\255\119\009\138\009\
    \255\255\140\009\130\009\144\009\255\255\157\009\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \144\009\021\009\154\009\145\009\168\009\228\008\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\165\009\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\125\009\123\009\169\009\184\009\124\009\211\009\
    \199\009\190\009\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\249\009\250\009\196\009\248\009\
    \247\009\125\009\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\194\009\255\255\
    \015\009\255\255\197\009\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\027\009\059\009\059\009\
    \200\009\160\009\059\009\174\009\255\255\207\009\204\009\125\009\
    \123\009\255\255\255\255\124\009\035\009\208\009\209\009\212\009\
    \226\009\227\009\184\009\225\009\224\009\059\009\255\255\255\255\
    \160\009\233\009\174\009\233\009\002\010\068\009\125\009\067\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\118\009\255\255\255\255\115\009\
    \245\009\246\009\006\010\244\009\243\009\011\010\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \069\009\013\010\070\009\014\010\065\009\009\010\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \017\010\018\010\255\255\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\241\009\242\009\019\010\240\009\239\009\064\009\
    \063\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\062\009\061\009\021\010\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\060\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\066\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\255\255\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\010\010\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\142\009\
    \140\009\010\010\027\010\141\009\025\010\073\009\037\010\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\001\010\005\010\032\010\032\010\038\010\142\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\025\010\039\010\039\010\028\010\065\009\040\010\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\066\010\067\010\082\010\082\010\
    \088\009\097\010\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \097\010\078\009\077\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\076\009\075\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\015\010\
    \070\010\074\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\000\000\000\000\000\000\000\000\015\010\070\010\
    \000\000\000\000\000\000\087\009\086\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\085\009\
    \084\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\000\000\000\000\255\255\
    \255\255\000\000\000\000\083\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\000\000\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\000\000\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\000\000\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\116\010\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\000\000\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\000\000\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\142\009\140\009\000\000\000\000\141\009\000\000\073\009\
    \000\000\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\113\010\114\010\000\000\112\010\111\010\
    \142\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\000\000\000\000\000\000\000\000\
    \088\009\000\000\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\090\010\091\010\000\000\089\010\088\010\
    \255\255\000\000\000\000\093\009\092\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\091\009\
    \090\009\000\000\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\089\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \000\000\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \000\000\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\111\009\111\009\151\009\
    \174\009\111\009\111\009\111\009\116\010\153\009\111\009\156\009\
    \154\009\130\009\128\009\155\009\026\010\129\009\042\010\156\009\
    \154\009\044\010\160\009\155\009\111\009\000\000\110\009\174\009\
    \000\000\111\009\000\000\255\255\000\000\065\010\156\009\000\000\
    \130\009\000\000\108\009\132\009\134\009\133\009\156\009\112\009\
    \137\009\160\009\136\009\171\009\169\009\193\009\019\010\170\009\
    \000\000\162\009\167\009\165\009\167\009\165\009\166\009\164\009\
    \166\009\171\009\169\009\000\000\070\010\170\009\000\000\196\009\
    \197\009\000\000\171\009\198\009\193\009\019\010\199\009\200\009\
    \011\010\167\009\201\009\167\009\000\000\000\000\079\010\020\010\
    \171\009\177\009\190\009\070\010\080\010\191\009\196\009\068\010\
    \208\009\209\009\000\000\135\009\210\009\199\009\000\000\011\010\
    \213\009\213\009\021\010\010\010\213\009\079\010\022\010\000\000\
    \177\009\015\010\178\009\080\010\000\000\000\000\068\010\208\009\
    \186\009\193\009\194\009\000\000\000\000\195\009\187\009\213\009\
    \067\010\021\010\109\010\110\010\000\000\108\010\107\010\000\000\
    \015\010\192\009\000\000\000\000\000\000\079\010\161\009\000\000\
    \193\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\000\000\012\010\189\009\000\000\184\009\
    \000\000\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\185\009\
    \184\009\184\009\184\009\206\009\204\009\192\009\254\009\205\009\
    \016\010\206\009\204\009\000\000\000\000\205\009\036\010\032\010\
    \036\010\032\010\035\010\000\000\035\010\042\010\116\010\152\009\
    \044\010\042\010\206\009\000\000\044\010\254\009\000\000\253\009\
    \206\009\000\000\025\010\000\000\000\000\036\010\068\010\036\010\
    \202\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\000\000\107\009\255\255\000\000\255\255\
    \048\010\255\255\255\255\050\010\046\010\068\010\105\010\106\010\
    \131\009\104\010\103\010\000\000\255\255\000\000\000\000\069\010\
    \000\000\163\009\183\009\182\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\181\009\180\009\
    \000\000\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\179\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \188\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\000\000\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\000\000\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\000\000\203\009\000\000\000\000\255\255\000\000\214\009\
    \203\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\255\255\255\255\000\000\000\000\255\255\
    \000\000\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\000\000\000\000\255\255\000\000\
    \184\009\000\000\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\000\000\193\009\194\009\000\000\
    \000\000\195\009\000\000\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\193\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\000\000\219\009\218\009\000\000\000\000\000\000\
    \000\000\048\010\000\000\184\009\050\010\000\000\000\000\217\009\
    \216\009\192\009\214\009\000\000\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\000\000\000\000\
    \000\000\000\000\000\000\215\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\000\000\
    \000\000\000\000\000\000\184\009\000\000\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\220\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\000\000\000\000\059\010\000\000\233\009\058\010\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\000\000\219\009\218\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\217\009\216\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\000\000\000\000\215\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\000\000\
    \232\009\231\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\230\009\229\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\000\000\255\255\
    \228\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\000\000\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\000\000\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\255\255\000\000\000\000\000\000\000\000\000\000\214\009\
    \000\000\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\000\000\000\000\000\000\000\000\000\000\
    \000\000\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\000\000\000\000\000\000\000\000\
    \184\009\000\000\184\009\184\009\184\009\184\009\221\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\219\009\218\009\000\000\000\000\000\000\
    \000\000\000\000\000\000\184\009\000\000\000\000\000\000\217\009\
    \216\009\000\000\214\009\000\000\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\000\000\000\000\
    \000\000\000\000\000\000\215\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\000\000\
    \000\000\000\000\000\000\184\009\000\000\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\222\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\219\009\218\009\
    \000\000\000\000\000\000\000\000\000\000\000\000\184\009\000\000\
    \000\000\000\000\217\009\216\009\000\000\214\009\000\000\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\000\000\000\000\000\000\000\000\000\000\215\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\000\000\000\000\000\000\000\000\184\009\000\000\
    \184\009\184\009\184\009\184\009\223\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\000\000\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\219\009\218\009\000\000\000\000\000\000\000\000\000\000\
    \000\000\184\009\000\000\000\000\000\000\217\009\216\009\000\000\
    \214\009\000\000\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\000\000\000\000\000\000\000\000\
    \000\000\215\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\054\010\000\000\054\010\
    \058\010\184\009\058\010\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\000\000\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\000\000\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\219\009\218\009\059\010\000\000\
    \000\000\058\010\000\000\000\000\233\009\000\000\000\000\000\000\
    \217\009\216\009\000\000\214\009\000\000\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\000\000\
    \000\000\000\000\000\000\000\000\215\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \000\000\000\000\000\000\000\000\233\009\000\000\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\000\000\
    \000\000\000\000\000\000\255\255\000\000\255\255\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\000\000\238\009\
    \237\009\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\236\009\235\009\000\000\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\234\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\255\255\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\000\000\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\034\010\032\010\041\010\
    \042\010\033\010\042\010\044\010\048\010\044\010\000\000\050\010\
    \056\010\054\010\059\010\059\010\055\010\058\010\058\010\056\010\
    \054\010\000\000\054\010\057\010\034\010\062\010\041\010\054\010\
    \000\000\000\000\058\010\000\000\000\000\000\000\000\000\056\010\
    \000\000\000\000\000\000\045\010\000\000\047\010\056\010\047\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\000\000\000\000\
    \000\000\031\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\053\010\060\010\061\010\000\000\
    \000\000\000\000\000\000\053\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\000\000\000\000\
    \000\000\000\000\078\010\000\000\078\010\078\010\078\010\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\077\010\076\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\075\010\074\010\000\000\255\255\000\000\255\255\000\000\
    \255\255\000\000\255\255\000\000\000\000\000\000\000\000\255\255\
    \255\255\255\255\000\000\000\000\000\000\073\010\255\255\000\000\
    \255\255\000\000\000\000\000\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\255\255\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \000\000\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \080\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\000\000\000\000\080\010\
    \000\000\000\000\000\000\000\000\000\000\000\000\082\010\000\000\
    \000\000\000\000\000\000\000\000\000\000\081\010\000\000\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\079\010\000\000\000\000\000\000\000\000\000\000\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\000\000\000\000\000\000\000\000\082\010\000\000\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\000\000\000\000\000\000\000\000\
    \097\010\000\000\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \000\000\087\010\086\010\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\085\010\084\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\000\000\
    \000\000\083\010\000\000\000\000\000\000\000\000\080\010\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\096\010\095\010\080\010\000\000\000\000\
    \000\000\000\000\000\000\000\000\082\010\000\000\000\000\094\010\
    \093\010\000\000\000\000\081\010\000\000\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\079\010\
    \000\000\000\000\000\000\092\010\000\000\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \000\000\000\000\000\000\000\000\082\010\000\000\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\087\010\
    \086\010\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\085\010\084\010\000\000\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\083\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\000\000\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\000\000\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\000\000\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\000\000\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\080\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\080\010\000\000\000\000\000\000\000\000\000\000\000\000\
    \097\010\000\000\000\000\000\000\000\000\000\000\000\000\081\010\
    \000\000\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\079\010\000\000\000\000\000\000\000\000\
    \000\000\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\000\000\000\000\000\000\000\000\
    \097\010\000\000\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\102\010\101\010\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\100\010\
    \099\010\000\000\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\098\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \000\000\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \000\000\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\025\000\026\000\000\000\025\000\050\000\
    \056\000\056\000\068\000\107\000\056\000\069\000\070\000\071\000\
    \072\000\077\000\086\000\087\000\088\000\114\000\089\000\090\000\
    \000\000\097\000\125\000\000\000\116\000\168\004\172\004\056\000\
    \000\000\068\000\024\000\049\000\069\000\070\000\071\000\072\000\
    \077\000\086\000\087\000\088\000\092\000\089\000\090\000\104\000\
    \097\000\111\000\122\000\116\000\087\000\088\000\089\000\089\000\
    \090\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\156\002\209\003\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\207\002\
    \150\003\166\004\174\004\003\000\028\000\144\004\169\004\203\006\
    \014\007\005\000\007\000\010\000\007\000\008\000\013\000\009\000\
    \012\000\018\000\012\000\030\000\012\000\013\000\011\000\004\000\
    \006\000\009\000\014\000\015\000\010\000\008\000\012\000\013\000\
    \007\000\010\000\016\000\012\000\012\000\017\000\013\000\014\000\
    \017\000\015\000\022\000\019\000\020\000\017\000\021\000\017\000\
    \029\000\021\000\016\000\019\000\022\000\020\000\020\000\023\000\
    \031\000\022\000\023\000\032\000\019\000\032\000\023\000\019\000\
    \027\000\027\000\023\000\033\000\027\000\023\000\034\000\035\000\
    \036\000\043\000\046\000\053\000\039\000\046\000\040\000\054\000\
    \034\000\032\000\037\000\033\000\037\000\059\000\037\000\027\000\
    \035\000\039\000\027\000\038\000\040\000\035\000\041\000\027\000\
    \037\000\042\000\038\000\047\000\042\000\037\000\037\000\045\000\
    \060\000\042\000\055\000\042\000\038\000\047\000\041\000\062\000\
    \045\000\045\000\047\000\038\000\063\000\064\000\065\000\066\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\055\000\067\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\044\000\052\000\
    \048\000\073\000\058\000\048\000\051\000\051\000\044\000\048\000\
    \051\000\074\000\061\000\048\000\075\000\076\000\048\000\044\000\
    \078\000\079\000\044\000\080\000\081\000\082\000\052\000\083\000\
    \000\000\058\000\084\000\051\000\085\000\094\000\051\000\095\000\
    \052\000\061\000\096\000\051\000\069\000\070\000\071\000\078\000\
    \099\000\100\000\087\000\088\000\101\000\089\000\098\000\102\000\
    \108\000\109\000\115\000\055\000\118\000\119\000\055\000\120\000\
    \128\000\131\000\133\000\091\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\098\000\103\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\091\000\105\000\105\000\110\000\117\000\105\000\
    \052\000\121\000\134\000\052\000\091\000\103\000\112\000\112\000\
    \130\000\129\000\112\000\130\000\123\000\123\000\135\000\103\000\
    \123\000\058\000\105\000\058\000\110\000\117\000\126\000\129\000\
    \121\000\061\000\126\000\061\000\132\000\112\000\110\000\078\000\
    \136\000\137\000\121\000\123\000\138\000\139\000\132\000\130\000\
    \140\000\143\000\141\000\141\000\144\000\145\000\141\000\130\000\
    \130\000\146\000\147\000\149\000\130\000\098\000\150\000\151\000\
    \152\000\153\000\154\000\155\000\156\000\157\000\158\000\159\000\
    \098\000\141\000\160\000\161\000\162\000\163\000\164\000\165\000\
    \166\000\167\000\168\000\169\000\170\000\171\000\172\000\173\000\
    \174\000\175\000\176\000\177\000\179\000\180\000\181\000\182\000\
    \183\000\184\000\185\000\186\000\187\000\188\000\189\000\190\000\
    \191\000\193\000\192\000\194\000\195\000\194\000\196\000\197\000\
    \117\000\192\000\198\000\199\000\200\000\201\000\202\000\197\000\
    \203\000\204\000\205\000\206\000\207\000\208\000\209\000\210\000\
    \211\000\212\000\213\000\214\000\210\000\213\000\215\000\216\000\
    \217\000\218\000\199\000\200\000\219\000\220\000\221\000\222\000\
    \223\000\224\000\225\000\226\000\227\000\228\000\229\000\230\000\
    \231\000\230\000\232\000\233\000\236\000\237\000\234\000\238\000\
    \239\000\235\000\240\000\219\000\220\000\241\000\242\000\243\000\
    \244\000\245\000\246\000\247\000\248\000\249\000\200\000\200\000\
    \200\000\200\000\250\000\200\000\200\000\234\000\251\000\200\000\
    \235\000\252\000\200\000\253\000\254\000\200\000\200\000\200\000\
    \200\000\255\000\000\001\001\001\002\001\220\000\003\001\004\001\
    \005\001\250\000\006\001\007\001\008\001\251\000\009\001\007\001\
    \010\001\011\001\012\001\013\001\014\001\015\001\016\001\017\001\
    \018\001\019\001\020\001\021\001\022\001\023\001\024\001\025\001\
    \026\001\027\001\091\000\028\001\235\000\029\001\030\001\031\001\
    \032\001\033\001\034\001\035\001\036\001\037\001\038\001\039\001\
    \251\000\040\001\251\000\041\001\042\001\043\001\251\000\044\001\
    \045\001\046\001\047\001\048\001\251\000\049\001\051\001\251\000\
    \050\001\251\000\251\000\052\001\050\001\053\001\054\001\055\001\
    \056\001\056\001\057\001\058\001\059\001\060\001\061\001\062\001\
    \063\001\064\001\056\001\065\001\066\001\067\001\068\001\069\001\
    \070\001\071\001\072\001\073\001\074\001\075\001\076\001\077\001\
    \078\001\079\001\080\001\081\001\082\001\083\001\084\001\085\001\
    \086\001\087\001\088\001\066\001\067\001\089\001\090\001\091\001\
    \092\001\093\001\094\001\095\001\096\001\097\001\098\001\099\001\
    \100\001\093\001\101\001\082\001\083\001\102\001\103\001\104\001\
    \105\001\107\001\108\001\106\001\109\001\110\001\091\001\092\001\
    \111\001\112\001\113\001\114\001\115\001\116\001\117\001\118\001\
    \119\001\120\001\121\001\122\001\123\001\124\001\125\001\105\001\
    \067\001\067\001\106\001\126\001\127\001\128\001\067\001\067\001\
    \129\001\130\001\131\001\115\001\116\001\132\001\133\001\134\001\
    \135\001\136\001\137\001\138\001\139\001\140\001\143\001\083\001\
    \144\001\141\001\142\001\145\001\146\001\147\001\092\001\148\001\
    \149\001\150\001\151\001\152\001\153\001\154\001\155\001\156\001\
    \157\001\158\001\162\001\106\001\159\001\116\001\160\001\116\001\
    \141\001\142\001\106\001\161\001\161\001\163\001\164\001\165\001\
    \166\001\167\001\168\001\169\001\170\001\170\001\171\001\170\001\
    \173\001\116\001\174\001\159\001\172\001\160\001\175\001\172\001\
    \176\001\177\001\178\001\179\001\170\001\180\001\181\001\182\001\
    \183\001\180\001\142\001\184\001\185\001\186\001\187\001\188\001\
    \189\001\190\001\191\001\192\001\193\001\194\001\195\001\196\001\
    \177\001\199\001\200\001\201\001\197\001\198\001\202\001\204\001\
    \160\001\160\001\160\001\160\001\205\001\160\001\160\001\206\001\
    \207\001\160\001\208\001\209\001\160\001\210\001\211\001\160\001\
    \160\001\160\001\160\001\197\001\198\001\212\001\213\001\214\001\
    \215\001\216\001\217\001\218\001\219\001\220\001\221\001\222\001\
    \223\001\224\001\225\001\226\001\227\001\228\001\229\001\230\001\
    \231\001\233\001\234\001\236\001\237\001\232\001\235\001\238\001\
    \234\001\239\001\240\001\241\001\242\001\243\001\244\001\232\001\
    \233\001\225\001\226\001\232\001\235\001\232\001\245\001\246\001\
    \247\001\248\001\249\001\250\001\251\001\252\001\253\001\254\001\
    \198\001\240\001\241\001\255\001\000\002\001\002\002\002\003\002\
    \004\002\005\002\006\002\007\002\008\002\009\002\010\002\011\002\
    \012\002\011\002\013\002\014\002\015\002\016\002\017\002\020\002\
    \018\002\019\002\021\002\000\002\001\002\022\002\023\002\024\002\
    \025\002\026\002\027\002\028\002\029\002\226\001\030\002\031\002\
    \032\002\033\002\034\002\035\002\036\002\037\002\241\001\018\002\
    \019\002\038\002\039\002\040\002\041\002\030\002\042\002\043\002\
    \044\002\030\002\045\002\046\002\047\002\048\002\049\002\050\002\
    \051\002\001\002\052\002\053\002\054\002\055\002\056\002\057\002\
    \058\002\055\002\059\002\060\002\061\002\062\002\059\002\063\002\
    \064\002\065\002\066\002\067\002\068\002\069\002\070\002\072\002\
    \073\002\073\002\074\002\075\002\073\002\076\002\019\002\077\002\
    \078\002\079\002\073\002\080\002\081\002\082\002\083\002\084\002\
    \085\002\086\002\087\002\088\002\089\002\090\002\091\002\092\002\
    \093\002\094\002\095\002\096\002\083\002\094\002\098\002\099\002\
    \079\002\100\002\080\002\101\002\102\002\103\002\104\002\105\002\
    \106\002\107\002\108\002\109\002\110\002\111\002\112\002\113\002\
    \114\002\115\002\116\002\117\002\118\002\119\002\120\002\123\002\
    \124\002\121\002\122\002\125\002\126\002\127\002\128\002\129\002\
    \130\002\131\002\132\002\133\002\134\002\135\002\136\002\137\002\
    \136\002\138\002\139\002\080\002\140\002\141\002\080\002\080\002\
    \121\002\122\002\142\002\143\002\080\002\080\002\144\002\145\002\
    \146\002\136\002\147\002\148\002\149\002\150\002\151\002\152\002\
    \153\002\154\002\155\002\157\002\158\002\159\002\160\002\144\002\
    \161\002\162\002\163\002\164\002\165\002\166\002\167\002\168\002\
    \169\002\170\002\171\002\172\002\173\002\174\002\175\002\176\002\
    \177\002\178\002\179\002\180\002\181\002\182\002\177\002\122\002\
    \183\002\184\002\179\002\179\002\185\002\186\002\187\002\179\002\
    \188\002\189\002\183\002\190\002\191\002\192\002\193\002\195\002\
    \196\002\197\002\198\002\199\002\200\002\201\002\202\002\203\002\
    \204\002\205\002\206\002\208\002\209\002\210\002\211\002\212\002\
    \213\002\214\002\215\002\216\002\217\002\193\002\195\002\218\002\
    \219\002\218\002\221\002\222\002\220\002\224\002\225\002\224\002\
    \229\002\233\002\234\002\193\002\195\002\220\002\220\002\213\002\
    \214\002\223\002\227\002\225\002\223\002\218\002\224\002\224\002\
    \231\002\223\002\226\002\223\002\230\002\226\002\228\002\235\002\
    \236\002\228\002\231\002\227\002\230\002\228\002\237\002\238\002\
    \227\002\228\002\239\002\240\002\228\002\241\002\242\002\243\002\
    \230\002\214\002\214\002\214\002\214\002\214\002\214\002\214\002\
    \214\002\214\002\244\002\245\002\214\002\214\002\232\002\214\002\
    \214\002\214\002\214\002\214\002\214\002\232\002\214\002\246\002\
    \247\002\248\002\249\002\250\002\251\002\252\002\253\002\232\002\
    \254\002\255\002\193\002\195\002\193\002\195\002\232\002\000\003\
    \001\003\002\003\003\003\004\003\005\003\006\003\008\003\006\003\
    \007\003\009\003\010\003\011\003\252\002\253\002\012\003\007\003\
    \013\003\014\003\015\003\016\003\017\003\018\003\019\003\020\003\
    \021\003\022\003\023\003\024\003\025\003\026\003\027\003\028\003\
    \029\003\030\003\031\003\032\003\033\003\034\003\035\003\036\003\
    \038\003\039\003\037\003\041\003\042\003\043\003\040\003\039\003\
    \044\003\045\003\046\003\047\003\037\003\048\003\049\003\038\003\
    \030\003\031\003\037\003\253\002\040\003\050\003\051\003\052\003\
    \053\003\054\003\055\003\056\003\057\003\058\003\059\003\060\003\
    \045\003\046\003\061\003\062\003\063\003\064\003\065\003\066\003\
    \067\003\068\003\069\003\070\003\071\003\072\003\073\003\074\003\
    \075\003\076\003\077\003\078\003\079\003\076\003\080\003\079\003\
    \081\003\061\003\062\003\082\003\083\003\084\003\085\003\086\003\
    \087\003\088\003\081\003\089\003\031\003\090\003\091\003\092\003\
    \093\003\094\003\095\003\096\003\097\003\046\003\098\003\099\003\
    \100\003\101\003\102\003\079\003\103\003\104\003\105\003\106\003\
    \107\003\108\003\109\003\110\003\079\003\193\002\195\002\062\003\
    \079\003\111\003\112\003\113\003\114\003\115\003\116\003\117\003\
    \118\003\119\003\120\003\121\003\122\003\123\003\124\003\125\003\
    \126\003\127\003\128\003\129\003\130\003\131\003\132\003\133\003\
    \135\003\134\003\136\003\137\003\132\003\134\003\138\003\139\003\
    \140\003\141\003\142\003\143\003\136\003\144\003\145\003\146\003\
    \147\003\148\003\149\003\151\003\152\003\153\003\154\003\155\003\
    \156\003\157\003\158\003\159\003\160\003\161\003\162\003\163\003\
    \164\003\165\003\166\003\167\003\168\003\169\003\170\003\171\003\
    \172\003\173\003\174\003\175\003\176\003\177\003\178\003\179\003\
    \180\003\178\003\181\003\182\003\183\003\184\003\185\003\186\003\
    \187\003\188\003\189\003\168\003\169\003\190\003\191\003\192\003\
    \193\003\194\003\195\003\196\003\197\003\198\003\199\003\200\003\
    \201\003\202\003\203\003\204\003\205\003\206\003\205\003\207\003\
    \208\003\210\003\211\003\212\003\190\003\191\003\213\003\214\003\
    \215\003\216\003\217\003\218\003\219\003\220\003\221\003\222\003\
    \223\003\224\003\225\003\226\003\227\003\169\003\228\003\229\003\
    \230\003\231\003\232\003\233\003\234\003\235\003\236\003\237\003\
    \238\003\239\003\240\003\241\003\242\003\243\003\244\003\245\003\
    \246\003\230\003\247\003\248\003\249\003\250\003\191\003\251\003\
    \252\003\253\003\191\003\254\003\255\003\000\004\001\004\002\004\
    \003\004\244\003\004\004\005\004\006\004\007\004\008\004\009\004\
    \010\004\011\004\012\004\013\004\014\004\015\004\016\004\017\004\
    \018\004\019\004\020\004\020\004\021\004\022\004\023\004\013\004\
    \024\004\022\004\025\004\026\004\027\004\028\004\029\004\030\004\
    \031\004\032\004\020\004\033\004\034\004\035\004\036\004\037\004\
    \038\004\039\004\040\004\041\004\042\004\043\004\044\004\045\004\
    \046\004\047\004\048\004\049\004\050\004\051\004\052\004\053\004\
    \054\004\055\004\056\004\057\004\058\004\059\004\060\004\061\004\
    \039\004\040\004\060\004\062\004\063\004\064\004\065\004\066\004\
    \067\004\068\004\069\004\070\004\071\004\072\004\073\004\074\004\
    \075\004\076\004\077\004\078\004\079\004\080\004\081\004\082\004\
    \083\004\084\004\085\004\086\004\087\004\088\004\089\004\090\004\
    \091\004\092\004\093\004\094\004\095\004\096\004\097\004\098\004\
    \099\004\100\004\101\004\102\004\103\004\104\004\105\004\106\004\
    \107\004\108\004\109\004\110\004\111\004\040\004\112\004\113\004\
    \092\004\093\004\114\004\115\004\116\004\117\004\118\004\119\004\
    \113\004\120\004\119\004\121\004\122\004\123\004\124\004\125\004\
    \126\004\127\004\128\004\129\004\130\004\131\004\132\004\133\004\
    \134\004\135\004\136\004\137\004\138\004\139\004\141\004\141\004\
    \143\004\131\004\141\004\145\004\146\004\147\004\148\004\151\004\
    \148\004\149\004\150\004\152\004\160\004\154\004\178\004\180\004\
    \093\004\151\004\158\004\181\004\154\004\141\004\157\004\143\004\
    \141\004\149\004\155\004\150\004\148\004\141\004\154\004\153\004\
    \150\004\153\004\158\004\153\004\157\004\154\004\159\004\155\004\
    \170\004\159\004\177\004\162\004\163\004\153\004\159\004\163\004\
    \159\004\177\004\153\004\153\004\162\004\162\004\141\004\141\004\
    \141\004\141\004\141\004\141\004\141\004\141\004\141\004\170\004\
    \183\004\141\004\141\004\141\004\141\004\141\004\141\004\141\004\
    \141\004\141\004\141\004\141\004\156\004\161\004\164\004\156\004\
    \167\004\167\004\184\004\156\004\167\004\161\004\185\004\156\004\
    \164\004\186\004\156\004\171\004\171\004\164\004\161\004\171\004\
    \187\004\161\004\175\004\175\004\182\004\188\004\175\004\167\004\
    \179\004\184\004\179\004\189\004\182\004\185\004\190\004\191\004\
    \192\004\193\004\171\004\194\004\195\004\196\004\197\004\198\004\
    \199\004\175\004\196\004\199\004\170\004\200\004\201\004\202\004\
    \203\004\204\004\207\004\205\004\206\004\208\004\209\004\210\004\
    \211\004\212\004\213\004\214\004\216\004\217\004\218\004\217\004\
    \185\004\185\004\185\004\185\004\219\004\185\004\185\004\220\004\
    \223\004\185\004\205\004\206\004\185\004\221\004\222\004\185\004\
    \185\004\185\004\185\004\224\004\167\004\225\004\226\004\227\004\
    \228\004\229\004\230\004\231\004\232\004\233\004\234\004\171\004\
    \235\004\236\004\237\004\238\004\221\004\222\004\239\004\240\004\
    \241\004\242\004\243\004\244\004\206\004\245\004\246\004\247\004\
    \248\004\249\004\250\004\251\004\252\004\253\004\250\004\254\004\
    \255\004\237\004\238\004\000\005\001\005\002\005\003\005\004\005\
    \005\005\006\005\007\005\008\005\009\005\010\005\011\005\012\005\
    \013\005\014\005\015\005\016\005\017\005\018\005\019\005\020\005\
    \021\005\222\004\022\005\023\005\024\005\025\005\026\005\027\005\
    \028\005\029\005\030\005\031\005\032\005\238\004\033\005\238\004\
    \034\005\035\005\036\005\238\004\037\005\141\004\038\005\039\005\
    \037\005\238\004\040\005\041\005\238\004\042\005\238\004\238\004\
    \043\005\044\005\045\005\045\005\046\005\047\005\048\005\049\005\
    \050\005\051\005\052\005\053\005\045\005\054\005\055\005\056\005\
    \057\005\058\005\059\005\060\005\061\005\062\005\063\005\064\005\
    \065\005\066\005\067\005\068\005\069\005\070\005\071\005\072\005\
    \073\005\074\005\075\005\076\005\077\005\055\005\056\005\078\005\
    \079\005\080\005\081\005\082\005\083\005\084\005\085\005\086\005\
    \087\005\088\005\089\005\082\005\090\005\071\005\072\005\091\005\
    \092\005\093\005\094\005\096\005\097\005\095\005\098\005\099\005\
    \080\005\081\005\100\005\101\005\102\005\103\005\104\005\105\005\
    \106\005\107\005\108\005\109\005\110\005\111\005\112\005\113\005\
    \114\005\094\005\056\005\056\005\095\005\115\005\116\005\117\005\
    \056\005\056\005\118\005\119\005\120\005\104\005\105\005\121\005\
    \122\005\123\005\124\005\125\005\126\005\127\005\128\005\129\005\
    \132\005\072\005\133\005\130\005\131\005\134\005\135\005\136\005\
    \081\005\137\005\138\005\139\005\140\005\141\005\142\005\144\005\
    \145\005\146\005\147\005\148\005\149\005\095\005\150\005\105\005\
    \151\005\105\005\130\005\131\005\095\005\152\005\153\005\154\005\
    \154\005\155\005\156\005\157\005\158\005\159\005\160\005\161\005\
    \162\005\163\005\163\005\105\005\163\005\164\005\165\005\166\005\
    \167\005\165\005\168\005\169\005\152\005\153\005\170\005\171\005\
    \172\005\163\005\173\005\174\005\131\005\175\005\173\005\176\005\
    \177\005\178\005\179\005\180\005\181\005\182\005\183\005\184\005\
    \185\005\186\005\187\005\188\005\189\005\170\005\190\005\191\005\
    \192\005\193\005\194\005\195\005\197\005\198\005\199\005\200\005\
    \153\005\153\005\153\005\153\005\201\005\153\005\153\005\202\005\
    \203\005\153\005\204\005\205\005\153\005\190\005\191\005\153\005\
    \153\005\153\005\153\005\206\005\207\005\208\005\209\005\210\005\
    \211\005\212\005\213\005\214\005\215\005\216\005\217\005\218\005\
    \219\005\220\005\221\005\222\005\223\005\224\005\226\005\227\005\
    \229\005\230\005\225\005\228\005\231\005\227\005\232\005\233\005\
    \234\005\235\005\236\005\237\005\225\005\226\005\218\005\219\005\
    \225\005\228\005\225\005\238\005\239\005\240\005\241\005\242\005\
    \243\005\244\005\191\005\245\005\246\005\247\005\233\005\234\005\
    \248\005\249\005\250\005\251\005\252\005\253\005\254\005\255\005\
    \000\006\001\006\002\006\003\006\004\006\005\006\004\006\006\006\
    \007\006\008\006\009\006\010\006\013\006\011\006\012\006\014\006\
    \249\005\250\005\015\006\016\006\017\006\018\006\019\006\020\006\
    \021\006\022\006\219\005\023\006\024\006\025\006\026\006\027\006\
    \028\006\029\006\030\006\234\005\011\006\012\006\031\006\032\006\
    \033\006\034\006\023\006\035\006\036\006\037\006\023\006\038\006\
    \039\006\040\006\041\006\042\006\043\006\044\006\250\005\045\006\
    \046\006\047\006\048\006\049\006\050\006\051\006\048\006\052\006\
    \053\006\054\006\055\006\052\006\056\006\057\006\058\006\059\006\
    \060\006\061\006\062\006\063\006\065\006\066\006\067\006\068\006\
    \065\006\067\006\069\006\012\006\070\006\071\006\072\006\073\006\
    \074\006\075\006\076\006\077\006\078\006\068\006\079\006\080\006\
    \072\006\081\006\082\006\083\006\084\006\085\006\086\006\087\006\
    \088\006\089\006\090\006\091\006\092\006\067\006\093\006\094\006\
    \095\006\096\006\097\006\098\006\099\006\100\006\067\006\101\006\
    \102\006\103\006\067\006\104\006\105\006\106\006\107\006\108\006\
    \109\006\110\006\111\006\112\006\113\006\114\006\115\006\116\006\
    \117\006\118\006\119\006\120\006\120\006\121\006\122\006\120\006\
    \123\006\124\006\125\006\126\006\127\006\120\006\128\006\129\006\
    \130\006\131\006\132\006\133\006\134\006\135\006\136\006\137\006\
    \138\006\139\006\140\006\141\006\142\006\143\006\130\006\141\006\
    \145\006\146\006\126\006\127\006\147\006\148\006\149\006\150\006\
    \151\006\152\006\153\006\154\006\155\006\156\006\157\006\158\006\
    \159\006\160\006\161\006\162\006\163\006\164\006\165\006\166\006\
    \167\006\170\006\168\006\169\006\171\006\172\006\173\006\174\006\
    \175\006\176\006\177\006\178\006\179\006\180\006\181\006\182\006\
    \183\006\184\006\183\006\185\006\127\006\186\006\187\006\127\006\
    \127\006\168\006\169\006\188\006\189\006\127\006\127\006\190\006\
    \191\006\192\006\193\006\183\006\194\006\195\006\196\006\197\006\
    \198\006\199\006\200\006\201\006\202\006\204\006\205\006\206\006\
    \207\006\191\006\208\006\209\006\210\006\211\006\212\006\213\006\
    \214\006\215\006\216\006\217\006\218\006\219\006\220\006\221\006\
    \222\006\223\006\224\006\225\006\226\006\227\006\228\006\229\006\
    \169\006\230\006\231\006\232\006\234\006\235\006\236\006\237\006\
    \238\006\239\006\240\006\241\006\242\006\243\006\244\006\222\006\
    \223\006\241\006\245\006\246\006\247\006\243\006\243\006\248\006\
    \249\006\250\006\243\006\251\006\252\006\246\006\253\006\254\006\
    \255\006\000\007\001\007\002\007\003\007\004\007\005\007\006\007\
    \007\007\008\007\009\007\010\007\011\007\012\007\248\006\249\006\
    \013\007\015\007\016\007\017\007\018\007\019\007\020\007\022\007\
    \023\007\024\007\026\007\027\007\248\006\249\006\025\007\223\006\
    \025\007\028\007\029\007\036\007\027\007\027\007\030\007\038\007\
    \031\007\030\007\031\007\032\007\034\007\020\007\030\007\021\007\
    \030\007\038\007\033\007\040\007\025\007\033\007\041\007\042\007\
    \032\007\031\007\031\007\043\007\035\007\034\007\039\007\035\007\
    \037\007\044\007\034\007\035\007\045\007\039\007\021\007\035\007\
    \037\007\046\007\035\007\047\007\048\007\049\007\050\007\039\007\
    \051\007\052\007\053\007\054\007\037\007\055\007\039\007\056\007\
    \057\007\058\007\059\007\061\007\062\007\060\007\063\007\064\007\
    \065\007\066\007\067\007\248\006\249\006\248\006\249\006\021\007\
    \021\007\021\007\021\007\021\007\021\007\021\007\021\007\021\007\
    \068\007\059\007\021\007\021\007\060\007\021\007\021\007\021\007\
    \021\007\021\007\021\007\069\007\021\007\069\007\070\007\071\007\
    \072\007\073\007\074\007\075\007\076\007\070\007\077\007\078\007\
    \079\007\080\007\081\007\082\007\083\007\084\007\085\007\086\007\
    \087\007\088\007\089\007\090\007\091\007\092\007\093\007\094\007\
    \095\007\096\007\097\007\098\007\099\007\101\007\102\007\100\007\
    \104\007\105\007\060\007\103\007\102\007\106\007\107\007\108\007\
    \109\007\100\007\110\007\111\007\101\007\093\007\094\007\100\007\
    \112\007\103\007\113\007\114\007\115\007\116\007\117\007\118\007\
    \119\007\120\007\121\007\122\007\123\007\124\007\108\007\109\007\
    \125\007\126\007\127\007\128\007\129\007\130\007\131\007\132\007\
    \133\007\134\007\135\007\136\007\137\007\138\007\139\007\140\007\
    \141\007\143\007\139\007\142\007\124\007\144\007\142\007\125\007\
    \145\007\146\007\147\007\148\007\149\007\150\007\151\007\144\007\
    \152\007\094\007\153\007\154\007\155\007\156\007\157\007\158\007\
    \159\007\160\007\161\007\109\007\162\007\163\007\248\006\249\006\
    \164\007\165\007\142\007\166\007\167\007\168\007\169\007\170\007\
    \171\007\172\007\173\007\142\007\125\007\174\007\175\007\142\007\
    \176\007\177\007\178\007\179\007\180\007\181\007\182\007\183\007\
    \184\007\185\007\186\007\187\007\188\007\189\007\190\007\191\007\
    \192\007\193\007\194\007\195\007\196\007\198\007\197\007\199\007\
    \200\007\195\007\197\007\201\007\202\007\203\007\204\007\205\007\
    \206\007\199\007\207\007\208\007\209\007\210\007\211\007\212\007\
    \213\007\214\007\215\007\216\007\217\007\218\007\219\007\220\007\
    \221\007\222\007\223\007\224\007\225\007\226\007\227\007\228\007\
    \229\007\230\007\231\007\232\007\233\007\234\007\235\007\236\007\
    \237\007\238\007\239\007\240\007\241\007\242\007\243\007\241\007\
    \244\007\245\007\246\007\247\007\248\007\249\007\250\007\251\007\
    \252\007\231\007\232\007\253\007\254\007\255\007\000\008\001\008\
    \002\008\003\008\004\008\005\008\006\008\007\008\008\008\009\008\
    \010\008\011\008\012\008\013\008\012\008\014\008\015\008\016\008\
    \017\008\018\008\253\007\254\007\019\008\020\008\021\008\022\008\
    \023\008\024\008\025\008\026\008\027\008\028\008\029\008\030\008\
    \031\008\032\008\033\008\232\007\034\008\035\008\036\008\037\008\
    \038\008\039\008\040\008\041\008\042\008\043\008\044\008\045\008\
    \046\008\047\008\048\008\049\008\050\008\051\008\052\008\053\008\
    \037\008\054\008\055\008\056\008\254\007\057\008\058\008\059\008\
    \254\007\060\008\061\008\062\008\063\008\064\008\065\008\066\008\
    \051\008\067\008\068\008\069\008\070\008\071\008\072\008\073\008\
    \074\008\075\008\076\008\077\008\078\008\079\008\080\008\081\008\
    \082\008\083\008\083\008\084\008\085\008\086\008\076\008\087\008\
    \085\008\088\008\089\008\090\008\091\008\092\008\093\008\094\008\
    \095\008\083\008\096\008\097\008\098\008\099\008\100\008\101\008\
    \102\008\103\008\104\008\105\008\106\008\107\008\108\008\109\008\
    \110\008\111\008\112\008\113\008\114\008\115\008\116\008\117\008\
    \118\008\119\008\120\008\121\008\122\008\123\008\124\008\102\008\
    \103\008\123\008\125\008\126\008\127\008\128\008\129\008\130\008\
    \131\008\132\008\133\008\134\008\135\008\136\008\137\008\138\008\
    \139\008\140\008\141\008\142\008\143\008\144\008\145\008\146\008\
    \147\008\148\008\149\008\150\008\151\008\152\008\153\008\154\008\
    \155\008\156\008\157\008\158\008\159\008\160\008\161\008\162\008\
    \164\008\163\008\165\008\157\008\163\008\166\008\167\008\168\008\
    \169\008\170\008\171\008\172\008\103\008\173\008\174\008\175\008\
    \176\008\177\008\178\008\179\008\180\008\181\008\182\008\183\008\
    \185\008\185\008\189\008\175\008\185\008\190\008\191\008\191\008\
    \192\008\192\008\191\008\203\008\192\008\210\008\193\008\196\008\
    \194\008\194\008\195\008\195\008\194\008\211\008\195\008\185\008\
    \212\008\219\008\221\008\198\008\198\008\191\008\199\008\192\008\
    \230\008\185\008\200\008\201\008\185\008\193\008\196\008\194\008\
    \202\008\195\008\191\008\203\008\192\008\206\008\231\008\193\008\
    \196\008\194\008\198\008\185\008\194\008\199\008\195\008\197\008\
    \197\008\200\008\201\008\197\008\198\008\205\008\205\008\199\008\
    \209\008\205\008\215\008\200\008\206\008\207\008\207\008\202\008\
    \202\008\207\008\216\008\214\008\214\008\216\008\197\008\214\008\
    \217\008\222\008\234\008\217\008\205\008\225\008\185\008\209\008\
    \197\008\215\008\232\008\197\008\207\008\235\008\233\008\238\008\
    \240\008\205\008\214\008\239\008\209\008\239\008\249\008\226\008\
    \222\008\207\008\226\008\247\008\225\008\251\008\247\008\214\008\
    \255\008\255\008\000\009\005\009\010\009\222\008\014\009\020\009\
    \207\008\225\008\026\009\238\008\028\009\232\008\018\009\226\008\
    \226\008\226\008\226\008\018\009\226\008\226\008\237\008\226\008\
    \241\008\029\009\036\009\010\009\234\008\226\008\226\008\226\008\
    \226\008\226\008\226\008\226\008\226\008\226\008\226\008\207\008\
    \235\008\226\008\233\008\216\008\037\009\226\008\226\008\226\008\
    \226\008\226\008\226\008\226\008\226\008\226\008\226\008\226\008\
    \226\008\226\008\226\008\226\008\226\008\226\008\226\008\226\008\
    \226\008\226\008\226\008\226\008\226\008\226\008\226\008\226\008\
    \226\008\038\009\226\008\237\008\226\008\039\009\226\008\226\008\
    \226\008\226\008\226\008\226\008\226\008\226\008\226\008\226\008\
    \226\008\226\008\226\008\226\008\226\008\226\008\226\008\226\008\
    \226\008\226\008\226\008\226\008\226\008\226\008\226\008\226\008\
    \226\008\236\008\236\008\246\008\040\009\241\008\246\008\254\008\
    \254\008\001\009\001\009\254\008\041\009\001\009\002\009\002\009\
    \003\009\003\009\002\009\042\009\003\009\004\009\004\009\045\009\
    \052\009\004\009\064\009\067\009\012\009\068\009\254\008\185\008\
    \001\009\050\009\012\009\054\009\078\009\002\009\254\008\003\009\
    \001\009\203\008\087\009\093\009\004\009\002\009\033\009\003\009\
    \119\009\033\009\043\009\043\009\004\009\120\009\043\009\236\008\
    \024\009\124\009\198\008\024\009\046\009\046\009\110\009\024\009\
    \046\009\059\009\059\009\109\009\114\009\059\009\202\008\060\009\
    \060\009\043\009\060\009\060\009\033\009\129\009\033\009\109\009\
    \108\009\108\009\132\009\046\009\246\008\133\009\074\009\074\009\
    \059\009\074\009\074\009\130\009\134\009\083\009\083\009\209\008\
    \083\009\083\009\033\009\135\009\207\008\089\009\089\009\108\009\
    \089\009\089\009\214\008\136\009\116\009\229\008\116\009\137\009\
    \222\008\141\009\130\009\145\009\225\008\151\009\229\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \143\009\018\009\155\009\143\009\162\009\226\008\229\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \229\008\229\008\166\009\229\008\050\009\229\008\054\009\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \229\008\242\008\122\009\122\009\170\009\183\009\122\009\186\009\
    \189\009\191\009\242\008\242\008\242\008\242\008\242\008\242\008\
    \242\008\242\008\242\008\242\008\179\009\179\009\192\009\179\009\
    \179\009\122\009\242\008\242\008\242\008\242\008\242\008\242\008\
    \242\008\242\008\242\008\242\008\242\008\242\008\242\008\242\008\
    \242\008\242\008\242\008\242\008\242\008\242\008\242\008\242\008\
    \242\008\242\008\242\008\242\008\242\008\242\008\195\009\242\008\
    \012\009\242\008\198\009\242\008\242\008\242\008\242\008\242\008\
    \242\008\242\008\242\008\242\008\242\008\242\008\242\008\242\008\
    \242\008\242\008\242\008\242\008\242\008\242\008\242\008\242\008\
    \242\008\242\008\242\008\242\008\242\008\024\009\057\009\057\009\
    \201\009\160\009\057\009\174\009\110\009\204\009\205\009\125\009\
    \125\009\109\009\114\009\125\009\033\009\207\009\210\009\211\009\
    \215\009\215\009\219\009\215\009\215\009\057\009\112\009\112\009\
    \160\009\232\009\174\009\238\009\001\010\057\009\125\009\057\009\
    \081\009\081\009\081\009\081\009\081\009\081\009\081\009\081\009\
    \081\009\081\009\081\009\081\009\116\009\112\009\229\008\108\009\
    \228\009\228\009\005\010\228\009\228\009\009\010\057\009\057\009\
    \057\009\057\009\057\009\057\009\057\009\057\009\057\009\057\009\
    \057\009\057\009\057\009\057\009\057\009\057\009\057\009\057\009\
    \057\009\057\009\057\009\057\009\057\009\057\009\057\009\057\009\
    \057\009\012\010\057\009\013\010\057\009\007\010\057\009\057\009\
    \057\009\057\009\057\009\057\009\057\009\057\009\057\009\057\009\
    \057\009\057\009\057\009\057\009\057\009\057\009\057\009\057\009\
    \057\009\057\009\057\009\057\009\057\009\057\009\057\009\057\009\
    \079\009\079\009\079\009\079\009\079\009\079\009\079\009\079\009\
    \079\009\079\009\079\009\079\009\079\009\079\009\079\009\079\009\
    \016\010\017\010\242\008\061\009\061\009\061\009\061\009\061\009\
    \061\009\061\009\061\009\061\009\061\009\061\009\061\009\061\009\
    \061\009\061\009\061\009\061\009\061\009\061\009\061\009\061\009\
    \061\009\061\009\061\009\061\009\061\009\061\009\061\009\061\009\
    \061\009\061\009\061\009\061\009\061\009\061\009\061\009\061\009\
    \061\009\061\009\061\009\061\009\061\009\061\009\061\009\061\009\
    \061\009\061\009\234\009\234\009\018\010\234\009\234\009\057\009\
    \057\009\082\009\082\009\082\009\082\009\082\009\082\009\082\009\
    \082\009\082\009\082\009\057\009\057\009\020\010\062\009\062\009\
    \062\009\062\009\062\009\062\009\062\009\062\009\062\009\062\009\
    \062\009\062\009\062\009\062\009\062\009\062\009\062\009\057\009\
    \062\009\062\009\062\009\062\009\062\009\062\009\062\009\062\009\
    \062\009\062\009\062\009\062\009\062\009\062\009\062\009\062\009\
    \062\009\062\009\062\009\062\009\062\009\062\009\062\009\062\009\
    \062\009\062\009\062\009\062\009\062\009\057\009\063\009\063\009\
    \063\009\063\009\063\009\063\009\063\009\063\009\063\009\063\009\
    \063\009\063\009\063\009\063\009\063\009\063\009\063\009\063\009\
    \063\009\063\009\063\009\063\009\063\009\112\009\063\009\063\009\
    \063\009\063\009\063\009\063\009\063\009\063\009\063\009\063\009\
    \063\009\063\009\063\009\063\009\063\009\063\009\063\009\063\009\
    \063\009\063\009\063\009\063\009\063\009\063\009\063\009\063\009\
    \063\009\063\009\063\009\063\009\063\009\007\010\063\009\063\009\
    \063\009\063\009\063\009\063\009\063\009\063\009\065\009\139\009\
    \139\009\022\010\026\010\139\009\028\010\065\009\031\010\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\255\009\003\010\033\010\035\010\037\010\139\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\027\010\040\010\038\010\027\010\065\009\038\010\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\073\009\073\009\073\009\073\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\073\009\073\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\073\009\073\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\065\010\066\010\077\010\087\010\
    \073\009\096\010\073\009\073\009\073\009\073\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\073\009\073\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\073\009\073\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\096\009\096\009\096\009\096\009\
    \096\009\096\009\096\009\096\009\096\009\096\009\096\009\096\009\
    \102\010\065\009\065\009\097\009\097\009\097\009\097\009\097\009\
    \097\009\097\009\097\009\097\009\097\009\065\009\065\009\094\009\
    \094\009\094\009\094\009\094\009\094\009\094\009\094\009\094\009\
    \094\009\094\009\094\009\094\009\094\009\094\009\094\009\014\010\
    \069\010\065\009\098\009\098\009\098\009\098\009\098\009\098\009\
    \098\009\098\009\098\009\098\009\098\009\098\009\098\009\098\009\
    \098\009\098\009\255\255\255\255\255\255\255\255\014\010\069\010\
    \255\255\255\255\255\255\073\009\073\009\101\009\101\009\101\009\
    \101\009\101\009\101\009\101\009\101\009\101\009\101\009\073\009\
    \073\009\100\009\100\009\100\009\100\009\100\009\100\009\100\009\
    \100\009\100\009\100\009\100\009\100\009\255\255\255\255\255\009\
    \003\010\255\255\255\255\073\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\075\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\075\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\075\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\075\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\075\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\076\009\076\009\076\009\076\009\076\009\
    \076\009\076\009\076\009\076\009\076\009\076\009\076\009\076\009\
    \076\009\076\009\076\009\076\009\255\255\076\009\076\009\076\009\
    \076\009\076\009\076\009\076\009\076\009\076\009\076\009\076\009\
    \076\009\076\009\076\009\076\009\076\009\076\009\076\009\076\009\
    \076\009\076\009\076\009\076\009\076\009\076\009\076\009\076\009\
    \076\009\076\009\077\009\077\009\077\009\077\009\077\009\077\009\
    \077\009\077\009\077\009\077\009\077\009\077\009\077\009\077\009\
    \077\009\077\009\077\009\077\009\077\009\077\009\077\009\077\009\
    \077\009\255\255\077\009\077\009\077\009\077\009\077\009\077\009\
    \077\009\077\009\077\009\077\009\077\009\077\009\077\009\077\009\
    \077\009\077\009\077\009\077\009\077\009\077\009\077\009\077\009\
    \077\009\077\009\077\009\077\009\077\009\077\009\077\009\077\009\
    \077\009\255\255\077\009\077\009\077\009\077\009\077\009\077\009\
    \077\009\077\009\080\009\080\009\080\009\080\009\080\009\080\009\
    \080\009\080\009\080\009\080\009\080\009\080\009\080\009\080\009\
    \080\009\080\009\080\009\080\009\080\009\080\009\080\009\080\009\
    \080\009\080\009\080\009\080\009\080\009\080\009\080\009\080\009\
    \080\009\080\009\080\009\080\009\080\009\080\009\080\009\080\009\
    \080\009\080\009\080\009\080\009\080\009\080\009\080\009\080\009\
    \080\009\080\009\080\009\080\009\080\009\080\009\080\009\080\009\
    \080\009\080\009\080\009\080\009\080\009\080\009\080\009\080\009\
    \080\009\080\009\084\009\084\009\084\009\084\009\084\009\084\009\
    \084\009\084\009\084\009\084\009\084\009\084\009\084\009\084\009\
    \084\009\084\009\084\009\084\009\084\009\084\009\084\009\084\009\
    \084\009\084\009\084\009\084\009\084\009\084\009\084\009\084\009\
    \084\009\084\009\084\009\084\009\084\009\084\009\084\009\084\009\
    \084\009\084\009\084\009\084\009\084\009\084\009\084\009\084\009\
    \084\009\085\009\085\009\085\009\085\009\085\009\085\009\085\009\
    \085\009\085\009\085\009\085\009\085\009\085\009\085\009\085\009\
    \085\009\085\009\115\010\085\009\085\009\085\009\085\009\085\009\
    \085\009\085\009\085\009\085\009\085\009\085\009\085\009\085\009\
    \085\009\085\009\085\009\085\009\085\009\085\009\085\009\085\009\
    \085\009\085\009\085\009\085\009\085\009\085\009\085\009\085\009\
    \086\009\086\009\086\009\086\009\086\009\086\009\086\009\086\009\
    \086\009\086\009\086\009\086\009\086\009\086\009\086\009\086\009\
    \086\009\086\009\086\009\086\009\086\009\086\009\086\009\255\255\
    \086\009\086\009\086\009\086\009\086\009\086\009\086\009\086\009\
    \086\009\086\009\086\009\086\009\086\009\086\009\086\009\086\009\
    \086\009\086\009\086\009\086\009\086\009\086\009\086\009\086\009\
    \086\009\086\009\086\009\086\009\086\009\086\009\086\009\255\255\
    \086\009\086\009\086\009\086\009\086\009\086\009\086\009\086\009\
    \088\009\142\009\142\009\255\255\255\255\142\009\255\255\088\009\
    \255\255\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\073\010\073\010\255\255\073\010\073\010\
    \142\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\255\255\255\255\255\255\255\255\
    \088\009\255\255\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\088\009\088\009\088\009\088\009\
    \088\009\088\009\088\009\088\009\090\009\090\009\090\009\090\009\
    \090\009\090\009\090\009\090\009\090\009\090\009\090\009\090\009\
    \090\009\090\009\090\009\090\009\090\009\090\009\090\009\090\009\
    \090\009\090\009\090\009\090\009\090\009\090\009\090\009\090\009\
    \090\009\090\009\090\009\090\009\090\009\090\009\090\009\090\009\
    \090\009\090\009\090\009\090\009\090\009\090\009\090\009\090\009\
    \090\009\090\009\090\009\102\009\102\009\102\009\102\009\102\009\
    \102\009\102\009\102\009\102\009\102\009\102\009\102\009\102\009\
    \102\009\102\009\102\009\083\010\083\010\255\255\083\010\083\010\
    \115\010\255\255\255\255\088\009\088\009\105\009\105\009\105\009\
    \105\009\105\009\105\009\105\009\105\009\105\009\105\009\088\009\
    \088\009\255\255\091\009\091\009\091\009\091\009\091\009\091\009\
    \091\009\091\009\091\009\091\009\091\009\091\009\091\009\091\009\
    \091\009\091\009\091\009\088\009\091\009\091\009\091\009\091\009\
    \091\009\091\009\091\009\091\009\091\009\091\009\091\009\091\009\
    \091\009\091\009\091\009\091\009\091\009\091\009\091\009\091\009\
    \091\009\091\009\091\009\091\009\091\009\091\009\091\009\091\009\
    \091\009\092\009\092\009\092\009\092\009\092\009\092\009\092\009\
    \092\009\092\009\092\009\092\009\092\009\092\009\092\009\092\009\
    \092\009\092\009\092\009\092\009\092\009\092\009\092\009\092\009\
    \255\255\092\009\092\009\092\009\092\009\092\009\092\009\092\009\
    \092\009\092\009\092\009\092\009\092\009\092\009\092\009\092\009\
    \092\009\092\009\092\009\092\009\092\009\092\009\092\009\092\009\
    \092\009\092\009\092\009\092\009\092\009\092\009\092\009\092\009\
    \255\255\092\009\092\009\092\009\092\009\092\009\092\009\092\009\
    \092\009\095\009\095\009\095\009\095\009\095\009\095\009\095\009\
    \095\009\095\009\095\009\095\009\095\009\095\009\095\009\095\009\
    \095\009\095\009\095\009\095\009\095\009\095\009\095\009\095\009\
    \095\009\095\009\095\009\095\009\095\009\095\009\095\009\095\009\
    \095\009\095\009\095\009\095\009\095\009\095\009\095\009\095\009\
    \095\009\095\009\095\009\095\009\095\009\095\009\095\009\095\009\
    \095\009\095\009\095\009\095\009\095\009\095\009\095\009\095\009\
    \095\009\095\009\095\009\095\009\095\009\095\009\095\009\095\009\
    \095\009\099\009\099\009\099\009\099\009\099\009\099\009\099\009\
    \099\009\099\009\099\009\099\009\099\009\099\009\099\009\099\009\
    \099\009\099\009\099\009\099\009\099\009\099\009\099\009\099\009\
    \099\009\099\009\099\009\099\009\099\009\099\009\099\009\099\009\
    \099\009\099\009\099\009\099\009\099\009\099\009\099\009\099\009\
    \099\009\099\009\099\009\099\009\099\009\099\009\099\009\099\009\
    \099\009\099\009\099\009\099\009\099\009\099\009\099\009\099\009\
    \099\009\099\009\099\009\099\009\099\009\099\009\099\009\099\009\
    \099\009\103\009\103\009\103\009\103\009\103\009\103\009\103\009\
    \103\009\103\009\103\009\103\009\103\009\103\009\103\009\103\009\
    \103\009\103\009\103\009\103\009\103\009\103\009\103\009\103\009\
    \103\009\103\009\103\009\103\009\103\009\103\009\103\009\103\009\
    \103\009\103\009\103\009\103\009\103\009\103\009\103\009\103\009\
    \103\009\103\009\103\009\103\009\103\009\103\009\103\009\103\009\
    \103\009\103\009\103\009\103\009\103\009\103\009\103\009\103\009\
    \103\009\103\009\103\009\103\009\103\009\103\009\103\009\103\009\
    \103\009\104\009\104\009\104\009\104\009\104\009\104\009\104\009\
    \104\009\104\009\104\009\104\009\104\009\106\009\106\009\149\009\
    \172\009\106\009\111\009\111\009\117\010\149\009\111\009\153\009\
    \153\009\126\009\126\009\153\009\023\010\126\009\043\010\156\009\
    \156\009\043\010\158\009\156\009\106\009\255\255\106\009\172\009\
    \255\255\111\009\255\255\111\009\255\255\063\010\153\009\255\255\
    \126\009\255\255\106\009\126\009\126\009\126\009\156\009\111\009\
    \126\009\158\009\126\009\161\009\161\009\193\009\019\010\161\009\
    \255\255\158\009\164\009\164\009\167\009\167\009\164\009\158\009\
    \167\009\171\009\171\009\255\255\070\010\171\009\255\255\196\009\
    \196\009\255\255\161\009\196\009\193\009\019\010\199\009\199\009\
    \011\010\164\009\199\009\167\009\255\255\255\255\079\010\019\010\
    \171\009\175\009\175\009\070\010\080\010\175\009\196\009\067\010\
    \208\009\208\009\255\255\126\009\208\009\199\009\255\255\011\010\
    \212\009\212\009\021\010\021\010\212\009\079\010\021\010\255\255\
    \175\009\015\010\175\009\080\010\255\255\255\255\067\010\208\009\
    \175\009\190\009\190\009\255\255\255\255\190\009\175\009\212\009\
    \067\010\021\010\092\010\092\010\255\255\092\010\092\010\255\255\
    \015\010\193\009\255\255\255\255\255\255\080\010\158\009\255\255\
    \190\009\175\009\175\009\175\009\175\009\175\009\175\009\175\009\
    \175\009\175\009\175\009\175\009\175\009\175\009\175\009\175\009\
    \175\009\175\009\175\009\175\009\175\009\175\009\175\009\175\009\
    \175\009\175\009\175\009\255\255\011\010\175\009\255\255\175\009\
    \255\255\175\009\175\009\175\009\175\009\175\009\175\009\175\009\
    \175\009\175\009\175\009\175\009\175\009\175\009\175\009\175\009\
    \175\009\175\009\175\009\175\009\175\009\175\009\175\009\175\009\
    \175\009\175\009\175\009\187\009\187\009\190\009\251\009\187\009\
    \015\010\206\009\206\009\255\255\255\255\206\009\034\010\034\010\
    \036\010\036\010\034\010\255\255\036\010\044\010\118\010\149\009\
    \044\010\045\010\187\009\255\255\045\010\251\009\255\255\251\009\
    \206\009\255\255\023\010\255\255\255\255\034\010\068\010\036\010\
    \187\009\227\009\227\009\227\009\227\009\227\009\227\009\227\009\
    \227\009\227\009\227\009\255\255\106\009\063\010\255\255\172\009\
    \049\010\111\009\117\010\049\010\045\010\068\010\098\010\098\010\
    \126\009\098\010\098\010\255\255\043\010\255\255\255\255\068\010\
    \255\255\158\009\175\009\175\009\242\009\242\009\242\009\242\009\
    \242\009\242\009\242\009\242\009\242\009\242\009\175\009\175\009\
    \255\255\181\009\181\009\181\009\181\009\181\009\181\009\181\009\
    \181\009\181\009\181\009\181\009\181\009\181\009\181\009\181\009\
    \181\009\181\009\175\009\181\009\181\009\181\009\181\009\181\009\
    \181\009\181\009\181\009\181\009\181\009\181\009\181\009\181\009\
    \181\009\181\009\181\009\181\009\181\009\181\009\181\009\181\009\
    \181\009\181\009\181\009\181\009\181\009\181\009\181\009\181\009\
    \175\009\180\009\180\009\180\009\180\009\180\009\180\009\180\009\
    \180\009\180\009\180\009\180\009\180\009\180\009\180\009\180\009\
    \180\009\180\009\180\009\180\009\180\009\180\009\180\009\180\009\
    \180\009\180\009\180\009\180\009\180\009\180\009\180\009\180\009\
    \180\009\180\009\180\009\180\009\180\009\180\009\180\009\180\009\
    \180\009\180\009\180\009\180\009\180\009\180\009\180\009\180\009\
    \182\009\182\009\182\009\182\009\182\009\182\009\182\009\182\009\
    \182\009\182\009\182\009\182\009\182\009\182\009\182\009\182\009\
    \182\009\182\009\182\009\182\009\182\009\182\009\182\009\255\255\
    \182\009\182\009\182\009\182\009\182\009\182\009\182\009\182\009\
    \182\009\182\009\182\009\182\009\182\009\182\009\182\009\182\009\
    \182\009\182\009\182\009\182\009\182\009\182\009\182\009\182\009\
    \182\009\182\009\182\009\182\009\182\009\182\009\182\009\255\255\
    \182\009\182\009\182\009\182\009\182\009\182\009\182\009\182\009\
    \184\009\255\255\187\009\255\255\255\255\251\009\255\255\184\009\
    \206\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\044\010\118\010\255\255\255\255\045\010\
    \255\255\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\255\255\255\255\049\010\255\255\
    \184\009\255\255\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\184\009\184\009\184\009\184\009\
    \184\009\184\009\184\009\184\009\255\255\194\009\194\009\255\255\
    \255\255\194\009\255\255\217\009\217\009\217\009\217\009\217\009\
    \217\009\217\009\217\009\217\009\217\009\217\009\217\009\217\009\
    \217\009\217\009\217\009\217\009\194\009\217\009\217\009\217\009\
    \217\009\217\009\217\009\217\009\217\009\217\009\217\009\217\009\
    \217\009\217\009\217\009\217\009\217\009\217\009\217\009\217\009\
    \217\009\217\009\217\009\217\009\217\009\217\009\217\009\217\009\
    \217\009\217\009\224\009\224\009\224\009\224\009\224\009\224\009\
    \224\009\224\009\224\009\224\009\224\009\224\009\224\009\224\009\
    \224\009\224\009\255\255\184\009\184\009\255\255\255\255\255\255\
    \255\255\050\010\255\255\185\009\050\010\255\255\255\255\184\009\
    \184\009\194\009\185\009\255\255\185\009\185\009\185\009\185\009\
    \185\009\185\009\185\009\185\009\185\009\185\009\255\255\255\255\
    \255\255\255\255\255\255\184\009\185\009\185\009\185\009\185\009\
    \185\009\185\009\185\009\185\009\185\009\185\009\185\009\185\009\
    \185\009\185\009\185\009\185\009\185\009\185\009\185\009\185\009\
    \185\009\185\009\185\009\185\009\185\009\185\009\185\009\255\255\
    \255\255\255\255\255\255\185\009\255\255\185\009\185\009\185\009\
    \185\009\185\009\185\009\185\009\185\009\185\009\185\009\185\009\
    \185\009\185\009\185\009\185\009\185\009\185\009\185\009\185\009\
    \185\009\185\009\185\009\185\009\185\009\185\009\185\009\214\009\
    \214\009\214\009\214\009\214\009\214\009\214\009\214\009\214\009\
    \214\009\214\009\214\009\214\009\214\009\214\009\214\009\214\009\
    \214\009\214\009\214\009\214\009\214\009\214\009\214\009\214\009\
    \214\009\255\255\255\255\052\010\255\255\214\009\052\010\214\009\
    \214\009\214\009\214\009\214\009\214\009\214\009\214\009\214\009\
    \214\009\214\009\214\009\214\009\214\009\214\009\214\009\214\009\
    \214\009\214\009\214\009\214\009\214\009\214\009\214\009\214\009\
    \214\009\226\009\226\009\226\009\226\009\226\009\226\009\226\009\
    \226\009\226\009\226\009\226\009\226\009\255\255\185\009\185\009\
    \246\009\246\009\246\009\246\009\246\009\246\009\246\009\246\009\
    \246\009\246\009\185\009\185\009\239\009\239\009\239\009\239\009\
    \239\009\239\009\239\009\239\009\239\009\239\009\239\009\239\009\
    \239\009\239\009\239\009\239\009\255\255\255\255\185\009\241\009\
    \241\009\241\009\241\009\241\009\241\009\241\009\241\009\241\009\
    \241\009\241\009\241\009\245\009\245\009\245\009\245\009\245\009\
    \245\009\245\009\245\009\245\009\245\009\245\009\245\009\255\255\
    \214\009\214\009\250\009\250\009\250\009\250\009\250\009\250\009\
    \250\009\250\009\250\009\250\009\214\009\214\009\243\009\243\009\
    \243\009\243\009\243\009\243\009\243\009\243\009\243\009\243\009\
    \243\009\243\009\243\009\243\009\243\009\243\009\255\255\050\010\
    \214\009\216\009\216\009\216\009\216\009\216\009\216\009\216\009\
    \216\009\216\009\216\009\216\009\216\009\216\009\216\009\216\009\
    \216\009\216\009\216\009\216\009\216\009\216\009\216\009\216\009\
    \216\009\216\009\216\009\216\009\216\009\216\009\216\009\216\009\
    \216\009\216\009\216\009\216\009\216\009\216\009\216\009\216\009\
    \216\009\216\009\216\009\216\009\216\009\216\009\216\009\216\009\
    \218\009\218\009\218\009\218\009\218\009\218\009\218\009\218\009\
    \218\009\218\009\218\009\218\009\218\009\218\009\218\009\218\009\
    \218\009\218\009\218\009\218\009\218\009\218\009\218\009\255\255\
    \218\009\218\009\218\009\218\009\218\009\218\009\218\009\218\009\
    \218\009\218\009\218\009\218\009\218\009\218\009\218\009\218\009\
    \218\009\218\009\218\009\218\009\218\009\218\009\218\009\218\009\
    \218\009\218\009\218\009\218\009\218\009\218\009\218\009\255\255\
    \218\009\218\009\218\009\218\009\218\009\218\009\218\009\218\009\
    \220\009\052\010\255\255\255\255\255\255\255\255\255\255\220\009\
    \255\255\220\009\220\009\220\009\220\009\220\009\220\009\220\009\
    \220\009\220\009\220\009\255\255\255\255\255\255\255\255\255\255\
    \255\255\220\009\220\009\220\009\220\009\220\009\220\009\220\009\
    \220\009\220\009\220\009\220\009\220\009\220\009\220\009\220\009\
    \220\009\220\009\220\009\220\009\220\009\220\009\220\009\220\009\
    \220\009\220\009\220\009\220\009\255\255\255\255\255\255\255\255\
    \220\009\255\255\220\009\220\009\220\009\220\009\220\009\220\009\
    \220\009\220\009\220\009\220\009\220\009\220\009\220\009\220\009\
    \220\009\220\009\220\009\220\009\220\009\220\009\220\009\220\009\
    \220\009\220\009\220\009\220\009\225\009\225\009\225\009\225\009\
    \225\009\225\009\225\009\225\009\225\009\225\009\225\009\225\009\
    \225\009\225\009\225\009\225\009\225\009\225\009\225\009\225\009\
    \225\009\225\009\225\009\225\009\225\009\225\009\225\009\225\009\
    \225\009\225\009\225\009\225\009\225\009\225\009\225\009\225\009\
    \225\009\225\009\225\009\225\009\225\009\225\009\225\009\225\009\
    \225\009\225\009\225\009\225\009\225\009\225\009\225\009\225\009\
    \225\009\225\009\225\009\225\009\225\009\225\009\225\009\225\009\
    \225\009\225\009\225\009\225\009\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\220\009\220\009\255\255\255\255\255\255\
    \255\255\255\255\255\255\221\009\255\255\255\255\255\255\220\009\
    \220\009\255\255\221\009\255\255\221\009\221\009\221\009\221\009\
    \221\009\221\009\221\009\221\009\221\009\221\009\255\255\255\255\
    \255\255\255\255\255\255\220\009\221\009\221\009\221\009\221\009\
    \221\009\221\009\221\009\221\009\221\009\221\009\221\009\221\009\
    \221\009\221\009\221\009\221\009\221\009\221\009\221\009\221\009\
    \221\009\221\009\221\009\221\009\221\009\221\009\221\009\255\255\
    \255\255\255\255\255\255\221\009\255\255\221\009\221\009\221\009\
    \221\009\221\009\221\009\221\009\221\009\221\009\221\009\221\009\
    \221\009\221\009\221\009\221\009\221\009\221\009\221\009\221\009\
    \221\009\221\009\221\009\221\009\221\009\221\009\221\009\229\009\
    \229\009\229\009\229\009\229\009\229\009\229\009\229\009\229\009\
    \229\009\229\009\229\009\229\009\229\009\229\009\229\009\229\009\
    \229\009\229\009\229\009\229\009\229\009\229\009\229\009\229\009\
    \229\009\229\009\229\009\229\009\229\009\229\009\229\009\229\009\
    \229\009\229\009\229\009\229\009\229\009\229\009\229\009\229\009\
    \229\009\229\009\229\009\229\009\229\009\229\009\247\009\247\009\
    \247\009\247\009\247\009\247\009\247\009\247\009\247\009\247\009\
    \247\009\247\009\247\009\247\009\247\009\247\009\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\221\009\221\009\
    \255\255\255\255\255\255\255\255\255\255\255\255\222\009\255\255\
    \255\255\255\255\221\009\221\009\255\255\222\009\255\255\222\009\
    \222\009\222\009\222\009\222\009\222\009\222\009\222\009\222\009\
    \222\009\255\255\255\255\255\255\255\255\255\255\221\009\222\009\
    \222\009\222\009\222\009\222\009\222\009\222\009\222\009\222\009\
    \222\009\222\009\222\009\222\009\222\009\222\009\222\009\222\009\
    \222\009\222\009\222\009\222\009\222\009\222\009\222\009\222\009\
    \222\009\222\009\255\255\255\255\255\255\255\255\222\009\255\255\
    \222\009\222\009\222\009\222\009\222\009\222\009\222\009\222\009\
    \222\009\222\009\222\009\222\009\222\009\222\009\222\009\222\009\
    \222\009\222\009\222\009\222\009\222\009\222\009\222\009\222\009\
    \222\009\222\009\230\009\230\009\230\009\230\009\230\009\230\009\
    \230\009\230\009\230\009\230\009\230\009\230\009\230\009\230\009\
    \230\009\230\009\230\009\255\255\230\009\230\009\230\009\230\009\
    \230\009\230\009\230\009\230\009\230\009\230\009\230\009\230\009\
    \230\009\230\009\230\009\230\009\230\009\230\009\230\009\230\009\
    \230\009\230\009\230\009\230\009\230\009\230\009\230\009\230\009\
    \230\009\249\009\249\009\249\009\249\009\249\009\249\009\249\009\
    \249\009\249\009\249\009\249\009\249\009\090\010\090\010\090\010\
    \090\010\090\010\090\010\090\010\090\010\090\010\090\010\090\010\
    \090\010\222\009\222\009\255\255\255\255\255\255\255\255\255\255\
    \255\255\223\009\255\255\255\255\255\255\222\009\222\009\255\255\
    \223\009\255\255\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\255\255\255\255\255\255\255\255\
    \255\255\222\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\055\010\255\255\057\010\
    \055\010\223\009\057\010\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\231\009\231\009\231\009\
    \231\009\231\009\231\009\231\009\231\009\231\009\231\009\231\009\
    \231\009\231\009\231\009\231\009\231\009\231\009\231\009\231\009\
    \231\009\231\009\231\009\231\009\255\255\231\009\231\009\231\009\
    \231\009\231\009\231\009\231\009\231\009\231\009\231\009\231\009\
    \231\009\231\009\231\009\231\009\231\009\231\009\231\009\231\009\
    \231\009\231\009\231\009\231\009\231\009\231\009\231\009\231\009\
    \231\009\231\009\231\009\231\009\255\255\231\009\231\009\231\009\
    \231\009\231\009\231\009\231\009\231\009\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\223\009\223\009\058\010\255\255\
    \255\255\058\010\255\255\255\255\233\009\255\255\255\255\255\255\
    \223\009\223\009\255\255\233\009\255\255\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\255\255\
    \255\255\255\255\255\255\255\255\223\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \255\255\255\255\255\255\255\255\233\009\255\255\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \233\009\233\009\233\009\233\009\233\009\233\009\233\009\233\009\
    \235\009\235\009\235\009\235\009\235\009\235\009\235\009\235\009\
    \235\009\235\009\235\009\235\009\235\009\235\009\235\009\235\009\
    \235\009\235\009\235\009\235\009\235\009\235\009\235\009\235\009\
    \235\009\235\009\235\009\235\009\235\009\235\009\235\009\235\009\
    \235\009\235\009\235\009\235\009\235\009\235\009\235\009\235\009\
    \235\009\235\009\235\009\235\009\235\009\235\009\235\009\255\255\
    \255\255\255\255\255\255\055\010\255\255\057\010\088\010\088\010\
    \088\010\088\010\088\010\088\010\088\010\088\010\088\010\088\010\
    \088\010\088\010\088\010\088\010\088\010\088\010\255\255\233\009\
    \233\009\091\010\091\010\091\010\091\010\091\010\091\010\091\010\
    \091\010\091\010\091\010\233\009\233\009\255\255\236\009\236\009\
    \236\009\236\009\236\009\236\009\236\009\236\009\236\009\236\009\
    \236\009\236\009\236\009\236\009\236\009\236\009\236\009\233\009\
    \236\009\236\009\236\009\236\009\236\009\236\009\236\009\236\009\
    \236\009\236\009\236\009\236\009\236\009\236\009\236\009\236\009\
    \236\009\236\009\236\009\236\009\236\009\236\009\236\009\236\009\
    \236\009\236\009\236\009\236\009\236\009\237\009\237\009\237\009\
    \237\009\237\009\237\009\237\009\237\009\237\009\237\009\237\009\
    \237\009\237\009\237\009\237\009\237\009\237\009\237\009\237\009\
    \237\009\237\009\237\009\237\009\058\010\237\009\237\009\237\009\
    \237\009\237\009\237\009\237\009\237\009\237\009\237\009\237\009\
    \237\009\237\009\237\009\237\009\237\009\237\009\237\009\237\009\
    \237\009\237\009\237\009\237\009\237\009\237\009\237\009\237\009\
    \237\009\237\009\237\009\237\009\255\255\237\009\237\009\237\009\
    \237\009\237\009\237\009\237\009\237\009\240\009\240\009\240\009\
    \240\009\240\009\240\009\240\009\240\009\240\009\240\009\240\009\
    \240\009\240\009\240\009\240\009\240\009\240\009\240\009\240\009\
    \240\009\240\009\240\009\240\009\240\009\240\009\240\009\240\009\
    \240\009\240\009\240\009\240\009\240\009\240\009\240\009\240\009\
    \240\009\240\009\240\009\240\009\240\009\240\009\240\009\240\009\
    \240\009\240\009\240\009\240\009\240\009\240\009\240\009\240\009\
    \240\009\240\009\240\009\240\009\240\009\240\009\240\009\240\009\
    \240\009\240\009\240\009\240\009\240\009\244\009\244\009\244\009\
    \244\009\244\009\244\009\244\009\244\009\244\009\244\009\244\009\
    \244\009\244\009\244\009\244\009\244\009\244\009\244\009\244\009\
    \244\009\244\009\244\009\244\009\244\009\244\009\244\009\244\009\
    \244\009\244\009\244\009\244\009\244\009\244\009\244\009\244\009\
    \244\009\244\009\244\009\244\009\244\009\244\009\244\009\244\009\
    \244\009\244\009\244\009\244\009\244\009\244\009\244\009\244\009\
    \244\009\244\009\244\009\244\009\244\009\244\009\244\009\244\009\
    \244\009\244\009\244\009\244\009\244\009\248\009\248\009\248\009\
    \248\009\248\009\248\009\248\009\248\009\248\009\248\009\248\009\
    \248\009\248\009\248\009\248\009\248\009\248\009\248\009\248\009\
    \248\009\248\009\248\009\248\009\248\009\248\009\248\009\248\009\
    \248\009\248\009\248\009\248\009\248\009\248\009\248\009\248\009\
    \248\009\248\009\248\009\248\009\248\009\248\009\248\009\248\009\
    \248\009\248\009\248\009\248\009\248\009\248\009\248\009\248\009\
    \248\009\248\009\248\009\248\009\248\009\248\009\248\009\248\009\
    \248\009\248\009\248\009\248\009\248\009\029\010\029\010\041\010\
    \041\010\029\010\046\010\041\010\047\010\046\010\255\255\047\010\
    \051\010\051\010\053\010\060\010\051\010\053\010\060\010\056\010\
    \056\010\255\255\061\010\056\010\029\010\061\010\041\010\062\010\
    \255\255\255\255\062\010\255\255\255\255\255\255\255\255\051\010\
    \255\255\255\255\255\255\041\010\255\255\046\010\056\010\047\010\
    \103\010\103\010\103\010\103\010\103\010\103\010\103\010\103\010\
    \103\010\103\010\103\010\103\010\103\010\103\010\103\010\103\010\
    \105\010\105\010\105\010\105\010\105\010\105\010\105\010\105\010\
    \105\010\105\010\105\010\105\010\106\010\106\010\106\010\106\010\
    \106\010\106\010\106\010\106\010\106\010\106\010\255\255\255\255\
    \255\255\029\010\110\010\110\010\110\010\110\010\110\010\110\010\
    \110\010\110\010\110\010\110\010\051\010\053\010\060\010\255\255\
    \255\255\255\255\255\255\056\010\071\010\071\010\071\010\071\010\
    \071\010\071\010\071\010\071\010\071\010\071\010\071\010\071\010\
    \071\010\071\010\071\010\071\010\071\010\071\010\071\010\071\010\
    \071\010\071\010\071\010\071\010\071\010\071\010\255\255\255\255\
    \255\255\255\255\071\010\255\255\071\010\071\010\071\010\071\010\
    \071\010\071\010\071\010\071\010\071\010\071\010\071\010\071\010\
    \071\010\071\010\071\010\071\010\071\010\071\010\071\010\071\010\
    \071\010\071\010\071\010\071\010\071\010\071\010\074\010\074\010\
    \074\010\074\010\074\010\074\010\074\010\074\010\074\010\074\010\
    \074\010\074\010\074\010\074\010\074\010\074\010\074\010\074\010\
    \074\010\074\010\074\010\074\010\074\010\074\010\074\010\074\010\
    \074\010\074\010\074\010\074\010\074\010\074\010\074\010\074\010\
    \074\010\074\010\074\010\074\010\074\010\074\010\074\010\074\010\
    \074\010\074\010\074\010\074\010\074\010\107\010\107\010\107\010\
    \107\010\107\010\107\010\107\010\107\010\107\010\107\010\107\010\
    \107\010\107\010\107\010\107\010\107\010\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\071\010\071\010\114\010\
    \114\010\114\010\114\010\114\010\114\010\114\010\114\010\114\010\
    \114\010\071\010\071\010\255\255\029\010\255\255\041\010\255\255\
    \046\010\255\255\047\010\255\255\255\255\255\255\255\255\051\010\
    \053\010\060\010\255\255\255\255\255\255\071\010\056\010\255\255\
    \061\010\255\255\255\255\255\255\255\255\062\010\255\255\255\255\
    \255\255\255\255\075\010\075\010\075\010\075\010\075\010\075\010\
    \075\010\075\010\075\010\075\010\075\010\075\010\075\010\075\010\
    \075\010\075\010\075\010\071\010\075\010\075\010\075\010\075\010\
    \075\010\075\010\075\010\075\010\075\010\075\010\075\010\075\010\
    \075\010\075\010\075\010\075\010\075\010\075\010\075\010\075\010\
    \075\010\075\010\075\010\075\010\075\010\075\010\075\010\075\010\
    \075\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \255\255\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \078\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \076\010\109\010\109\010\109\010\109\010\109\010\109\010\109\010\
    \109\010\109\010\109\010\109\010\109\010\255\255\255\255\078\010\
    \255\255\255\255\255\255\255\255\255\255\255\255\078\010\255\255\
    \255\255\255\255\255\255\255\255\255\255\078\010\255\255\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\255\255\255\255\255\255\255\255\255\255\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\255\255\255\255\255\255\255\255\078\010\255\255\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\081\010\081\010\081\010\081\010\081\010\081\010\
    \081\010\081\010\081\010\081\010\081\010\081\010\081\010\081\010\
    \081\010\081\010\081\010\081\010\081\010\081\010\081\010\081\010\
    \081\010\081\010\081\010\081\010\255\255\255\255\255\255\255\255\
    \081\010\255\255\081\010\081\010\081\010\081\010\081\010\081\010\
    \081\010\081\010\081\010\081\010\081\010\081\010\081\010\081\010\
    \081\010\081\010\081\010\081\010\081\010\081\010\081\010\081\010\
    \081\010\081\010\081\010\081\010\113\010\113\010\113\010\113\010\
    \113\010\113\010\113\010\113\010\113\010\113\010\113\010\113\010\
    \255\255\078\010\078\010\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\078\010\078\010\111\010\
    \111\010\111\010\111\010\111\010\111\010\111\010\111\010\111\010\
    \111\010\111\010\111\010\111\010\111\010\111\010\111\010\255\255\
    \255\255\078\010\255\255\255\255\255\255\255\255\082\010\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\081\010\081\010\082\010\255\255\255\255\
    \255\255\255\255\255\255\255\255\082\010\255\255\255\255\081\010\
    \081\010\255\255\255\255\082\010\255\255\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \255\255\255\255\255\255\081\010\255\255\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \255\255\255\255\255\255\255\255\082\010\255\255\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \084\010\084\010\084\010\084\010\084\010\084\010\084\010\084\010\
    \084\010\084\010\084\010\084\010\084\010\084\010\084\010\084\010\
    \084\010\084\010\084\010\084\010\084\010\084\010\084\010\084\010\
    \084\010\084\010\084\010\084\010\084\010\084\010\084\010\084\010\
    \084\010\084\010\084\010\084\010\084\010\084\010\084\010\084\010\
    \084\010\084\010\084\010\084\010\084\010\084\010\084\010\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\082\010\
    \082\010\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\082\010\082\010\255\255\085\010\085\010\
    \085\010\085\010\085\010\085\010\085\010\085\010\085\010\085\010\
    \085\010\085\010\085\010\085\010\085\010\085\010\085\010\082\010\
    \085\010\085\010\085\010\085\010\085\010\085\010\085\010\085\010\
    \085\010\085\010\085\010\085\010\085\010\085\010\085\010\085\010\
    \085\010\085\010\085\010\085\010\085\010\085\010\085\010\085\010\
    \085\010\085\010\085\010\085\010\085\010\086\010\086\010\086\010\
    \086\010\086\010\086\010\086\010\086\010\086\010\086\010\086\010\
    \086\010\086\010\086\010\086\010\086\010\086\010\086\010\086\010\
    \086\010\086\010\086\010\086\010\255\255\086\010\086\010\086\010\
    \086\010\086\010\086\010\086\010\086\010\086\010\086\010\086\010\
    \086\010\086\010\086\010\086\010\086\010\086\010\086\010\086\010\
    \086\010\086\010\086\010\086\010\086\010\086\010\086\010\086\010\
    \086\010\086\010\086\010\086\010\255\255\086\010\086\010\086\010\
    \086\010\086\010\086\010\086\010\086\010\089\010\089\010\089\010\
    \089\010\089\010\089\010\089\010\089\010\089\010\089\010\089\010\
    \089\010\089\010\089\010\089\010\089\010\089\010\089\010\089\010\
    \089\010\089\010\089\010\089\010\089\010\089\010\089\010\089\010\
    \089\010\089\010\089\010\089\010\089\010\089\010\089\010\089\010\
    \089\010\089\010\089\010\089\010\089\010\089\010\089\010\089\010\
    \089\010\089\010\089\010\089\010\089\010\089\010\089\010\089\010\
    \089\010\089\010\089\010\089\010\089\010\089\010\089\010\089\010\
    \089\010\089\010\089\010\089\010\089\010\093\010\093\010\093\010\
    \093\010\093\010\093\010\093\010\093\010\093\010\093\010\093\010\
    \093\010\093\010\093\010\093\010\093\010\093\010\093\010\093\010\
    \093\010\093\010\093\010\093\010\093\010\093\010\093\010\093\010\
    \093\010\093\010\093\010\093\010\093\010\093\010\093\010\093\010\
    \093\010\093\010\093\010\093\010\093\010\093\010\093\010\093\010\
    \093\010\093\010\093\010\093\010\094\010\094\010\094\010\094\010\
    \094\010\094\010\094\010\094\010\094\010\094\010\094\010\094\010\
    \094\010\094\010\094\010\094\010\094\010\255\255\094\010\094\010\
    \094\010\094\010\094\010\094\010\094\010\094\010\094\010\094\010\
    \094\010\094\010\094\010\094\010\094\010\094\010\094\010\094\010\
    \094\010\094\010\094\010\094\010\094\010\094\010\094\010\094\010\
    \094\010\094\010\094\010\095\010\095\010\095\010\095\010\095\010\
    \095\010\095\010\095\010\095\010\095\010\095\010\095\010\095\010\
    \095\010\095\010\095\010\095\010\095\010\095\010\095\010\095\010\
    \095\010\095\010\255\255\095\010\095\010\095\010\095\010\095\010\
    \095\010\095\010\095\010\095\010\095\010\095\010\095\010\095\010\
    \095\010\095\010\095\010\095\010\095\010\095\010\095\010\095\010\
    \095\010\095\010\095\010\095\010\095\010\095\010\095\010\095\010\
    \095\010\095\010\097\010\095\010\095\010\095\010\095\010\095\010\
    \095\010\095\010\095\010\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\097\010\255\255\255\255\255\255\255\255\255\255\255\255\
    \097\010\255\255\255\255\255\255\255\255\255\255\255\255\097\010\
    \255\255\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\255\255\255\255\255\255\255\255\
    \255\255\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\255\255\255\255\255\255\255\255\
    \097\010\255\255\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\097\010\097\010\097\010\097\010\
    \097\010\097\010\097\010\097\010\099\010\099\010\099\010\099\010\
    \099\010\099\010\099\010\099\010\099\010\099\010\099\010\099\010\
    \099\010\099\010\099\010\099\010\099\010\099\010\099\010\099\010\
    \099\010\099\010\099\010\099\010\099\010\099\010\099\010\099\010\
    \099\010\099\010\099\010\099\010\099\010\099\010\099\010\099\010\
    \099\010\099\010\099\010\099\010\099\010\099\010\099\010\099\010\
    \099\010\099\010\099\010\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\097\010\097\010\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\097\010\
    \097\010\255\255\100\010\100\010\100\010\100\010\100\010\100\010\
    \100\010\100\010\100\010\100\010\100\010\100\010\100\010\100\010\
    \100\010\100\010\100\010\097\010\100\010\100\010\100\010\100\010\
    \100\010\100\010\100\010\100\010\100\010\100\010\100\010\100\010\
    \100\010\100\010\100\010\100\010\100\010\100\010\100\010\100\010\
    \100\010\100\010\100\010\100\010\100\010\100\010\100\010\100\010\
    \100\010\101\010\101\010\101\010\101\010\101\010\101\010\101\010\
    \101\010\101\010\101\010\101\010\101\010\101\010\101\010\101\010\
    \101\010\101\010\101\010\101\010\101\010\101\010\101\010\101\010\
    \255\255\101\010\101\010\101\010\101\010\101\010\101\010\101\010\
    \101\010\101\010\101\010\101\010\101\010\101\010\101\010\101\010\
    \101\010\101\010\101\010\101\010\101\010\101\010\101\010\101\010\
    \101\010\101\010\101\010\101\010\101\010\101\010\101\010\101\010\
    \255\255\101\010\101\010\101\010\101\010\101\010\101\010\101\010\
    \101\010\104\010\104\010\104\010\104\010\104\010\104\010\104\010\
    \104\010\104\010\104\010\104\010\104\010\104\010\104\010\104\010\
    \104\010\104\010\104\010\104\010\104\010\104\010\104\010\104\010\
    \104\010\104\010\104\010\104\010\104\010\104\010\104\010\104\010\
    \104\010\104\010\104\010\104\010\104\010\104\010\104\010\104\010\
    \104\010\104\010\104\010\104\010\104\010\104\010\104\010\104\010\
    \104\010\104\010\104\010\104\010\104\010\104\010\104\010\104\010\
    \104\010\104\010\104\010\104\010\104\010\104\010\104\010\104\010\
    \104\010\108\010\108\010\108\010\108\010\108\010\108\010\108\010\
    \108\010\108\010\108\010\108\010\108\010\108\010\108\010\108\010\
    \108\010\108\010\108\010\108\010\108\010\108\010\108\010\108\010\
    \108\010\108\010\108\010\108\010\108\010\108\010\108\010\108\010\
    \108\010\108\010\108\010\108\010\108\010\108\010\108\010\108\010\
    \108\010\108\010\108\010\108\010\108\010\108\010\108\010\108\010\
    \108\010\108\010\108\010\108\010\108\010\108\010\108\010\108\010\
    \108\010\108\010\108\010\108\010\108\010\108\010\108\010\108\010\
    \108\010\112\010\112\010\112\010\112\010\112\010\112\010\112\010\
    \112\010\112\010\112\010\112\010\112\010\112\010\112\010\112\010\
    \112\010\112\010\112\010\112\010\112\010\112\010\112\010\112\010\
    \112\010\112\010\112\010\112\010\112\010\112\010\112\010\112\010\
    \112\010\112\010\112\010\112\010\112\010\112\010\112\010\112\010\
    \112\010\112\010\112\010\112\010\112\010\112\010\112\010\112\010\
    \112\010\112\010\112\010\112\010\112\010\112\010\112\010\112\010\
    \112\010\112\010\112\010\112\010\112\010\112\010\112\010\112\010\
    \112\010\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255";
  Lexing.lex_base_code =
   "";
  Lexing.lex_backtrk_code =
   "";
  Lexing.lex_default_code =
   "";
  Lexing.lex_trans_code =
   "";
  Lexing.lex_check_code =
   "";
  Lexing.lex_code =
   "";
}

let rec coq_bol lexbuf =
   __ocaml_lex_coq_bol_rec lexbuf 0
and __ocaml_lex_coq_bol_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 446 "tools/coqdoc/cpretty.mll"
      ( if not (!in_proof <> None && (!Cdglobals.gallina || !Cdglobals.light))
        then Output.empty_line_of_code ();
        coq_bol lexbuf )
# 3450 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 450 "tools/coqdoc/cpretty.mll"
      ( Output.end_coq (); Output.start_doc ();
	let eol = doc_bol lexbuf in
	  Output.end_doc (); Output.start_coq ();
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3458 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 455 "tools/coqdoc/cpretty.mll"
      ( Output.end_coq (); Output.start_doc (); comments lexbuf; Output.end_doc ();
	Output.start_coq (); coq lexbuf )
# 3464 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 458 "tools/coqdoc/cpretty.mll"
      ( skip_hide lexbuf; coq_bol lexbuf )
# 3469 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 460 "tools/coqdoc/cpretty.mll"
      ( begin_show (); coq_bol lexbuf )
# 3474 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 462 "tools/coqdoc/cpretty.mll"
      ( end_show (); coq_bol lexbuf )
# 3479 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 464 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  if !Cdglobals.light && section_or_end s then
	    let eol = skip_to_dot lexbuf in
	      if eol then (coq_bol lexbuf) else coq lexbuf
	  else
	    begin
	      output_indented_keyword s lexbuf;
	      let eol = body lexbuf in
	      if eol then coq_bol lexbuf else coq lexbuf
	    end )
# 3493 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 475 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
        output_indented_keyword s lexbuf;
        let eol = body lexbuf in
	in_proof := Some eol;
	if eol then coq_bol lexbuf else coq lexbuf )
# 3502 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 481 "tools/coqdoc/cpretty.mll"
      ( in_proof := Some true;
	let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body_bol lexbuf end
	  else
	    let s = lexeme lexbuf in
	      if s.[String.length s - 1] = '.' then false
	      else skip_to_dot lexbuf
	in if eol then coq_bol lexbuf else coq lexbuf )
# 3515 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 490 "tools/coqdoc/cpretty.mll"
                  (
      let eol =
	if not (only_gallina ()) then
	  begin backtrack lexbuf; body_bol lexbuf end
	else skip_to_dot lexbuf
      in
	in_proof := None;
	if eol then coq_bol lexbuf else coq lexbuf )
# 3527 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 499 "tools/coqdoc/cpretty.mll"
      (
	in_proof := None;
	let s = lexeme lexbuf in
	output_indented_keyword s lexbuf;
	let eol= body lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3537 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 506 "tools/coqdoc/cpretty.mll"
      (
	in_proof := None;
	let s = lexeme lexbuf in
	output_indented_keyword s lexbuf;
	let eol= body lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3547 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 513 "tools/coqdoc/cpretty.mll"
      (	let s = lexeme lexbuf in
	output_indented_keyword s lexbuf;
	let eol= start_notation_string lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3555 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 519 "tools/coqdoc/cpretty.mll"
      ( let tok = lexeme lexbuf in
	let s = printing_token_body lexbuf in
	  add_printing_token tok s;
	  coq_bol lexbuf )
# 3563 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 524 "tools/coqdoc/cpretty.mll"
      ( eprintf "warning: bad 'printing' command at character %d\n"
	  (lexeme_start lexbuf); flush stderr;
	comment_level := 1;
	ignore (comment lexbuf);
	coq_bol lexbuf )
# 3572 "tools/coqdoc/cpretty.ml"

  | 15 ->
# 531 "tools/coqdoc/cpretty.mll"
      ( remove_printing_token (lexeme lexbuf);
	coq_bol lexbuf )
# 3578 "tools/coqdoc/cpretty.ml"

  | 16 ->
# 534 "tools/coqdoc/cpretty.mll"
      ( eprintf "warning: bad 'remove printing' command at character %d\n"
	  (lexeme_start lexbuf); flush stderr;
	comment_level := 1;
	ignore (comment lexbuf);
	coq_bol lexbuf )
# 3587 "tools/coqdoc/cpretty.ml"

  | 17 ->
# 540 "tools/coqdoc/cpretty.mll"
      ( comment_level := 1;
        let eol =
          if parse_comments () then begin
            let s = lexeme lexbuf in
            let nbsp, isp = count_spaces s in
            Output.indentation nbsp;
            Output.start_comment ();
            comment lexbuf
          end else skipped_comment lexbuf in
        if eol then coq_bol lexbuf else coq lexbuf )
# 3601 "tools/coqdoc/cpretty.ml"

  | 18 ->
# 550 "tools/coqdoc/cpretty.mll"
                (
    let eol = begin backtrack lexbuf; body_bol lexbuf end
    in if eol then coq_bol lexbuf else coq lexbuf )
# 3608 "tools/coqdoc/cpretty.ml"

  | 19 ->
# 554 "tools/coqdoc/cpretty.mll"
      ( () )
# 3613 "tools/coqdoc/cpretty.ml"

  | 20 ->
# 556 "tools/coqdoc/cpretty.mll"
      ( let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body_bol lexbuf end
	  else
	    skip_to_dot_or_brace lexbuf
	in
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3624 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_coq_bol_rec lexbuf __ocaml_lex_state

and coq lexbuf =
   __ocaml_lex_coq_rec lexbuf 1165
and __ocaml_lex_coq_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 568 "tools/coqdoc/cpretty.mll"
      ( if not (only_gallina ()) then Output.line_break(); coq_bol lexbuf )
# 3636 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 570 "tools/coqdoc/cpretty.mll"
      ( Output.end_coq (); Output.start_doc ();
	let eol = doc_bol lexbuf in
	  Output.end_doc (); Output.start_coq ();
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3644 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 575 "tools/coqdoc/cpretty.mll"
      ( comment_level := 1;
        let eol =
          if parse_comments () then begin
            Output.start_comment ();
            comment lexbuf
          end else skipped_comment lexbuf in
        if eol then coq_bol lexbuf else coq lexbuf )
# 3655 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 583 "tools/coqdoc/cpretty.mll"
      ( if not !formatted then
	begin
	  (* Isn't this an anomaly *)
	  let s = lexeme lexbuf in
	  let nlsp,s = remove_newline s in
	  let nbsp,isp = count_spaces s in
	  Output.indentation nbsp;
	  let loc = lexeme_start lexbuf + isp + nlsp in
	  Output.sublexer ']' loc;
	  Output.sublexer ']' (loc+1);
	  coq lexbuf
	end )
# 3671 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 596 "tools/coqdoc/cpretty.mll"
      ( () )
# 3676 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 598 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  if !Cdglobals.light && section_or_end s then
	    begin
	      let eol = skip_to_dot lexbuf in
		if eol then coq_bol lexbuf else coq lexbuf
	    end
	  else
	    begin
	      Output.ident s None;
	      let eol=body lexbuf in
		if eol then coq_bol lexbuf else coq lexbuf
	    end )
# 3692 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 611 "tools/coqdoc/cpretty.mll"
      ( let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body lexbuf end
	  else
	    let s = lexeme lexbuf in
	    let eol =
	      if s.[String.length s - 1] = '.' then false
	      else skip_to_dot lexbuf
	    in
	      eol
	in if eol then coq_bol lexbuf else coq lexbuf )
# 3707 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 622 "tools/coqdoc/cpretty.mll"
           (
      let eol =
	if not !Cdglobals.gallina then
	  begin backtrack lexbuf; body lexbuf end
	else
	  let eol = skip_to_dot lexbuf in
	    if !in_proof <> Some true && eol then
	      Output.line_break ();
	    eol
      in
	in_proof := None;
	if eol then coq_bol lexbuf else coq lexbuf )
# 3723 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 635 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  Output.ident s None;
	let eol = body lexbuf in
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3731 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 640 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	Output.ident s None;
	let eol= start_notation_string lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3739 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 645 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  Output.ident s None;
	let eol = body lexbuf in
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3747 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 650 "tools/coqdoc/cpretty.mll"
      ( ignore(lexeme lexbuf);
        Output.char '#'; Output.char '[';
        let eol = body lexbuf in
          if eol then coq_bol lexbuf else coq lexbuf )
# 3755 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 654 "tools/coqdoc/cpretty.mll"
           ( Output.char ' '; coq lexbuf )
# 3760 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 656 "tools/coqdoc/cpretty.mll"
      ( () )
# 3765 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 657 "tools/coqdoc/cpretty.mll"
      (	let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body lexbuf end
	  else
	    skip_to_dot_or_brace lexbuf
	in
	  if eol then coq_bol lexbuf else coq lexbuf)
# 3776 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_coq_rec lexbuf __ocaml_lex_state

and doc_bol lexbuf =
   __ocaml_lex_doc_bol_rec lexbuf 2233
and __ocaml_lex_doc_bol_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 669 "tools/coqdoc/cpretty.mll"
      ( let eol, lex = strip_eol (lexeme lexbuf) in
        let lev, s = sec_title lex in
          if (!Cdglobals.lib_subtitles) &&
             (subtitle (Output.get_module false) s) then
            ()
          else
            Output.section lev (fun () -> ignore (doc None (from_string s)));
	    if eol then doc_bol lexbuf else doc None lexbuf )
# 3795 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 678 "tools/coqdoc/cpretty.mll"
      ( let buf' = lexeme lexbuf in
        let bufs = Str.split_delim (Str.regexp "['\n']") buf' in
        let lines = (List.length bufs) - 1 in
        let line = 
          match bufs with
          | [] -> eprintf "Internal error bad_split1 - please report\n";
                  exit 1
          | _ -> List.nth bufs lines
        in
          match check_start_list line with
          | Neither -> backtrack_past_newline lexbuf; doc None lexbuf
          | List n -> if lines > 0 then Output.paragraph ();
                      Output.item 1; doc (Some [n]) lexbuf
          | Rule -> Output.rule (); doc None lexbuf
      )
# 3814 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 694 "tools/coqdoc/cpretty.mll"
      ( Output.paragraph (); doc_bol lexbuf )
# 3819 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 696 "tools/coqdoc/cpretty.mll"
      ( Output.start_verbatim false; verbatim 0 false lexbuf; doc_bol lexbuf )
# 3824 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 698 "tools/coqdoc/cpretty.mll"
      ( true )
# 3829 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 700 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '_' else start_emph ();
        doc None lexbuf )
# 3835 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 703 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf; doc None lexbuf )
# 3840 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_doc_bol_rec lexbuf __ocaml_lex_state

and doc_list_bol indents lexbuf =
   __ocaml_lex_doc_list_bol_rec indents lexbuf 2255
and __ocaml_lex_doc_list_bol_rec indents lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 708 "tools/coqdoc/cpretty.mll"
      ( let (n_spaces,_) = count_spaces (lexeme lexbuf) in
        match find_level indents n_spaces with
        | Before -> backtrack lexbuf; doc_bol lexbuf
        | StartLevel n -> Output.item n; doc (Some (take n indents)) lexbuf
        | InLevel (n,true) ->
            let items = List.length indents in
              Output.item (items+1);
              doc (Some (List.append indents [n_spaces])) lexbuf
        | InLevel (_,false) ->
            backtrack lexbuf; doc_bol lexbuf
      )
# 3862 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 720 "tools/coqdoc/cpretty.mll"
      ( Output.start_verbatim false;
        verbatim 0 false lexbuf;
        doc_list_bol indents lexbuf )
# 3869 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 724 "tools/coqdoc/cpretty.mll"
      ( formatted := true;
        Output.start_inline_coq_block ();
        ignore(body_bol lexbuf);
        Output.end_inline_coq_block ();
        formatted := false;
        doc_list_bol indents lexbuf )
# 3879 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 731 "tools/coqdoc/cpretty.mll"
      ( inf_rules (Some indents) lexbuf )
# 3884 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 733 "tools/coqdoc/cpretty.mll"
      ( (* Like in the doc_bol production, these two productions
           exist only to deal properly with whitespace *)
        Output.paragraph ();
        backtrack_past_newline lexbuf;
        doc_list_bol indents lexbuf )
# 3893 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 739 "tools/coqdoc/cpretty.mll"
      ( let buf' = lexeme lexbuf in
        let buf =
          let bufs = Str.split_delim (Str.regexp "['\n']") buf' in
            match bufs with
              | (_ :: s :: []) -> s
              | (_ :: _ :: s :: _) -> s
              | _ -> eprintf "Internal error bad_split2 - please report\n";
                     exit 1
        in
        let (n_spaces,_) = count_spaces buf in
        match find_level indents n_spaces with
        | StartLevel 1 | Before ->
        (* Here we were at the beginning of a line, and it was blank.
           The next line started before any list items.  So: insert
           a paragraph for the empty line, rewind to whatever's just
           after the newline, then toss over to doc_bol for whatever
           comes next. *)
            Output.stop_item ();
            Output.paragraph ();
            backtrack_past_newline lexbuf;
            doc_bol lexbuf
        | StartLevel _ | InLevel _ ->
            Output.paragraph ();
            backtrack_past_newline lexbuf;
            doc_list_bol indents lexbuf

      )
# 3924 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 767 "tools/coqdoc/cpretty.mll"
      ( let (n_spaces,_) = count_spaces (lexeme lexbuf) in
        match find_level indents n_spaces with
        | Before -> Output.stop_item (); backtrack lexbuf;
                    doc_bol lexbuf
        | StartLevel n ->
            Output.reach_item_level (n-1);
            backtrack lexbuf;
            doc (Some (take (n-1) indents)) lexbuf
        | InLevel (n,_) ->
            Output.reach_item_level n;
            backtrack lexbuf;
            doc (Some (take n indents)) lexbuf
      )
# 3941 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_doc_list_bol_rec indents lexbuf __ocaml_lex_state

and doc indents lexbuf =
   __ocaml_lex_doc_rec indents lexbuf 2274
and __ocaml_lex_doc_rec indents lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 784 "tools/coqdoc/cpretty.mll"
      ( Output.char '\n';
        match indents with
        | Some ls -> doc_list_bol ls lexbuf
        | None -> doc_bol lexbuf )
# 3956 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 789 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments
        then (Output.char '['; Output.char '['; doc indents lexbuf)
        else (formatted := true;
              Output.start_inline_coq_block ();
	      let eol = body_bol lexbuf in
	        Output.end_inline_coq_block (); formatted := false;
	        if eol then
		  match indents with
		  | Some ls -> doc_list_bol ls lexbuf
		  | None -> doc_bol lexbuf
		else doc indents lexbuf))
# 3971 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 801 "tools/coqdoc/cpretty.mll"
      ( inf_rules indents lexbuf )
# 3976 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 803 "tools/coqdoc/cpretty.mll"
      ( Output.proofbox (); doc indents lexbuf )
# 3981 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 804 "tools/coqdoc/cpretty.mll"
         ( url lexbuf; doc indents lexbuf )
# 3986 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 806 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '['
        else (brackets := 1;  Output.start_inline_coq (); escaped_coq lexbuf;
              Output.end_inline_coq ()); doc indents lexbuf )
# 3993 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 810 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf ;
        let bol_parse = match indents with
                        | Some is -> doc_list_bol is
                        | None   -> doc_bol
        in
        let eol =
          if !Cdglobals.parse_comments then comment lexbuf
          else skipped_comment lexbuf in
        if eol then bol_parse lexbuf else doc indents lexbuf )
# 4006 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 820 "tools/coqdoc/cpretty.mll"
      ((match indents with
        | Some _ -> Output.stop_item ()
        | None -> ());
       (* this says - if there is a blank line between the two comments,
          insert one in the output too *)
       let lines = List.length (Str.split_delim (Str.regexp "['\n']") 
                                                (lexeme lexbuf))
       in 
         if lines > 2 then Output.paragraph ();
       doc_bol lexbuf
      )
# 4021 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 832 "tools/coqdoc/cpretty.mll"
      ( true )
# 4026 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 834 "tools/coqdoc/cpretty.mll"
      ( false )
# 4031 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 836 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '$'
        else (Output.start_latex_math (); escaped_math_latex lexbuf);
        doc indents lexbuf )
# 4038 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 840 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '$';
        Output.char '$'; doc indents lexbuf )
# 4044 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 843 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '%'
        else escaped_latex lexbuf; doc indents lexbuf )
# 4050 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 846 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '%';
        Output.char '%'; doc indents lexbuf )
# 4056 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 849 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '#'
        else escaped_html lexbuf; doc indents lexbuf )
# 4062 "tools/coqdoc/cpretty.ml"

  | 15 ->
# 852 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '#';
        Output.char '#'; doc indents lexbuf )
# 4068 "tools/coqdoc/cpretty.ml"

  | 16 ->
# 855 "tools/coqdoc/cpretty.mll"
      ( List.iter (fun x -> Output.char (lexeme_char lexbuf x)) [0;1;2];
        doc indents lexbuf)
# 4074 "tools/coqdoc/cpretty.ml"

  | 17 ->
# 858 "tools/coqdoc/cpretty.mll"
      ( Output.char (lexeme_char lexbuf 0);
        if !Cdglobals.plain_comments then Output.char '_' else  start_emph () ;
        doc indents lexbuf )
# 4081 "tools/coqdoc/cpretty.ml"

  | 18 ->
# 862 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '_' else stop_emph () ;
        Output.char (lexeme_char lexbuf 1);
        doc indents lexbuf )
# 4088 "tools/coqdoc/cpretty.ml"

  | 19 ->
# 866 "tools/coqdoc/cpretty.mll"
      ( Output.start_verbatim true; verbatim 0 true lexbuf; doc_bol lexbuf )
# 4093 "tools/coqdoc/cpretty.ml"

  | 20 ->
# 868 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments 
	then Output.char '"' 
	else if in_quote ()
	then stop_quote () 
	else start_quote ();
        doc indents lexbuf )
# 4103 "tools/coqdoc/cpretty.ml"

  | 21 ->
# 875 "tools/coqdoc/cpretty.mll"
      ( false )
# 4108 "tools/coqdoc/cpretty.ml"

  | 22 ->
# 877 "tools/coqdoc/cpretty.mll"
      ( Output.char (lexeme_char lexbuf 0); doc indents lexbuf )
# 4113 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_doc_rec indents lexbuf __ocaml_lex_state

and escaped_math_latex lexbuf =
   __ocaml_lex_escaped_math_latex_rec lexbuf 2316
and __ocaml_lex_escaped_math_latex_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 882 "tools/coqdoc/cpretty.mll"
        ( Output.stop_latex_math () )
# 4125 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 883 "tools/coqdoc/cpretty.mll"
        ( Output.stop_latex_math () )
# 4130 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 885 "tools/coqdoc/cpretty.mll"
        ( Output.stop_latex_math (); backtrack lexbuf )
# 4135 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 886 "tools/coqdoc/cpretty.mll"
        ( Output.latex_char (lexeme_char lexbuf 0); escaped_math_latex lexbuf )
# 4140 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_escaped_math_latex_rec lexbuf __ocaml_lex_state

and escaped_latex lexbuf =
   __ocaml_lex_escaped_latex_rec lexbuf 2322
and __ocaml_lex_escaped_latex_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 889 "tools/coqdoc/cpretty.mll"
        ( () )
# 4152 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 890 "tools/coqdoc/cpretty.mll"
        ( () )
# 4157 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 892 "tools/coqdoc/cpretty.mll"
        ( backtrack lexbuf )
# 4162 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 893 "tools/coqdoc/cpretty.mll"
        ( Output.latex_char (lexeme_char lexbuf 0); escaped_latex lexbuf )
# 4167 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_escaped_latex_rec lexbuf __ocaml_lex_state

and escaped_html lexbuf =
   __ocaml_lex_escaped_html_rec lexbuf 2328
and __ocaml_lex_escaped_html_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 896 "tools/coqdoc/cpretty.mll"
        ( () )
# 4179 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 898 "tools/coqdoc/cpretty.mll"
        ( Output.html_char '&'; Output.html_char '#'; escaped_html lexbuf )
# 4184 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 900 "tools/coqdoc/cpretty.mll"
        ( Output.html_char '#'; escaped_html lexbuf )
# 4189 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 901 "tools/coqdoc/cpretty.mll"
        ( () )
# 4194 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 903 "tools/coqdoc/cpretty.mll"
        ( backtrack lexbuf )
# 4199 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 904 "tools/coqdoc/cpretty.mll"
        ( Output.html_char (lexeme_char lexbuf 0); escaped_html lexbuf )
# 4204 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_escaped_html_rec lexbuf __ocaml_lex_state

and verbatim depth inline lexbuf =
   __ocaml_lex_verbatim_rec depth inline lexbuf 2337
and __ocaml_lex_verbatim_rec depth inline lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 907 "tools/coqdoc/cpretty.mll"
                      ( Output.verbatim_char inline '\n'; Output.stop_verbatim inline )
# 4216 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 908 "tools/coqdoc/cpretty.mll"
            ( Output.verbatim_char inline '\n'; Output.stop_verbatim inline )
# 4221 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 909 "tools/coqdoc/cpretty.mll"
         ( Output.stop_verbatim inline )
# 4226 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 910 "tools/coqdoc/cpretty.mll"
         ( Output.verbatim_char inline '(';
           Output.verbatim_char inline '*';
           verbatim (depth+1) inline lexbuf )
# 4233 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 913 "tools/coqdoc/cpretty.mll"
         ( if (depth == 0)
           then (Output.stop_verbatim inline; backtrack lexbuf)
           else (Output.verbatim_char inline '*';
                 Output.verbatim_char inline ')';
                 verbatim (depth-1) inline lexbuf) )
# 4242 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 918 "tools/coqdoc/cpretty.mll"
        ( Output.stop_verbatim inline )
# 4247 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 919 "tools/coqdoc/cpretty.mll"
      ( Output.verbatim_char inline (lexeme_char lexbuf 0); verbatim depth inline lexbuf )
# 4252 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_verbatim_rec depth inline lexbuf __ocaml_lex_state

and url lexbuf =
   __ocaml_lex_url_rec lexbuf 2354
and __ocaml_lex_url_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 922 "tools/coqdoc/cpretty.mll"
         ( Output.url (Buffer.contents url_buffer) None; Buffer.clear url_buffer )
# 4264 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 923 "tools/coqdoc/cpretty.mll"
        ( url_name lexbuf )
# 4269 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 924 "tools/coqdoc/cpretty.mll"
      ( Buffer.add_char url_buffer (lexeme_char lexbuf 0); url lexbuf )
# 4274 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_url_rec lexbuf __ocaml_lex_state

and url_name lexbuf =
   __ocaml_lex_url_name_rec lexbuf 2358
and __ocaml_lex_url_name_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 927 "tools/coqdoc/cpretty.mll"
        ( Output.url (Buffer.contents url_buffer) (Some (Buffer.contents url_name_buffer));
	  Buffer.clear url_buffer; Buffer.clear url_name_buffer )
# 4287 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 929 "tools/coqdoc/cpretty.mll"
      ( Buffer.add_char url_name_buffer (lexeme_char lexbuf 0); url_name lexbuf )
# 4292 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_url_name_rec lexbuf __ocaml_lex_state

and escaped_coq lexbuf =
   __ocaml_lex_escaped_coq_rec lexbuf 2361
and __ocaml_lex_escaped_coq_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 935 "tools/coqdoc/cpretty.mll"
      ( decr brackets;
	if !brackets > 0 then
	  (Output.sublexer_in_doc ']'; escaped_coq lexbuf)
	else Tokens.flush_sublexer () )
# 4307 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 940 "tools/coqdoc/cpretty.mll"
      ( incr brackets;
        Output.sublexer_in_doc '['; escaped_coq lexbuf )
# 4313 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 943 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer (); comment_level := 1;
        ignore (if !Cdglobals.parse_comments then comment lexbuf
                else skipped_comment lexbuf);
        escaped_coq lexbuf )
# 4321 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 948 "tools/coqdoc/cpretty.mll"
      ( (* likely to be a syntax error: we escape *) backtrack lexbuf )
# 4326 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 950 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer () )
# 4331 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 952 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        Output.ident (lexeme lexbuf) None;
        escaped_coq lexbuf )
# 4338 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 956 "tools/coqdoc/cpretty.mll"
      ( let str = lexeme lexbuf in
          Tokens.flush_sublexer();
          (if !Cdglobals.inline_notmono then () 
                                        else Output.end_inline_coq ()); 
          String.iter Output.char str; 
          (if !Cdglobals.inline_notmono then () 
                                        else Output.start_inline_coq ());
          escaped_coq lexbuf )
# 4350 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 965 "tools/coqdoc/cpretty.mll"
      ( Output.sublexer_in_doc (lexeme_char lexbuf 0);
        escaped_coq lexbuf )
# 4356 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_escaped_coq_rec lexbuf __ocaml_lex_state

and comments lexbuf =
   __ocaml_lex_comments_rec lexbuf 2410
and __ocaml_lex_comments_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 972 "tools/coqdoc/cpretty.mll"
      ( Output.char ' '; comments lexbuf )
# 4368 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 974 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	let s = String.sub s 1 (String.length s - 2) in
	ignore (doc None (from_string s)); comments lexbuf )
# 4375 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 978 "tools/coqdoc/cpretty.mll"
      ( escaped_coq (from_string (lexeme lexbuf)); comments lexbuf )
# 4380 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 980 "tools/coqdoc/cpretty.mll"
      ( () )
# 4385 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 982 "tools/coqdoc/cpretty.mll"
      ( () )
# 4390 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 984 "tools/coqdoc/cpretty.mll"
      ( Output.char (lexeme_char lexbuf 0); comments lexbuf )
# 4395 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_comments_rec lexbuf __ocaml_lex_state

and skipped_comment lexbuf =
   __ocaml_lex_skipped_comment_rec lexbuf 2420
and __ocaml_lex_skipped_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 988 "tools/coqdoc/cpretty.mll"
      ( incr comment_level;
	skipped_comment lexbuf )
# 4408 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 991 "tools/coqdoc/cpretty.mll"
      ( decr comment_level;
        if !comment_level > 0 then skipped_comment lexbuf else true )
# 4414 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 994 "tools/coqdoc/cpretty.mll"
      ( decr comment_level;
        if !comment_level > 0 then skipped_comment lexbuf else false )
# 4420 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 996 "tools/coqdoc/cpretty.mll"
        ( false )
# 4425 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 997 "tools/coqdoc/cpretty.mll"
      ( skipped_comment lexbuf )
# 4430 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_skipped_comment_rec lexbuf __ocaml_lex_state

and comment lexbuf =
   __ocaml_lex_comment_rec lexbuf 2430
and __ocaml_lex_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1001 "tools/coqdoc/cpretty.mll"
      ( incr comment_level;
        Output.start_comment ();
        comment lexbuf )
# 4444 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1005 "tools/coqdoc/cpretty.mll"
      ( Output.end_comment ();
        Output.line_break ();
        decr comment_level;
        if !comment_level > 0 then comment lexbuf else true )
# 4452 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1010 "tools/coqdoc/cpretty.mll"
      ( Output.end_comment ();
        decr comment_level;
        if !comment_level > 0 then comment lexbuf else false )
# 4459 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 1014 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '['
        else (brackets := 1; Output.start_inline_coq ();
              escaped_coq lexbuf; Output.end_inline_coq ());
        comment lexbuf )
# 4467 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 1019 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then (Output.char '['; Output.char '[')
        else (formatted := true;
              Output.start_inline_coq_block ();
              let _ = body_bol lexbuf in
                Output.end_inline_coq_block (); formatted := false);
        comment lexbuf )
# 4477 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 1026 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '$'
        else (Output.start_latex_math (); escaped_math_latex lexbuf);
        comment lexbuf )
# 4484 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 1030 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '$';
        Output.char '$';
        comment lexbuf )
# 4491 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 1034 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '%'
        else escaped_latex lexbuf;
        comment lexbuf )
# 4498 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 1038 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '%';
        Output.char '%';
        comment lexbuf )
# 4505 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 1042 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '#'
        else escaped_html lexbuf;
        comment lexbuf )
# 4512 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 1046 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '#';
        Output.char '#';
        comment lexbuf )
# 4519 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 1049 "tools/coqdoc/cpretty.mll"
         ( false )
# 4524 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 1051 "tools/coqdoc/cpretty.mll"
      ( Output.indentation (fst (count_spaces (lexeme lexbuf)));
        comment lexbuf )
# 4530 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 1054 "tools/coqdoc/cpretty.mll"
      ( Output.line_break ();
        comment lexbuf )
# 4536 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 1056 "tools/coqdoc/cpretty.mll"
      ( Output.char (lexeme_char lexbuf 0);
        comment lexbuf )
# 4542 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_comment_rec lexbuf __ocaml_lex_state

and skip_to_dot lexbuf =
   __ocaml_lex_skip_to_dot_rec lexbuf 2453
and __ocaml_lex_skip_to_dot_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1060 "tools/coqdoc/cpretty.mll"
                  ( true )
# 4554 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1061 "tools/coqdoc/cpretty.mll"
                     ( false )
# 4559 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1063 "tools/coqdoc/cpretty.mll"
      ( comment_level := 1;
        ignore (skipped_comment lexbuf);
        skip_to_dot lexbuf )
# 4566 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 1066 "tools/coqdoc/cpretty.mll"
      ( skip_to_dot lexbuf )
# 4571 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_skip_to_dot_rec lexbuf __ocaml_lex_state

and skip_to_dot_or_brace lexbuf =
   __ocaml_lex_skip_to_dot_or_brace_rec lexbuf 2462
and __ocaml_lex_skip_to_dot_or_brace_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1069 "tools/coqdoc/cpretty.mll"
                  ( true )
# 4583 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1070 "tools/coqdoc/cpretty.mll"
                     ( false )
# 4588 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1072 "tools/coqdoc/cpretty.mll"
      ( comment_level := 1;
        ignore (skipped_comment lexbuf);
        skip_to_dot_or_brace lexbuf )
# 4595 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 1076 "tools/coqdoc/cpretty.mll"
      ( true )
# 4600 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 1078 "tools/coqdoc/cpretty.mll"
      ( false )
# 4605 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 1080 "tools/coqdoc/cpretty.mll"
      ( skip_to_dot_or_brace lexbuf )
# 4610 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 1081 "tools/coqdoc/cpretty.mll"
      ( skip_to_dot lexbuf )
# 4615 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_skip_to_dot_or_brace_rec lexbuf __ocaml_lex_state

and body_bol lexbuf =
   __ocaml_lex_body_bol_rec lexbuf 2476
and __ocaml_lex_body_bol_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1085 "tools/coqdoc/cpretty.mll"
      ( Output.indentation (fst (count_spaces (lexeme lexbuf))); body lexbuf )
# 4627 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1086 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf; Output.indentation 0; body lexbuf )
# 4632 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_body_bol_rec lexbuf __ocaml_lex_state

and body lexbuf =
   __ocaml_lex_body_rec lexbuf 2479
and __ocaml_lex_body_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1089 "tools/coqdoc/cpretty.mll"
       (Tokens.flush_sublexer(); Output.line_break(); Lexing.new_line lexbuf; body_bol lexbuf)
# 4644 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1091 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        if not !formatted then
          begin
            let s = lexeme lexbuf in
            let nlsp,s = remove_newline s in
            let _,isp = count_spaces s in
            let loc = lexeme_start lexbuf + nlsp + isp in
            Output.sublexer ']' loc;
            Output.sublexer ']' (loc+1);
            Tokens.flush_sublexer();
            body lexbuf
          end
        else
          begin
            Output.paragraph ();
            true
          end )
# 4665 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1109 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        if not !formatted then
          begin
	    let loc = lexeme_start lexbuf in
	    Output.sublexer ']' loc;
	    Output.sublexer ']' (loc+1);
	    Tokens.flush_sublexer();
	    Output.line_break();
            body lexbuf
          end
        else
          begin
            Output.paragraph ();
            true
          end )
# 4684 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 1124 "tools/coqdoc/cpretty.mll"
        ( Tokens.flush_sublexer(); false )
# 4689 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 1126 "tools/coqdoc/cpretty.mll"
 ( Tokens.flush_sublexer(); Output.char '.'; Output.line_break();
	  if not !formatted then true else body_bol lexbuf )
# 4695 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 1129 "tools/coqdoc/cpretty.mll"
 ( Tokens.flush_sublexer(); Output.char '.';
        if not !formatted then
          begin
            eprintf "Error: stray ]] at %d\n"  (lexeme_start lexbuf);
            flush stderr;
            exit 1
          end
	  else
          begin
            Output.paragraph ();
            true
          end
      )
# 4712 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 1143 "tools/coqdoc/cpretty.mll"
        ( Tokens.flush_sublexer(); Output.char '.'; Output.char ' ';
	  if not !formatted then false else body lexbuf )
# 4718 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 1146 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.end_coq (); Output.start_doc ();
	let eol = doc_bol lexbuf in
	  Output.end_doc (); Output.start_coq ();
	  if eol then body_bol lexbuf else body lexbuf )
# 4726 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 1151 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); comment_level := 1;
        let eol =
          if parse_comments () then begin
            Output.start_comment ();
            comment lexbuf
          end else begin
            let eol = skipped_comment lexbuf in
            if eol then Output.line_break();
            eol
          end in
	if eol then body_bol lexbuf else body lexbuf )
# 4741 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 1163 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        Output.ident (lexeme lexbuf) None;
	start_notation_string lexbuf )
# 4748 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 1167 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        Output.ident (lexeme lexbuf) (Some (lexeme_start lexbuf));
	body lexbuf )
# 4755 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 1171 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char '.'; Output.char '.';
        body lexbuf )
# 4761 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 1174 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char '"';
        string lexbuf;
        body lexbuf )
# 4768 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 1178 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char (lexeme_char lexbuf 0);
        body lexbuf )
# 4774 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 1181 "tools/coqdoc/cpretty.mll"
      ( let c = lexeme_char lexbuf 0 in
	Output.sublexer c (lexeme_start lexbuf);
        body lexbuf )
# 4781 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_body_rec lexbuf __ocaml_lex_state

and start_notation_string lexbuf =
   __ocaml_lex_start_notation_string_rec lexbuf 2555
and __ocaml_lex_start_notation_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1186 "tools/coqdoc/cpretty.mll"
          ( Tokens.flush_sublexer(); Output.char (lexeme_char lexbuf 0);
	    start_notation_string lexbuf )
# 4794 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1189 "tools/coqdoc/cpretty.mll"
      ( Output.sublexer '"' (lexeme_start lexbuf);
        notation_string lexbuf;
	body lexbuf )
# 4801 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1193 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf; body lexbuf )
# 4806 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_start_notation_string_rec lexbuf __ocaml_lex_state

and notation_string lexbuf =
   __ocaml_lex_notation_string_rec lexbuf 2559
and __ocaml_lex_notation_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1197 "tools/coqdoc/cpretty.mll"
      ( Output.char '"'; Output.char '"'; (* Unlikely! *)
        notation_string lexbuf )
# 4819 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1200 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char '"' )
# 4824 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1201 "tools/coqdoc/cpretty.mll"
      ( let c = lexeme_char lexbuf 0 in
        Output.sublexer c (lexeme_start lexbuf);
        notation_string lexbuf )
# 4831 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_notation_string_rec lexbuf __ocaml_lex_state

and string lexbuf =
   __ocaml_lex_string_rec lexbuf 2563
and __ocaml_lex_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1206 "tools/coqdoc/cpretty.mll"
           ( Output.char '"'; Output.char '"'; string lexbuf )
# 4843 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1207 "tools/coqdoc/cpretty.mll"
           ( Output.char '"' )
# 4848 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1208 "tools/coqdoc/cpretty.mll"
           ( let c = lexeme_char lexbuf 0 in Output.char c; string lexbuf )
# 4853 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_string_rec lexbuf __ocaml_lex_state

and skip_hide lexbuf =
   __ocaml_lex_skip_hide_rec lexbuf 2567
and __ocaml_lex_skip_hide_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1211 "tools/coqdoc/cpretty.mll"
                   ( () )
# 4865 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1212 "tools/coqdoc/cpretty.mll"
      ( skip_hide lexbuf )
# 4870 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_skip_hide_rec lexbuf __ocaml_lex_state

and printing_token_body lexbuf =
   __ocaml_lex_printing_token_body_rec lexbuf 2583
and __ocaml_lex_printing_token_body_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1218 "tools/coqdoc/cpretty.mll"
 ( let s = Buffer.contents token_buffer in
	  Buffer.clear token_buffer;
	  s )
# 4884 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1221 "tools/coqdoc/cpretty.mll"
        ( Buffer.add_string token_buffer (lexeme lexbuf);
	  printing_token_body lexbuf )
# 4890 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_printing_token_body_rec lexbuf __ocaml_lex_state

and inf_rules indents lexbuf =
   __ocaml_lex_inf_rules_rec indents lexbuf 2589
and __ocaml_lex_inf_rules_rec indents lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1228 "tools/coqdoc/cpretty.mll"
      ( inf_rules indents lexbuf )
# 4902 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1230 "tools/coqdoc/cpretty.mll"
      ( match indents with
        | Some ls -> doc_list_bol ls lexbuf
        | None -> doc_bol lexbuf )
# 4909 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1234 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf;  (* anything else must be the first line in a rule *)
        inf_rules_assumptions indents [] lexbuf)
# 4915 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_inf_rules_rec indents lexbuf __ocaml_lex_state

and inf_rules_assumptions indents assumptions lexbuf =
   __ocaml_lex_inf_rules_assumptions_rec indents assumptions lexbuf 2601
and __ocaml_lex_inf_rules_assumptions_rec indents assumptions lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1243 "tools/coqdoc/cpretty.mll"
      ( let line = lexeme lexbuf in
        let (spaces,_) = count_spaces line in
        let dashes_and_name = 
               cut_head_tail_spaces (String.sub line 0 (String.length line - 1))
        in
        let ldn = String.length dashes_and_name in
        let (dashes,name) = 
          try (let i = String.index dashes_and_name ' ' in
               let d = String.sub dashes_and_name 0 i in
               let n = cut_head_tail_spaces 
                        (String.sub dashes_and_name (i+1) (ldn-i-1))
               in
                 (d, Some n))
          with _ -> (dashes_and_name, None)
            
        in
          inf_rules_conclusion indents (List.rev assumptions) 
                               (spaces, dashes, name) [] lexbuf )
# 4944 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1262 "tools/coqdoc/cpretty.mll"
      ( let line = lexeme lexbuf in
        let (spaces,_) = count_spaces line in
        let assumption = cut_head_tail_spaces 
                            (String.sub line 0 (String.length line - 1))
        in
          inf_rules_assumptions indents ((spaces,assumption)::assumptions)
                                lexbuf )
# 4955 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_inf_rules_assumptions_rec indents assumptions lexbuf __ocaml_lex_state

and inf_rules_conclusion indents assumptions middle conclusions lexbuf =
   __ocaml_lex_inf_rules_conclusion_rec indents assumptions middle conclusions lexbuf 2611
and __ocaml_lex_inf_rules_conclusion_rec indents assumptions middle conclusions lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1276 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf;
        Output.inf_rule assumptions middle (List.rev conclusions);
        inf_rules indents lexbuf )
# 4969 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1280 "tools/coqdoc/cpretty.mll"
      ( let line = lexeme lexbuf in
        let (spaces,_) = count_spaces line in
        let conc = cut_head_tail_spaces (String.sub line 0 
                                                    (String.length line - 1))
        in
          inf_rules_conclusion indents assumptions middle 
                               ((spaces,conc) :: conclusions) lexbuf
      )
# 4981 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_inf_rules_conclusion_rec indents assumptions middle conclusions lexbuf __ocaml_lex_state

and st_start m lexbuf =
   __ocaml_lex_st_start_rec m lexbuf 2623
and __ocaml_lex_st_start_rec m lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1292 "tools/coqdoc/cpretty.mll"
      ( st_modname m lexbuf )
# 4993 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1294 "tools/coqdoc/cpretty.mll"
      ( None )
# 4998 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_st_start_rec m lexbuf __ocaml_lex_state

and st_modname m lexbuf =
   __ocaml_lex_st_modname_rec m lexbuf 2631
and __ocaml_lex_st_modname_rec m lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1298 "tools/coqdoc/cpretty.mll"
      ( if subtitle m (lexeme lexbuf) then
          st_subtitle lexbuf
        else
          None
      )
# 5014 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1304 "tools/coqdoc/cpretty.mll"
      ( None )
# 5019 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_st_modname_rec m lexbuf __ocaml_lex_state

and st_subtitle lexbuf =
   __ocaml_lex_st_subtitle_rec lexbuf 2675
and __ocaml_lex_st_subtitle_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1308 "tools/coqdoc/cpretty.mll"
      ( let st = lexeme lexbuf in
        let i = try Str.search_forward (Str.regexp "\\**)") st 0 with
                   Not_found ->
                     (eprintf "unterminated comment at beginning of file\n";
                      exit 1)
        in
          Some (cut_head_tail_spaces (String.sub st 0 i))
      )
# 5038 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1317 "tools/coqdoc/cpretty.mll"
      ( None )
# 5043 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_st_subtitle_rec lexbuf __ocaml_lex_state

;;

# 1320 "tools/coqdoc/cpretty.mll"
 
  let coq_file f m =
    reset ();
    let c = open_in f in
    let lb = from_channel c in
      (Index.current_library := m;
       Output.initialize ();
       Output.start_module ();
       Output.start_coq (); coq_bol lb; Output.end_coq ();
       close_in c)

  let detect_subtitle f m =
    let c = open_in f in
    let lb = from_channel c in
    let sub = st_start m lb in
      close_in c;
      sub

# 5069 "tools/coqdoc/cpretty.ml"
